// htconst.cc:  implementations of functions for height bounds
//////////////////////////////////////////////////////////////////////////
//
// Copyright 1990-2005 John Cremona
// 
// This file is part of the mwrank package.
// 
// mwrank is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation; either version 2 of the License, or (at your
// option) any later version.
// 
// mwrank is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with mwrank; if not, write to the Free Software Foundation,
// Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
// 
//////////////////////////////////////////////////////////////////////////
 
// Here we implement (1) Silverman, (2) CPS (Cremona/Prickett/Siksek)
// bounds on the difference between naive and canonical height.

#include "curve.h"
#include "compproc.h"
#include "matrix.h"
#include "subspace.h"
#include "cperiods.h"
#include "points.h"
#include "sieve_search.h"
#include "polys.h"
#include "curvemod.h"
#include "pointsmod.h"
#include "saturate.h"
#include "mwprocs.h"
#include "htconst.h"
#include "egr.h"

#include "realroots.h"

// Code for Silverman bound

double logplus(double x)
{
  double ax = fabs(x);
  if(ax<1) return 0;
  return log(ax);
}

double hj(const Curvedata& CD, double& realjay)
{
  bigint c4, c6, njay, djay;
  c4=getc4(CD);
  c6=getc6(CD);
  njay = pow(c4,3);
  djay = getdiscr(CD);
  if((djay==0)||(njay==0)) {realjay=0; return 0;}

  double g = I2double(gcd(njay,djay));
  double xnjay = I2double(njay)/g;
  double xdjay = I2double(djay)/g;

  realjay = xnjay/xdjay;

  double x = log(fabs(xnjay));
  double y = log(fabs(xdjay));

  if(x<y) return y;
  else    return x;
}

double silverman_bound(const Curvedata& CD)
{
  static double log2 = log(2.0);
  bigint b2 = getb2(CD);
  bigint delta = getdiscr(CD);
  double realjay;
  double hjay = hj(CD,realjay);

// NB the constant 1.922 = 2*0.961 below is from Bremner's correction 
// to Silverman's paper; Silverman has 0.973 givin 2*0.973 = 1.946.

  double mu = 1.922  + hjay/12
                     + log(fabs(I2double(delta)))/6
                     + logplus(realjay)/6
	             + logplus(I2double(b2)/12);
  
  if(b2!=0) mu += log2;

  return mu;
}


// Cremona-Prickett-Siksek height bound, August 2002
// NB: We assume a minimal model here!

//#define DEBUG_CPS
//#define TEST_CPS

double cps_real(const bigfloat& b2, const bigfloat& b4, const bigfloat& b6, const bigfloat& b8);

double egr_height_constant(const Curvedata& CD)
{ 
  return cps_real(I2bigfloat(getb2(CD)),I2bigfloat(getb4(CD)),
		     I2bigfloat(getb6(CD)),I2bigfloat(getb8(CD)));
}

double cps_bound(const Curvedata& CD)
{
  double bd = cps_real(I2bigfloat(getb2(CD)),I2bigfloat(getb4(CD)),
			  I2bigfloat(getb6(CD)),I2bigfloat(getb8(CD)));
#ifdef DEBUG_CPS
  cout<<"In cps_bound() for "<<(Curve)CD<<endl;
  cout<<"cps_real = "<<bd<<endl;
#endif
  CurveRed CR(CD);
  vector<bigint> plist = getbad_primes(CD);
  for(unsigned int i=0; i<plist.size(); i++)
    {
      bigint q=plist[i];
      if(getc_p(CR,q)==1) 
	{
#ifdef DEBUG_CPS
      cout<<"q = "<<q<<", alpha = 0 since c_q=0"<<endl;
      cout<<"sum so far = "<<bd<<endl;
#endif
	  continue;      
	}
      double alpha =0;
      int m, Kc = getKodaira_code(CR,q).code;
      switch (Kc%10){
      case 0: // Im
	m = Kc/10; 
	alpha = (m%2 ? double(m*m-1)/double(4*m) : double(m)/4);
	break;
      case 1: // I*m
	m = (Kc - 1)/10;  
	alpha = (m==0? 1 : (getc_p(CR,q)==2? 1 : double(m+4)/4));
	break;
      case 3: // III
	alpha = 0.5; 
	break;
      case 4: // IV
	alpha = double(2)/3; 
	break;
      case 5: // IV*
	alpha = double(4)/3; 
	break;
      case 6: // III*
	alpha = 1.5; 
	break;
      default: // II, II*: c_p=1
	break;
      };
      bd += alpha*log(double(I2long(q)));
#ifdef DEBUG_CPS
      cout<<"q = "<<q<<", alpha = "<<alpha<<", q-term = "<< alpha*log(double(I2long(q))) <<endl;
      cout<<"sum so far = "<<bd<<endl;
#endif
    }
  return bd;
}

// Implementation for the real place originally by Nigel Smart, here
// rewritten by JC, 22/8/02

bigfloat calc_dvd_inf(const bigfloat& b2, const bigfloat& b4, const bigfloat& b6, const bigfloat& b8);
bigfloat calc_dv_inf(const bigfloat& b2, const bigfloat& b4, const bigfloat& b6, const bigfloat& b8);
bigfloat old_calc_dvd_inf(const bigfloat& b2, const bigfloat& b4, const bigfloat& b6, const bigfloat& b8, const bigfloat& del);
bigfloat old_calc_dv_inf(const bigfloat& b2, const bigfloat& b4, const bigfloat& b6, const bigfloat& b8, const bigfloat& del);

double cps_real(const bigfloat& b2, const bigfloat& b4, const bigfloat& b6, const bigfloat& b8)
{
  bigfloat dv=calc_dv_inf(b2,b4,b6,b8);     
  bigfloat dvd=calc_dvd_inf(b2,b4,b6,b8); 
#ifdef TEST_CPS
  bigfloat del = -b2*b2*b8-8*b4*b4*b4-27*b6*b6+9*b2*b4*b6;
  bigfloat dv2=old_calc_dv_inf(b2,b4,b6,b8,del);
  bigfloat dvd2=old_calc_dvd_inf(b2,b4,b6,b8,del); 
  bigfloat diff = dv-dv2;
  if(!is_zero(diff)) 
    cout<<"old_calc_dv gives "<<dv2<<"\nwhile new gives "<<dv<<endl<<"diff = "<<diff<<endl;
  diff = dvd-dvd2;
  if(!is_zero(diff)) 
    cout<<"old_calc_dvd gives "<<dvd2<<"\nwhile new gives "<<dvd<<endl<<"diff = "<<diff<<endl;
#endif
#ifdef DEBUG_CPS
  cout << "dv=" << dv << endl;
  cout << "dvd=" << dvd << endl;
#endif
  bigfloat htc=to_bigfloat(0);
  if(dv==-1)
    {
      if(dvd==-1) htc = to_bigfloat(0);
      else 
	{
	  if(dvd>0) htc = -log(dvd)/3;
	  else
	    {
	      cerr<<"Precision problem in cps_real(): dvd = "<<dvd<<" but should be >0"<<endl;
	      cerr<<"Height constant will not be correct"<<endl;
	      htc=0;
	    }
	}
    }
  else
    if(dvd==-1) 
      {
	if(dv>0) htc = -log(dv)/3;
	else
	  {
	    cerr<<"Precision problem in cps_real(): dv = "<<dv<<" but should be >0"<<endl;
	    cerr<<"Height constant will not be correct"<<endl;
	    htc=0;
	  }
      }
    else 
      {
	bigfloat mindv=min(dv,dvd);
	if(mindv>0) htc = -log(mindv)/3;
	else
	  {
	    cerr<<"Precision problem in cps_real(): min(dv,dvd) = "<<mindv<<" but should be >0"<<endl;
	    cerr<<"Height constant will not be correct"<<endl;
	    htc=0;
	  }
      }
  
#ifdef DEBUG_CPS
  cout<<"cps_real() returns -log(min(dv,dvd))/3 = "<<htc<<endl;
#endif

#ifdef MPFP
  double ans;
  doublify(htc,ans);
  return ans;
#else
  return htc;
#endif
}

// coeff has length 5 but may start with leading zeros
// returns real roots in [-1,1]
vector<bigfloat> roots11( const vector<bigfloat>& coeff );

inline vector<bigfloat> set_coeff(const bigfloat& c0, const bigfloat& c1, const bigfloat& c2, const bigfloat& c3, const bigfloat& c4)
{
  vector<bigfloat> coeff(5);
  coeff[0]=c0;  coeff[1]=c1;  coeff[2]=c2;  coeff[3]=c3;  coeff[4]=c4;
  return coeff;
}

inline vector<bigfloat> set_coeff(const bigfloat& c0, const bigfloat& c1, const bigfloat& c2, const bigfloat& c3)
{
  vector<bigfloat> coeff(4);
  coeff[0]=c0;  coeff[1]=c1;  coeff[2]=c2;  coeff[3]=c3;
  return coeff;
}

inline vector<bigfloat> set_coeff(const bigfloat& c0, const bigfloat& c1, const bigfloat& c2)
{
  vector<bigfloat> coeff(3);
  coeff[0]=c0;  coeff[1]=c1;  coeff[2]=c2;
  return coeff;
}

inline vector<bigfloat> set_reverse_coeff(const bigfloat& c0, const bigfloat& c1, const bigfloat& c2, const bigfloat& c3, const bigfloat& c4)
{
  return set_coeff(c4,c3,c2,c1,c0);
}

template <class T >
inline ostream& operator<<(ostream& os, const std::set<T>& v);

vector<bigfloat> reals_in ( vector<bigcomplex>& v);
vector<bigfloat> reals_in_11 ( vector<bigcomplex>& v);

//  Procedure to calculate dv' for the infinite prime

bigfloat calc_dvd_inf(const bigfloat& b2, const bigfloat& b4, const bigfloat& b6, const bigfloat& b8)
{ 
  bigfloat rx;
  bigfloat dvd=to_bigfloat(0),x,x2,Fx,Gx;

#ifdef DEBUG_CPS
  cout<<"\nIn calc_dvd_inf"<<endl;
  cout<<"b2="<<b2<<"\nb4="<<b4<<"\nb6="<<b6<<"\nb8="<<b8<<endl;
#endif

  vector<bigfloat> rts; // holds real roots of F, G, etc
  std::set<bigfloat> crit_pts; // use a set, so we don't get any duplicates
  std::set<bigfloat> F_roots;

  crit_pts.insert(to_bigfloat(1));
  crit_pts.insert(to_bigfloat(-1));
  
#ifdef DEBUG_CPS
  cout<<"crit_pts = "<<crit_pts<<endl;
#endif

  // Put the roots of G,F',G',F+G,F-G into crit_pts:
  // Keep the roots of F separate (see comment below)

  // Roots of F
  //  rts=roots11(set_reverse_coeff(to_bigfloat(0),to_bigfloat(1),b2/4.0,b4/2.0,b6/4.0));
  rts=realroots11(set_coeff(b6,2*b4,b2,to_bigfloat(4),to_bigfloat(0)));
  F_roots.insert(rts.begin(),rts.end());
#ifdef DEBUG_CPS
  cout<<"relevant roots of F: "<<rts<<endl;
  cout<<"After adding roots of F, F_roots = "<<F_roots<<endl;
#endif

  // Roots of G
  //  rts=roots11(set_reverse_coeff(to_bigfloat(1),to_bigfloat(0),-b4,-2*b6,-b8));
  rts=realroots11(set_coeff(-b8,-2*b6,-b4,to_bigfloat(0),to_bigfloat(1)));
  crit_pts.insert(rts.begin(),rts.end());
#ifdef DEBUG_CPS
  cout<<"relevant roots of G: "<<rts<<endl;
  cout<<"After adding roots of G, crit_pts = "<<crit_pts<<endl;
#endif

  // Roots of F+G
  //  rts=roots11(set_reverse_coeff(to_bigfloat(1),to_bigfloat(4),b2-b4,2.0*(b4-b6),b6-b8));
  rts=realroots11(set_coeff(b6-b8,2.0*(b4-b6),b2-b4,to_bigfloat(4),to_bigfloat(1)));
  crit_pts.insert(rts.begin(),rts.end());
#ifdef DEBUG_CPS
  cout<<"relevant roots of F+G: "<<rts<<endl;
  cout<<"After adding roots of F+G, crit_pts = "<<crit_pts<<endl;
#endif

  // Roots of G-F // Change from NPS
  //  rts=roots11(set_reverse_coeff(to_bigfloat(1),to_bigfloat(-4),-(b2+b4),-2.0*(b4+b6),-(b6+b8)));
  rts=realroots11(set_coeff(-(b6+b8),-2.0*(b4+b6),-(b2+b4),to_bigfloat(-4),to_bigfloat(1)));
  crit_pts.insert(rts.begin(),rts.end());
#ifdef DEBUG_CPS
  cout<<"relevant roots of F-G: "<<rts<<endl;
  cout<<"After adding roots of F-G, crit_pts = "<<crit_pts<<endl;
#endif

  // Roots of G'
  //  rts=roots11(set_coeff(to_bigfloat(0),-2*b8,-3*b6,-b4,to_bigfloat(0)));
  rts=realroots11(set_coeff(-2*b8,-3*b6,-b4,to_bigfloat(0)));
  crit_pts.insert(rts.begin(),rts.end());
#ifdef DEBUG_CPS
  cout<<"relevant roots of G': "<<rts<<endl;
  cout<<"After adding roots of G', crit_pts = "<<crit_pts<<endl;
#endif

  // Roots of F'
  //  rts=roots11(set_coeff(to_bigfloat(0),2*b6,3*b4,b2,to_bigfloat(2)));
  rts=realroots11(set_coeff(2*b6,3*b4,b2,to_bigfloat(2)));
  crit_pts.insert(rts.begin(),rts.end());
#ifdef DEBUG_CPS
  cout<<"relevant roots of F': "<<rts<<endl;
  cout<<"After adding roots of F', crit_pts = "<<crit_pts<<endl;
#endif

  // Evaluate max(|F(x)|,|G(x)|) at each of the x-values in array
  // crit_pts for which F(x)>=0: First take max|G(x)| over the roots
  // of F in [-1,1] -- this avoids the possibility that evaluating f
  // at one of its roots gives a slightly negative value causing a
  // possibly relevant value of G(x) to be ignored: observation of
  // Samir 25/07/04

  int first=1;
  std::set<bigfloat>::const_iterator xi = F_roots.begin();
  while(xi!=F_roots.end())
    {
      x=*xi++;
      x2=x*x;
      Gx=abs(1-b4*x2-2.0*b6*x*x2-b8*x2*x2);
#ifdef DEBUG_CPS
      cout<<"x="<<x<<", Gx="<<Gx<<endl;
#endif
      if (first)  { dvd=Gx; first=0;}
      else if (dvd>Gx) { dvd=Gx; }
#ifdef DEBUG_CPS
      cout<<"dvd so far ="<<dvd<<endl;
#endif
    }
  xi = crit_pts.begin();
  while(xi!=crit_pts.end())
    { 
      x=*xi++;
      x2=x*x;
      Fx=(4.0*x+b2*x2+2.0*b4*x*x2+b6*x2*x2);
#ifdef DEBUG_CPS
      cout<<"x="<<x<<", Fx="<<Fx<<endl;
#endif
      if(Fx<0) continue;
      Gx=abs(1-b4*x2-2.0*b6*x*x2-b8*x2*x2);
#ifdef DEBUG_CPS
      cout<<"x="<<x<<", Gx="<<Gx<<endl;
#endif
      rx=max(Fx,Gx);
      if (first)  { dvd=rx; first=0;}
      else if (dvd>rx) { dvd=rx; }
#ifdef DEBUG_CPS
      cout<<"dvd so far ="<<dvd<<endl;
#endif
    }
  if(first) return to_bigfloat(-1);

  return dvd;

}

// Procedure to calculate dv for the infinite prime

bigfloat calc_dv_inf(const bigfloat& b2, const bigfloat& b4, const bigfloat& b6, const bigfloat& b8)
{
  bigfloat rx;
  bigfloat dv=to_bigfloat(0),x,x2,fx,gx;

#ifdef DEBUG_CPS
  cout<<"\nIn calc_dv_inf"<<endl;
  cout<<"b2="<<b2<<"\nb4="<<b4<<"\nb6="<<b6<<"\nb8="<<b8<<endl;
#endif

  vector<bigfloat> rts; // holds real roots of f, g, etc
  std::set<bigfloat> crit_pts; // use a set, so we don't get any duplicates
  std::set<bigfloat> f_roots;

  crit_pts.insert(to_bigfloat(1));
  crit_pts.insert(to_bigfloat(-1));
  
#ifdef DEBUG_CPS
  cout<<"crit_pts = "<<crit_pts<<endl;
#endif

  // Put the roots of g,f',g',f+g,f-g into crit_pts:
  // Keep the roots of f separate (see comment below)

  // Roots of f
  //  rts=roots11(set_coeff(to_bigfloat(0),to_bigfloat(1),b2/4.0,b4/2.0,b6/4.0));
  rts=realroots11(set_coeff(to_bigfloat(4),b2,2*b4,b6));
  f_roots.insert(rts.begin(),rts.end());
#ifdef DEBUG_CPS
  cout<<"relevant roots of f: "<<rts<<endl;
  cout<<"After adding roots of f, f_roots = "<<f_roots<<endl;
#endif

  // Roots of g
  //  rts=roots11(set_coeff(to_bigfloat(1),to_bigfloat(0),-b4,-2*b6,-b8));
  rts=realroots11(set_coeff(to_bigfloat(1),to_bigfloat(0),-b4,-2*b6,-b8));
  crit_pts.insert(rts.begin(),rts.end());
#ifdef DEBUG_CPS
  cout<<"relevant roots of g: "<<rts<<endl;
  cout<<"After adding roots of g, crit_pts = "<<crit_pts<<endl;
#endif

  // Roots of f+g
  //  rts=roots11(set_coeff(to_bigfloat(1),to_bigfloat(4),b2-b4,2.0*(b4-b6),b6-b8));
  rts=realroots11(set_coeff(to_bigfloat(1),to_bigfloat(4),b2-b4,2.0*(b4-b6),b6-b8));
  crit_pts.insert(rts.begin(),rts.end());
#ifdef DEBUG_CPS
  cout<<"relevant roots of f+g: "<<rts<<endl;
  cout<<"After adding roots of f+g, crit_pts = "<<crit_pts<<endl;
#endif

  // Roots of g-f // Change from NPS
  //  rts=roots11(set_coeff(to_bigfloat(1),to_bigfloat(-4),-(b2+b4),-2.0*(b4+b6),-(b6+b8)));
  rts=realroots11(set_coeff(to_bigfloat(1),to_bigfloat(-4),-(b2+b4),-2.0*(b4+b6),-(b6+b8)));
  crit_pts.insert(rts.begin(),rts.end());
#ifdef DEBUG_CPS
  cout<<"relevant roots of f-g: "<<rts<<endl;
  cout<<"After adding roots of f-g, crit_pts = "<<crit_pts<<endl;
#endif

  // Roots of g'
  //  rts=roots11(set_coeff(to_bigfloat(0),to_bigfloat(1),to_bigfloat(0),-b4/2.0,-b6/2.0));
  rts=realroots11(set_coeff(to_bigfloat(2),to_bigfloat(0),-b4,-b6));
  crit_pts.insert(rts.begin(),rts.end());
#ifdef DEBUG_CPS
  cout<<"relevant roots of g': "<<rts<<endl;
  cout<<"After adding roots of g', crit_pts = "<<crit_pts<<endl;
#endif

  // Roots of f'
  //  rts=roots11(set_coeff(to_bigfloat(0),to_bigfloat(0),to_bigfloat(1),b2/6,b4/6));
  rts=realroots11(set_coeff(to_bigfloat(6),b2,b4));
  crit_pts.insert(rts.begin(),rts.end());
#ifdef DEBUG_CPS
  cout<<"relevant roots of f': "<<rts<<endl;
  cout<<"After adding roots of f', crit_pts = "<<crit_pts<<endl;
#endif

  // Evaluate max(|f(x)|,|g(x)|) at each of the x-values in array
  // crit_pts for which f(x)>=0: First take max|g(x)| over the roots
  // of f in [-1,1] -- this avoids the possibility that evaluating f
  // at one of its roots gives a slightly negative value causing a
  // possibly relevant value of g(x) to be ignored: observation of
  // Samir 25/07/04

  int first=1;
  std::set<bigfloat>::const_iterator xi = f_roots.begin();
  while(xi!=f_roots.end())
    {
      x=*xi++;
      x2=x*x;
      gx=abs(x2*x2-b4*x2-2.0*b6*x-b8);
#ifdef DEBUG_CPS
      cout<<"x="<<x<<", gx="<<gx<<endl;
#endif
      if (first)  { dv=gx; first=0;}
      else if (dv>gx) { dv=gx; }
#ifdef DEBUG_CPS
      cout<<"dv so far ="<<dv<<endl;
#endif
    }
  xi = crit_pts.begin();
  while(xi!=crit_pts.end())
    { 
      x=*xi++;
      x2=x*x;
      fx=(4.0*x*x2+b2*x2+2.0*b4*x+b6);
#ifdef DEBUG_CPS
      cout<<"x="<<x<<", fx="<<fx<<endl;
#endif
      if(fx<0) continue;
      gx=abs(x2*x2-b4*x2-2.0*b6*x-b8);
#ifdef DEBUG_CPS
      cout<<"x="<<x<<", gx="<<gx<<endl;
#endif
      rx=max(fx,gx);
      if (first)  { dv=rx; first=0;}
      else if (dv>rx) { dv=rx; }
#ifdef DEBUG_CPS
      cout<<"dv so far ="<<dv<<endl;
#endif
    }
  if(first) return to_bigfloat(-1);
  return dv;
}


// PAST THIS POINT the Smart method is implemented

int interval_test(const bigfloat& x, const vector<bigfloat> rts, int debug=0);

vector<bigfloat> reals_in_interval ( vector<bigcomplex>& v, const vector<bigfloat> rts);

// Inserts from C in to S the elements which are real and between -1
// and +1
void include_real_11(std::set<bigfloat>& S, const vector<bigcomplex>& C);


// test if rr lies in one of the first numint intervals out of [i1l.i1r],[i2l.i2r]

int is_in_int(const bigfloat rr,const bigfloat i1l,const bigfloat i1r,
              const bigfloat i2l,const bigfloat i2r,const long numint)
{
if (numint>0)
  { if (rr<=i1r && rr>=i1l) { return 1; }
    if (numint==2 && rr<=i2r && rr>=i2l) { return 1; }
  }
return 0;
}

// test if rr lies in one of the numint intervals  [intervals[i][0],intervals[i][1]]

int is_in_int2(const bigfloat rr,bigfloat** intervals,const long numint)
{
  long i;
  for (i=0; i<numint; i++)
    { if ((rr>=intervals[i][0]) && (rr<=intervals[i][1]))
	{ return 1; }
    }
  return 0;
}

bigfloat old_calc_dv_inf(const bigfloat& b2, const bigfloat& b4, const bigfloat& b6, const bigfloat& b8, const bigfloat& del)
{
  bigfloat rr=to_bigfloat(0),rx,rn;

#ifdef DEBUG_CPS
  cout<<"\nIn old_calc_dv_inf"<<endl;
  cout<<"b2="<<b2<<"\nb4="<<b4<<"\nb6="<<b6<<"\nb8="<<b8<<endl;
#endif
  vector<bigcomplex> rt = solvecubic(b2/4,b4/2,b6/4);

  bigfloat i1l,i1u,i2l=to_bigfloat(0),i2u=to_bigfloat(0);  // Bounds on Intervals
  long i,numint;
  if (del<0)
    { for (i=0; i<3; i++)
	{ if (is_approx_zero(rt[i].imag()))
	     { rr=rt[i].real(); i=3; }
	}
      numint=0; 
      if (rr<=1.0);
	{ numint=1;  
          i1l=rr;  i1u=1.0;
	  if (rr<-1.0) { i1l=-1.0; }
	}
    }
  else
    { rn=rt[0].real();
      rx=rt[1].real();
	// Want rn<rr<rx
      if (rn>rx) { rr=rx; rx=rn; rn=rr; }
      rr=rt[2].real();
      if (rr<rn)
        { rr=rn; rn=rt[2].real(); }
      else if (rx<rr)
        { rr=rx; rx=rt[2].real(); }

      numint=2;
      i1l=rn; i1u=rr; i2l=rx; i2u=1.0;
	// Deal With Right Most Interval
      if (i2l>i2u) { numint=1; }
      if (i2l<-1.0) 
	{ numint=1;
	  i1l=-1.0; i1u=i2u;
        }
	// Now Deal With Left Most Interval
      if (i1u>1.0)  { i1u=1.0; }
      if (i1l<-1.0) { i1l=-1.0; }
      if (i1l>i1u) 
	{ numint=numint-1;
	  i1l=-1.0; i1u=i2u;
	}
    }

  if(numint==0) return to_bigfloat(-1);  // code for "infinity"

  bigfloat* te = new bigfloat[36];
  long tec=2*numint;
  if (numint>0)
    { te[0]=i1l; te[1]=i1u;
      if (numint==2)
        { te[2]=i2l; te[3]=i2u; }
    }

  if (numint!=0)
    { // Roots of g
      rt=solvequartic(to_bigfloat(0),-b4,-2.0*b6,-b8);
      for (i=0; i<4; i++)
        { if (is_approx_zero(rt[i].imag()))
	    { rr=rt[i].real();
              if (is_in_int(rr,i1l,i1u,i2l,i2u,numint))
		{ te[tec]=rr; tec=tec+1; }
            }
        }

      // Roots of f (again!)
      rt=solvecubic(b2/4.0,b4/2.0,b6/4.0);
      for (i=0; i<3; i++)
        { if (is_approx_zero(rt[i].imag()))
            { rr=rt[i].real();
              if (is_in_int(rr,i1l,i1u,i2l,i2u,numint))
                { te[tec]=rr; tec=tec+1; }
            }
        }

      // Roots of f+g
      rt=solvequartic(to_bigfloat(4),b2-b4,2.0*(b4-b6),b6-b8);
      for (i=0; i<4; i++)
        { if (is_approx_zero(rt[i].imag()))
            { rr=rt[i].real();
              if (is_in_int(rr,i1l,i1u,i2l,i2u,numint))
                { te[tec]=rr; tec=tec+1; }
            }
        }

      // Roots of g-f // Change from NPS
      rt=solvequartic(to_bigfloat(-4),-(b2+b4),-2.0*(b4+b6),-(b6+b8));
      for (i=0; i<4; i++)
        { if (is_approx_zero(rt[i].imag()))
            { rr=rt[i].real();
              if (is_in_int(rr,i1l,i1u,i2l,i2u,numint))
                { te[tec]=rr; tec=tec+1; }
            }
        }

      // Roots of f'
      if (-24.0*b4+b2*b2>=0.0)
        { rn=sqrt(-24.0*b4+b2*b2);
	  rr=(-b2+rn)/12.0;
	  if (is_in_int(rr,i1l,i1u,i2l,i2u,numint))
                { te[tec]=rr; tec=tec+1; }
	  rr=(-b2-rn)/12.0;
	  if (is_in_int(rr,i1l,i1u,i2l,i2u,numint))
                { te[tec]=rr; tec=tec+1; }
	}

      // Roots of g'
      rt=solvecubic(to_bigfloat(0),-b4/2.0,-b6/2.0);
      for (i=0; i<3; i++)
        { if (is_approx_zero(rt[i].imag()))
            { rr=rt[i].real();
              if (is_in_int(rr,i1l,i1u,i2l,i2u,numint))
                { te[tec]=rr; tec=tec+1; }
            }
        }

    }

  bigfloat dv=to_bigfloat(0);
  // Evaluate |f|, |g| at each of the tec(>0) x-values in array te:
  for (i=0; i<tec; i++)
    { rr=te[i]*te[i];
      rn=4.0*rr*te[i]+b2*rr+2.0*b4*te[i]+b6;
#ifdef DEBUG_CPS
      cout<<"x="<<te[i]<<", fx="<<rn<<endl;
#endif
      rn=abs(rn);
      rx=rr*rr-b4*rr-2.0*b6*te[i]-b8;
      rx=abs(rx);
#ifdef DEBUG_CPS
      cout<<"x="<<te[i]<<", gx="<<rx<<endl;
#endif
      if (rn>rx) { rx=rn; }
      if (i==0)  { dv=rx; }
      else if (dv>rx) { dv=rx; }
#ifdef DEBUG_CPS
      cout<<"dv so far ="<<dv<<endl;
#endif
    }
  delete[] te;

  return dv;
}

bigfloat old_calc_dvd_inf(const bigfloat& b2, const bigfloat& b4, const bigfloat& b6, const bigfloat& b8, const bigfloat& del)
{ 
  bigfloat rr=to_bigfloat(0),rn;

#ifdef DEBUG_CPS
  cout<<"\nIn old_calc_dvd_inf"<<endl;
  cout<<"b2="<<b2<<"\nb4="<<b4<<"\nb6="<<b6<<"\nb8="<<b8<<endl;
#endif
  vector<bigcomplex> rt = solvecubic(b2/4,b4/2,b6/4);
  long i,j,numrt;
  bigfloat rrt[4];
  rrt[0]=to_bigfloat(0);
  if (del<0)
    { for (i=0; i<3; i++)
        { if (is_approx_zero(rt[i].imag()))
             { rr=rt[i].real(); i=3; }
        }
      if (is_approx_zero(rr)) { numrt=1; }
      else                     { rrt[1]=1.0/rr; numrt=2; }
    }
  else
    { numrt=1;
      for (i=0; i<3; i++)
	{ if (!is_approx_zero(rt[i].real()))
	    { rrt[numrt]=1.0/rt[i].real(); 
	      numrt++;
            }
	}
    }
 
  // Sort the real roots 
  long fl=0;
  while (fl==0)
    { fl=1;
      for (i=0; i<numrt-1; i++)
        { if (rrt[i]>rrt[i+1])
	    { rr=rrt[i+1];
	      rrt[i+1]=rrt[i];
	      rrt[i]=rr;
	      fl=0;
	    }
        }
    }

  // Make Intervals
  bigfloat **intervals=new bigfloat*[3];
  for (i=0; i<3; i++)
    { intervals[i]=new bigfloat[2]; }

  long numint=0;
  if (b6<0.0)
    { for (i=0; i<numrt; i=i+2)
	{ intervals[numint][1]=rrt[numrt-i-1];
	  intervals[numint][0]=rrt[numrt-i-2];
	  numint=numint+1;
	}
    }
  else
    { intervals[numint][1]=rrt[numrt-1]+2;
      intervals[numint][0]=rrt[numrt-1];
      numint=numint+1;
      for (i=1; i<numrt; i=i+2)
        { intervals[numint][1]=rrt[numrt-i-1];
	  if (numrt-i-2>0) 
	    { intervals[numint][0]=rrt[numrt-i-2]; }
	  else
	    { intervals[numint][0]=-1.0; }
          numint=numint+1;
        }
    }
  for (i=0; i<numint; i++)
    { if (intervals[i][1]>1.0)  { intervals[i][1]=1.0; }
      if (intervals[i][0]<-1.0) { intervals[i][0]=-1.0; }
    }
  i=0;
  while (i<numint)
    { fl=0;
      while (fl==0 && i<numint)
	{ fl=1;
          if (intervals[i][0]>intervals[i][1])
             { fl=0; 
	       for (j=i; j<numint-1; j++)
		  { intervals[j][0]=intervals[j+1][0];
		    intervals[j][1]=intervals[j+1][1];
		  }
	       numint=numint-1;
	     }
	}
      i=i+1;
    }

  bigfloat *te=new bigfloat[36];
  long tec=2*numint;
  for (i=0; i<numint; i++)
     { te[2*i]=intervals[i][0];
       te[2*i+1]=intervals[i][1];
     }

  if (numint!=0)
    { // Roots of g
      rt=solvequartic(to_bigfloat(0),-b4,-2.0*b6,-b8);
      for (i=0; i<4; i++)
	{ if (!is_approx_zero(rt[i]))
	    { rt[i]=to_bigfloat(1)/rt[i];
              if (is_approx_zero(rt[i].imag()))
                 { rr=rt[i].real();
                   if (is_in_int2(rr,intervals,numint))
                     { te[tec]=rr; tec=tec+1; }
		 }
            }
        }

      // Roots of f
      rr=0.0;
      if (is_in_int2(rr,intervals,numint))
                     { te[tec]=rr; tec=tec+1; }
      rt=solvecubic(b2/4.0,b4/2.0,b6/4.0);
      for (i=0; i<3; i++)
        { if (!is_approx_zero(rt[i]))
            { rt[i]=to_bigfloat(1)/rt[i];
              if (is_approx_zero(rt[i].imag()))
                 { rr=rt[i].real();
                   if (is_in_int2(rr,intervals,numint))
                     { te[tec]=rr; tec=tec+1; }
                 }
            }
        }

      // Roots of f+g
      rt=solvequartic(to_bigfloat(4),b2-b4,2.0*(b4-b6),b6-b8);
      for (i=0; i<4; i++)
        { if (!is_approx_zero(rt[i]))
            { rt[i]=to_bigfloat(1)/rt[i];
              if (is_approx_zero(rt[i].imag()))
                 { rr=rt[i].real();
                   if (is_in_int2(rr,intervals,numint))
                     { te[tec]=rr; tec=tec+1; }
		 }
            }
        }

      // Roots of f-g
      rt=solvequartic(to_bigfloat(-4),-b2-b4,2.0*(-b4-b6),-b6-b8);
      for (i=0; i<4; i++)
        { if (!is_approx_zero(rt[i]))
            { rt[i]=to_bigfloat(1)/rt[i];
	      if (is_approx_zero(rt[i].imag()))
                 { rr=rt[i].real();
                   if (is_in_int2(rr,intervals,numint))
                     { te[tec]=rr; tec=tec+1; }
		 }
            }
        }

      // Roots of f'
      rt=solvecubic(b2/2.0,3.0*b4/2.0,b6);
      for (i=0; i<3; i++)
        { if (!is_approx_zero(rt[i]))
            { if (is_approx_zero(rt[i].imag()))
                { rr=rt[i].real();
                  if (is_in_int2(rr,intervals,numint))
                    { te[tec]=rr; tec=tec+1; }
		}
            }
        }

      // Roots of g'
      rr=0.0;
      if (is_in_int2(rr,intervals,numint))
          { te[tec]=rr; tec=tec+1; }
      bigfloat dr=9.0*b6*b6-8.0*b8*b4;
      if (dr>=0.0)
	{ dr=sqrt(dr);
	  rr=(3.0*b6+dr)/(-4.0*b8);
	  if (is_in_int2(rr,intervals,numint))
                { te[tec]=rr; tec=tec+1; }
	  rr=(3.0*b6-dr)/(-4.0*b8);
          if (is_in_int2(rr,intervals,numint))
                { te[tec]=rr; tec=tec+1; }
	}
    }

  bigfloat dvd=to_bigfloat(0);
  bigfloat f,g;
  for (i=0; i<tec; i++)
    { rr=te[i];
#ifdef DEBUG_CPS
      cout<<"x="<<te[i]<<", fx="<<rr<<endl;
#endif
      f=(((b6*rr+2.0*b4)*rr+b2)*rr+4.0)*rr;
      g=(((-b8*rr-2.0*b6)*rr-b4)*rr)*rr+1.0;
      f=abs(f); g=abs(g);
      rn=f;
#ifdef DEBUG_CPS
      cout<<"x="<<te[i]<<", gx="<<rn<<endl;
#endif
      if (g>f) { rn=g; }
      if (i==0)  { dvd=rn; }
      else if (dvd>rn) { dvd=rn; }
#ifdef DEBUG_CPS
      cout<<"dvd so far ="<<dvd<<endl;
#endif
    }
  delete[] te;
  for (i=0; i<3; i++)
    { delete[] intervals[i]; }
  delete[] intervals;

  return dvd;

}

// coeff has length 5 but may start with leading zeros
// returns real roots in [-1,1]
vector<bigfloat> roots11( const vector<bigfloat>& coeff )
{
  bigfloat a=coeff[0], a1;
  if(a!=0) 
    {
      a1=1/a;
      vector<bigcomplex> cr = solvequartic(a1*coeff[1],a1*coeff[2],a1*coeff[3],a1*coeff[4]);
      return reals_in_11(cr);
    }
  a=coeff[1];
  if(a!=0) 
    {
      a1=1/a;
      vector<bigcomplex> cr = solvecubic(a1*coeff[2],a1*coeff[3],a1*coeff[4]);
      return reals_in_11(cr);
    }
  vector<bigfloat> ans; // zero length
  a=coeff[2];
  if(a==0) cerr<<"Error in roots: degree<2"<<endl;
  bigfloat b=coeff[3], c=coeff[4], x;
  bigfloat d=b*b-4*a*c;
  if(d>=0) 
    {
      d=sqrt(d);  a1=1/(2*a);
      x=a1*(-b+d);
      if((x<=1)&&(x>=-1)) ans.push_back(x);
      x=a1*(-b-d);
      if((x<=1)&&(x>=-1)) ans.push_back(x);
    }
  return ans;
}

vector<bigfloat> reals_in ( vector<bigcomplex>& v)
{
  vector<bigfloat> vr;
  vector<bigcomplex>::iterator vi = v.begin();
  while(vi!=v.end()) 
    {
      if(is_real(*vi)) vr.push_back((*vi).real());
      vi++;
    }
  return vr;
}

vector<bigfloat> reals_in_11 ( vector<bigcomplex>& v)
{
  vector<bigfloat> vr;
  vector<bigcomplex>::iterator vi = v.begin();
  while(vi!=v.end()) 
    {
      if(is_real(*vi)) 
	{
	  bigfloat x = (*vi).real();
	  if((x<=1)&&(x>=-1)) vr.push_back(x);
	}
      vi++;
    }
  return vr;
}

int interval_test(const bigfloat& x, const vector<bigfloat> rts, int debug)
{
// rts will have size 1 or 3, and be ordered
  if(debug) cout<<"Interval test("<<x<<"), rts="<<rts<<endl;
  if(x>1) {if(debug) cout<<"\t returns 0\n"; return 0;}
  if(x<-1) {if(debug) cout<<"\t returns 0\n";  return 0;}
  int ans;
  if(rts.size()==1) 
    ans = (x>=rts[0]);
  else 
    ans =  ((x>=rts[0]) && (x<=rts[1])) || (x>=rts[2]); 
  if(debug) cout<<"\t returns "<<ans<<"\n";
  return ans;
}

template <class T >
inline ostream& operator<<(ostream& os, const std::set<T>& v)
{
  os <<"[";
  copy(v.begin(),v.end(), ostream_iterator<T>(cout, " "));
  os << "]";
  return os;
}

vector<bigfloat> reals_in_interval ( vector<bigcomplex>& v, const vector<bigfloat> rts)
{
  vector<bigfloat> vr;
  vector<bigcomplex>::iterator vi = v.begin();
  bigfloat x;
  while(vi!=v.end()) 
    {
      if(is_real(*vi)) 
	{
	  x=(*vi).real();
	  if(interval_test(x,rts,1)) vr.push_back(x);
	}
      vi++;
    }
  return vr;
}

// Inserts from C in to S the elements which are real and between -1
// and +1
void include_real_11(std::set<bigfloat>& S, const vector<bigcomplex>& C)
{
  bigfloat x;
  vector<bigcomplex>::const_iterator Ci = C.begin();
  while(Ci!=C.end())
    {
      if(is_real(*Ci))
	{
	  x = (*Ci).real();
	  if((x<=1)&&(x>=-1)) S.insert(x);
	}
      Ci++;
    }
}

inline int in_fund_region(const bigcomplex& z)
{
  return (imag(z)>0)&&(abs(z)>0.999)&&(abs(2*real(z))<1.001);
}

#define HTB_DEBUG

bigfloat lower_height_bound(const Curvedata& CD)
{
  static bigfloat log2 = log(to_bigfloat(2));
  bigcomplex w1, w2;
  Cperiods pers(CD);
  bigcomplex tau = gettau(pers);
#ifdef HTB_DEBUG
  cout<<"\nAfter normalizing, periods are: "<<pers<<endl;
  if(!in_fund_region(tau)) 
    cout<<"tau is not in the fundamental region!"<<endl;
#endif  
  bigfloat aq = abs(q(tau));
  pers.getwi(w1,w2);  //NB w1 is the minimal period here, not necessarily the real period!
  int realcase = abs(imag(w1))<0.0001;
#ifdef HTB_DEBUG
  cout<<"q  = "<<q(tau)<<endl;
  cout<<"aq = "<<aq<<endl;
  cout<<"w1 = "<< w1 << endl;
  cout<<"w2 = "<< w2 << endl;
  cout<<"real case?  ";if(realcase)cout<<"yes";else cout<<"no"; cout<<endl;
#endif  

  // Either (when w1 is real):
  // (1/6)log|Delta| - 2log2 - (1/6)log|q| - 4(|q|)/(1-|q|)

  // Or (when w1 is not real):
  // (1/6)log|Delta| - 2log2 + (1/12)log|q| - 2(1+|q|^2)/(1-|q|)

  bigfloat term1 = log(abs(I2bigfloat(getdiscr(CD))))/6;
  bigfloat term2 = - 2*log2;
  bigfloat term3, term4;

  if(realcase) 
    {
      term3 = -log(aq)/6;
      term4 = -4*(aq)/(1-aq);
    }
  else
    {
      term3 = log(aq)/12;
      term4 = -2*(1+aq*aq)/(1-aq);
    }
  bigfloat bound = term1 + term2 + term3 + term4;
#ifdef HTB_DEBUG
  cout<<"term 1 = "<<term1<<endl;
  cout<<"term 2 = "<<term2<<endl;
  cout<<"term 3 = "<<term3<<endl;
  cout<<"term 4 = "<<term4<<endl;
#endif
  return bound;
}


static long fact_tab[13] = {1, 1, 2, 6, 24, 120, 720, 5040, 40320, 362880, 3628800, 39916800, 479001600};

long factorial(long n) // OK for n up to 12
{
  if(n<2) return 1;
  if(n>12) 
    {
      cerr<<"factorial(long) called with n = "<<n<<" -- too big!"<<endl;
      exit(1);
    }
  //  return n*factorial(n-1);
  return fact_tab[n];
}

bigfloat factorial(bigfloat n)
{
  if(n<2) return to_bigfloat(1);
  if(n<13) return to_bigfloat(fact_tab[longify(n)]);
  return n*factorial(n-1);
}

bigfloat lattice_const(int r)
{
  static bigfloat one   = to_bigfloat(1);
  static bigfloat two   = to_bigfloat(2);
  static bigfloat three = to_bigfloat(3);
  static bigfloat four  = to_bigfloat(4);
  static bigfloat five  = to_bigfloat(5);
  static bigfloat six   = to_bigfloat(6);
  static bigfloat seven = to_bigfloat(7);
  static bigfloat c1 = one;
  static bigfloat c2 = sqrt(four/three);
  static bigfloat c3 = pow(two,one/three);
  static bigfloat c4 = sqrt(two);
  static bigfloat c5 = pow(two,three/five);
  static bigfloat c6 = two/pow(three,one/six);
  static bigfloat c7 = pow(two,six/seven);
  static bigfloat c8 = two;
  switch(r) {
  case 0: return c1;
  case 1: return c1;
  case 2: return c2;
  case 3: return c3;
  case 4: return c4;
  case 5: return c5;
  case 6: return c6;
  case 7: return c7;
  case 8: return c8;
  default: 
    {
      bigfloat xr = to_bigfloat(r);
      if(r%2)
	{
	  return factorial(xr)*sqrt(Pi())/(factorial((xr-1)/2)*(1<<r));
	}
      else // r is even
	{
	  return factorial(xr/2);
	}
    }
  }
}

point_min_height_finder::point_min_height_finder(Curvedata* EE, int egr, int verb)
  :E(EE), egr_flag(egr), verbose(verb) 
{
  min_ht=to_bigfloat(0);
  Pmin=Point(E);
  E -> getai(a1,a2,a3,a4,a6);
  iso = !((a1==0)&&(a3==0));
  c.resize(4);
  if(iso)
    {
      c[0]=16*getb6(*E);
      c[1]= 8*getb4(*E);
      c[2]=   getb2(*E);
      c[3]=1;
    }
  else
    {
      c[0]=a6;
      c[1]=a4;
      c[2]=a2;
      c[3]=1;
    }
}

int point_min_height_finder::process(const bigint& x, const bigint& y, const bigint& z) 
{
  bigint rz; isqrt(z,rz);
  bigint x1=x*rz, y1=y, z1=z*rz;
  if(iso)
    {
      y1 -= (a1*x1+4*a3*z1);
      x1 *= 2;
      z1 *= 8;
    }
  Point P(E, x1,y1,z1);
  if(P.isvalid())
    {
      if(order(P)<0) {
      int egr=1; bigint p0;
      if(egr_flag) egr=HasGoodReduction(P,getbad_primes(*E),p0);
      if(!egr)
	{
	  if(verbose)
	    cout<<"Found point "<<P
		<<" but ignoring as not egr (bad reduction at "<<p0<<")"<<endl;
	}
      else
	{
	  bigfloat hP=height(P);
	  if(is_zero(hP)) return 0;
	  if(verbose) 
	    cout<<"Found point "<<P<<" with height "<<hP<<endl;
	  all_points.push_back(P);
	  
	  if((min_ht==0)||(hP<min_ht)) 
	    {
	      if(verbose) 
		cout<<"New minimum height = "<<hP<<endl;
	      min_ht=hP;
	      Pmin=P;
	    }
	}
    }
    }
  else
    {
      cout<<"Raw point       x,y,z = "<<x<<", "<<y<<", "<<z<<endl;
      cout<<"converted point P = "<<P<<" --not on curve!"<<endl;
    }
  return 0;
}

void point_min_height_finder::search(bigfloat h_lim)
{
  qsieve s(this, 3, c, h_lim, (verbose>1)); 
  bigcomplex c1(I2bigfloat(c[2])),
    c2(I2bigfloat(c[1])),
    c3(I2bigfloat(c[0]));
  vector<bigcomplex> roots=solvecubic(c1,c2,c3);
  //  cout<<"solvecubic("<<c1<<","<<c2<<","<<c3<<") returns "<<roots<<endl;
  vector<double> bnd(3);
  int nrr=order_real_roots(bnd,roots);
  s.set_intervals(bnd,nrr,1);
  s.search();
}



