"""
Elliptic curve constructor
"""

#*****************************************************************************
#       Copyright (C) 2005 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************


import sage.rings.rings as rings

import ell_generic, ell_finite_field, ell_rational_field

def EllipticCurve(x, y=None):
    """
    There are several ways to construct elliptic curves:

        -- EllipticCurve([a1,a2,a3,a4,a6]): Elliptic curve with given
           a-invariants.  The invariants are coerced into a the parent
           of the first element.  If all are integers, they are coerced
           into the rational numbers.

        -- EllipticCurve([a4,a6]): Same as above, but a1=a2=a3=0.

        -- EllipticCurve(label): Returns the elliptic curve over Q
           from the Cremoa database with the given label.  The label
           is a string, such as"11A" or "37B2".

        -- EllipticCurve(R, [a1,a2,a3,a4,a6]): Create the elliptic curve
           over R with given a-invariants.  Here R can be an arbitrary
           ring.  Note that addition need not be defined.


    EXAMPLES:
    We illustrate creating elliptic curves.

        sage: EllipticCurve([0,0,1,-1,0])
        Elliptic Curve defined by y^2 + y = x^3 - x over Rational Field

        sage: EllipticCurve([GF(5)(0),0,1,-1,0])
        Elliptic Curve defined by y^2 + y = x^3 - x over Finite field of size 5

    Alternatively, one can create the curve over the finite field as follows:
        sage: EllipticCurve(GF(5), [0, 0,1,-1,0])
        Elliptic Curve defined by y^2 + y = x^3 - x over Finite field of size 5

        sage: E = EllipticCurve([CC(0),0,1,-1,0])
        sage: E
        Elliptic Curve defined by y^2 + y = x^3 - x over Complex Field
        sage: E.j_invariant()
        2988.97297297297297297


        sage: E = EllipticCurve(ZZ, [0, 0,1,-1,0])
        sage: E
        Elliptic Curve defined by y^2 + y = x^3 - x over Integer Ring

    Of course, arithmetic on elliptic curves over Z need not be defined:
        sage: P = E([0,0])
        sage: P + P + P + P
        (2, -3)
        sage: P + P + P + P + P
        Traceback (most recent call last):
        ...
        ArithmeticError: Point (1/4, -5/8) is not on curve.
    """
    if isinstance(x, rings.Ring):
        if isinstance(x, rings.RationalField):
            return ell_rational_field.EllipticCurve_rational_field(x, y)
        elif isinstance(x, rings.FiniteField):
            return ell_finite_field.EllipticCurve_finite_field(x, y)
        else:
            return ell_generic.EllipticCurve_generic(x, y)
        
    if isinstance(x, str):
        return ell_rational_field.EllipticCurve_rational_field(x)

    if not isinstance(x,list) and (len(x)==5 or len(x)==2):
        raise TypeError, "invalid input to EllipticCurve constructor"

    if isinstance(x[0], (rings.Rational, rings.Integer, int, long)):
        return ell_rational_field.EllipticCurve_rational_field(x, y)
    elif isinstance(x[0], (rings.FiniteFieldElement, rings.IntegerMod)):
        return ell_finite_field.EllipticCurve_finite_field(x, y)
    else:
        return ell_generic.EllipticCurve_generic(x, y)


if __name__ ==  '__main__':
    from sage.rings.rings import *
    ZZ = IntegerRing()
    QQ = RationalField()
    CC = ComplexField()
    import doctest, sys
    doctest.testmod(sys.modules[__name__])
