"""
Integers
"""

#*****************************************************************************
#       Copyright (C) 2004 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

doc="""
Integers
"""


import operator

include "gmp.pxi"
include "interrupt.pxi"  # ctrl-c interrupt block support

cdef class Integer(_element.Element)

import sage.rings.integer_ring
import sage.rings.coerce
import sage.rings.complex_field
import sage.rings.integer_mod
import _rational as rational
import sage.libs.pari

cdef public int set_mpz(Integer self, mpz_t value):
    mpz_set(self.value, value)

cdef set_from_Integer(Integer self, Integer other):
    mpz_set(self.value, other.value)

cdef set_from_int(Integer self, int other):
    mpz_set_si(self.value, other)

cdef public mpz_t* get_value(Integer self):
    return &self.value


cdef extern from "Python.h":
    void* PyMem_Realloc(void *p, size_t n)

cdef void* pymem_realloc(void *ptr, size_t old_size, size_t new_size):
    return PyMem_Realloc(ptr, new_size)

cdef void pymem_free(void *ptr, size_t size):
    PyMem_Free(ptr)

cdef extern from "gmp.h":
    void mp_set_memory_functions (void *(*alloc_func_ptr) (size_t), void *(*realloc_func_ptr) (void *, size_t, size_t), void (*free_func_ptr) (void *, size_t))


def _pmem_malloc():
    """
    Switch to using the Python heap for GMP memory management.

    We do *not* do this by default, since it leads to weird crashes,
    unfortunately.  It's better to stick with malloc.

    Here's an example that leads to a weird crash,

        import _integer
        _integer.pmem_malloc()
        Z = _integer.Integer
        a=Z(19)**Z(1000000); b=a*a; c=a.inverse_mod(b)
        
    """
    mp_set_memory_functions(PyMem_Malloc, pymem_realloc, pymem_free)


cdef class Integer(_element.Element):
    """
    The \\class{Integer} class represents arbitrary precision
    integers.  It derives from the \\class{Element} class, so
    integers can be used as ring elements anywhere in SAGE.

    \\begin{notice}
    The class \\class{Integer} is implemented in Pyrex, as a wrapper
    of the GMP \\code{mpz_t} integer type.
    \\end{notice}
    """
    def __new__(self, x=None):
        mpz_init(self.value)

    def __pyxdoc__init__(self):
        """
        You can create an integer from an int, long, string literal, or
        integer modulo N.

        EXAMPLES:
            sage: Integer(495)
            495
            sage: Integer('495949209809328523')
            495949209809328523
            sage: Integer(Mod(3,7))
            3


        Integers also support the standard arithmetic operations, such
        as +,-,*,/, \code{abs}, \code{mod}, \code{float}
        """
    def __init__(self, x=None):
        if x != None:
            if isinstance(x, Integer):
                set_from_Integer(self, x) 
            elif isinstance(x, int):
                mpz_set_si(self.value, x)
            elif isinstance(x, long):
                s = "%x"%x
                mpz_set_str(self.value, s, 16)
            elif isinstance(x, str):
                if x.isdigit() or x[1:].isdigit():
                    mpz_set_str(self.value, x, 0)
            elif isinstance(x, rational.Rational):
                if x.denominator() != 1:
                    raise TypeError, "Unable to coerce rational (=%s) to an Integer."%x
                set_from_Integer(self, x.numer())
                
            elif isinstance(x, sage.rings.integer_mod.IntegerMod):
                set_from_Integer(self, x.lift())
                
            elif isinstance(x, sage.libs.pari.gen):
                if x.type() == 't_INTMOD':
                    x = x.lift()
                # TODO: figure out how to convert to pari integer in base 16 ?
                s = str(x)
                if mpz_set_str(self.value, s, 0):
                    raise TypeError, "Unable to coerce PARI %s to an Integer."%x
            else:
                raise TypeError, "Unable to coerce %s (of type %s) to an Integer."%(x,type(x))
    

    def __reduce__(self):
        # This single line below took me HOURS to figure out.
        # It is the *trick* needed to pickle pyrex extension types.
        # The trick is that you must Python a pure Python function
        # as the first argument, and that function must return
        # the result of unpickling with the argument in the second
        # tuple as input. All kinds of problems happen
        # if we don't do this.
        return sage.rings.integer_ring.Z, (long(self),)

    cdef int cmp(self, Integer x):
        cdef int i
        i = mpz_cmp(self.value, x.value)
        if i < 0:
            return -1
        elif i == 0:
            return 0
        else:
            return 1

    def __cmp__(Integer self, Integer x):
        return self.cmp(x)

    def __richcmp__(Integer self, x, int op):
        cdef int n
        if not isinstance(x, Integer):
            try:
                x = Integer(x)
            except TypeError:
                n = sage.rings.coerce.cmp(self, x)
            else:
                n = self.cmp(x)
        else:
            n = self.cmp(x)
        if op == 0:
            return bool(n < 0)
        elif op == 1:
            return bool(n <= 0)
        elif op == 2:
            return bool(n == 0)
        elif op == 3:
            return bool(n != 0)
        elif op == 4:
            return bool(n > 0)
        elif op == 5:
            return bool(n >= 0)

    def copy(self):
        """
        Return a copy of the integer.
        """
        cdef Integer z
        z = Integer()
        set_mpz(z,self.value)
        return z

    def  __dealloc__(self):
        mpz_clear(self.value)
        
    def __repr__(self):
        return self.str()

    def __str_malloc(self, int base=10):
        """
        Return the string representation of \\code{self} in the given
        base.  (Uses malloc then PyMem.  This is actually slightly
        faster than self.str() below, but it is unpythonic to use
        malloc.)  However, self.str() below is nice because we know
        the size of the string ahead of time, and can work around a
        bug in GMP nicely.  There seems to be a bug in GMP, where
        non-2-power base conversion for very large integers > 10
        million digits (?) crashes GMP.
        """
        _sig_on
        cdef char *s
        s = mpz_get_str(NULL, base, self.value)
        t = str(s)
        free(s)
        _sig_off
        return t    
        
    def str(self, int base=10):
        r"""
        Return the string representation of \code{self} in the given
        base.

        \begin{notice}
        String representation of integers with more than about {\em four
        million digits} is not available in bases other than a power of
        2. This is because of a bug in GMP.  To obtain the base-$b$
        expansion of an integer $n$ use \code{n.str(b)}.
        \end{notice}
        """
        if base < 2 or base > 36:
            raise ValueError, "base (=%s) must be between 2 and 36"%base
        cdef size_t n
        cdef char *s
        n = mpz_sizeinbase(self.value, base) + 2
        if n > 4200000 and base != 2 and base != 4 and base != 8 and base != 16 and base != 32:
            raise RuntimeError, "String representation of integers with more than 4200000 digits is not available in bases other than a power of 2. This is because of a bug in GMP.  To obtain the base-b expansion of an integer n use n.str(b)."
        s = <char *>PyMem_Malloc(n)
        if s == NULL:
            raise MemoryError, "Unable to allocate enough memory for the string representation of an integer."
        _sig_on
        mpz_get_str(s, base, self.value)
        _sig_off
        return PyString_FromString(s)
    
    def set_si(self, signed long int n):
        """
        Coerces $n$ to a C signed integer if possible, and sets self
        equal to $n$.

        """
        mpz_set_si(self.value, n)

    def set_str(self, s, base=10):
        """
        Set self equal to the number defined by the string s in the
        given base.
        """
        _sig_on
        valid = mpz_set_str(self.value, s, base)
        _sig_off

        if valid != 0:
            raise ValueError, "invalid literal:" + s

    cdef void set_from_mpz(Integer self, mpz_t value):
        mpz_set(self.value, value)

    cdef mpz_t* get_value(Integer self):
        return &self.value

    def __add(Integer self, Integer other):
        cdef Integer x
        x = Integer()
        mpz_add(x.value, self.value, other.value)
        return x
        
    def __add__(x, y):
        if isinstance(x, Integer) and isinstance(y, Integer):
            return x.__add(y)
        return sage.rings.coerce.bin_op(x, y, operator.add)

    def __sub(Integer self, Integer other):
        cdef Integer x
        x = Integer()
        mpz_sub(x.value, self.value, other.value)
        return x
        
    def __sub__(x, y):
        """
        EXAMPLES:
            sage: a = Integer(5); b = Integer(3)
            sage: b - a
            -2
        """
        if isinstance(x, Integer) and isinstance(y, Integer):
            return x.__sub(y)
        return sage.rings.coerce.bin_op(x, y, operator.sub)

    def __mul(Integer self, Integer other):
        cdef Integer x
        x = Integer()

        _sig_on
        mpz_mul(x.value, self.value, other.value)
        _sig_off

        return x
            
    def __mul__(x, y):
        if isinstance(x, list):
            return x * int(y)
        if isinstance(x, Integer) and isinstance(y, Integer):
            return x.__mul(y)
        return sage.rings.coerce.bin_op(x, y, operator.mul)
        
    def __div(Integer self, Integer other):
        cdef Integer x
        #if mpz_divisible_p(self.value, other.value):
        #    x = Integer()
        #    mpz_divexact(x.value, self.value, other.value)
        #    return x
        #else:
        return rational.Rational(self)/rational.Rational(other)
        #raise ArithmeticError, "Exact division impossible."
        
    def __div__(x, y):
        if isinstance(x, Integer) and isinstance(y, Integer):
            return x.__div(y)
        return sage.rings.coerce.bin_op(x, y, operator.div)

    def __floordiv(Integer self, Integer other):
        cdef Integer x
        x = Integer()
        

        _sig_on
        mpz_fdiv_q(x.value, self.value, other.value)
        _sig_off


        return x


    def __floordiv__(x, y):
        if isinstance(x, Integer) and isinstance(y, Integer):
            return x.__floordiv(y)
        return sage.rings.coerce.bin_op(x, y, operator.floordiv)


    def __pow__(self, n, dummy):
        cdef Integer _self, _n
        cdef unsigned int _nval
        _n = Integer(n)
        if _n < 0:
            return Integer(1)/(self**(-_n))
        _self = integer(self)
        cdef Integer x
        x = Integer()
        _nval = _n

        _sig_on
        mpz_pow_ui(x.value, _self.value, _nval)
        _sig_off

        return x

    def __pos__(self):
        return self

    def __neg__(self):
        cdef Integer x
        x = Integer()
        mpz_neg(x.value, self.value)
        return x

    def __abs__(self):
        """
        """
        cdef Integer x
        x = Integer()
        mpz_abs(x.value, self.value)
        return x

    def __mod__(self, modulus):
        cdef Integer _modulus, _self
        _modulus = integer(modulus)
        if not _modulus:
            raise ZeroDivisionError, "Integer modulo by zero"            
        _self = integer(self)
        
        cdef Integer x
        x = Integer()

        _sig_on
        mpz_mod(x.value, _self.value, _modulus.value)
        _sig_off
        
        return x

    def powermod(self, exp, mod):
        """
        powermod(self, Integer exp, Integer mod):
        Compute self**exp modulo mod.
        """
        cdef Integer x, _exp, _mod
        _exp = Integer(exp); _mod = Integer(mod)
        x = Integer()

        _sig_on
        mpz_powm(x.value, self.value, _exp.value, _mod.value)
        _sig_off
        
        return x

    def powermodm_ui(self, unsigned long int exp, mod):
        cdef Integer x, _mod
        _mod = Integer(mod)
        x = Integer()

        _sig_on
        mpz_powm_ui(x.value, self.value, exp, _mod.value)
        _sig_off
        
        return x

    def __int__(self):
        cdef char *s
        s = mpz_get_str(NULL, 16, self.value)
        n = int(s,16)
        free(s)
        return n

    def __nonzero__(self):
        return not self.is_zero()

    def __long__(self):
        cdef char *s
        s = mpz_get_str(NULL, 16, self.value)
        n = long(s,16)
        free(s)
        return n

    def __float__(self):
        return mpz_get_d(self.value)

    def __hash__(self):
        return hash(self.str(16))

    def factor(self):
        """
        Return the prime factorization of the integer as a list of
        pairs $(p,e)$, where $p$ is prime and $e$ is a positive integer.
        """
        return sage.rings.integer_ring.factor(self)

    def lcm(self, n):
        """
        Returns the least common multiple of self and $n$.
        """
        cdef Integer _n
        cdef mpz_t x

        _n = Integer(n)
        mpz_init(x)

        _sig_on
        mpz_lcm(x, self.value, _n.value)
        _sig_off

        
        cdef Integer z
        z = Integer()
        mpz_set(z.value,x)
        mpz_clear(x)
        return z
        
    def denominator(self):
        """
        Return the denominator of the integer.
        """
        return ONE

    def is_one(self):
        """
        Returns \\code{True} if the integers is $1$, otherwise \\code{False}.
        """
        return mpz_cmp_si(self.value, 1) == 0

    def is_zero(self):
        """
        Returns \\code{True} if the integers is $0$, otherwise \\code{False}.
        """
        return mpz_cmp_si(self.value, 0) == 0

    def is_square(self):
        return bool(self.pari().issquare())

    def is_square_free(self):
        """
        Returns True if this integer is not divisible by the square of
        any prime and False otherwise.
        EXAMPLES:

        """
        return self.pari().issquarefree()

    def pari(self):
        if self._pari is None:
            # better to do in hex, but I can't figure out
            # how to input/output a number in hex in PARI!!
            # TODO: (I could just think carefully about raw bytes and make this all much faster...)
            self._pari = sage.libs.pari.pari(str(self))
        return self._pari

    def parent(self):
        """
        Return the ring $\\Z$ of integers.
        """
        return sage.rings.integer_ring.Z

    def isqrt(self):
        """
        Returns the integer floor of the square root of self, or raises
        an \\exception{ValueError} if self is negative.
        
        EXAMPLE:
            sage: a = Integer(5)
            sage: a.isqrt()
            2
        """
        if self < 0:
            raise ValueError, "square root of negative number not defined."
        cdef Integer x
        x = Integer()

        _sig_on
        mpz_sqrt(x.value, self.value)
        _sig_off
        
        return x
    
    def sqrt(self):
        """
        Returns the square root of self if self is a perfect square,
        or raises a \\exception{ValueError} exception otherwise.
        
        EXAMPLE:
            sage.: a = Integer(4)
            sage.: a.sqrt()
            2.0
        """
        return sage.rings.complex_field.CC(str(self)).sqrt()
    
        #if self < 0:
        #    raise ValueError, "square root of negative number not defined."
        #cdef Integer x
        #x = self.isqrt()
        #if x*x == self:
        #    return x
        #raise ValueError, "%s is not a perfect square."%self
    
    def xgcd(self, n):
        """
        Return a triple $g, s, t \\in\\Z$ such that
        $$
           g = s \\cdot \\mbox{\\rm self} + t \\cdot n.
        $$
        """
        cdef mpz_t g, s, t
        cdef object g0, s0, t0
        cdef Integer _n
        _n = Integer(n)
        
        mpz_init(g)
        mpz_init(s)
        mpz_init(t)

        _sig_on
        mpz_gcdext(g, s, t, self.value, _n.value)
        _sig_off
        
        g0 = Integer()
        s0 = Integer()
        t0 = Integer()
        set_mpz(g0,g)
        set_mpz(s0,s)
        set_mpz(t0,t)
        mpz_clear(g)
        mpz_clear(s)
        mpz_clear(t)
        return g0, s0, t0

    def __lshift__(Integer self, n):
        """
        """
        n = int(n)
        return self*(Integer(2)**n)   # todo: optimize

    def __rshift__(Integer self, n):
        """
        """
        n = int(n)
        return self.__floordiv(Integer(2)**n)    # todo: optimize

    def _and(Integer self, Integer other):
        cdef Integer x
        x = Integer()
        mpz_and(x.value, self.value, other.value)
        return x

    def __and__(x, y):
        if isinstance(x, Integer) and isinstance(y, Integer):
            return x._and(y)
        return sage.rings.coerce.bin_op(x, y, operator.and_)


    def _or(Integer self, Integer other):
        cdef Integer x
        x = Integer()
        mpz_ior(x.value, self.value, other.value)
        return x

    def __or__(x, y):
        if isinstance(x, Integer) and isinstance(y, Integer):
            return x._or(y)
        return sage.rings.coerce.bin_op(x, y, operator.or_)


    def __invert__(self):
        """
        """
        return Integer(1)/self    # todo: optimize
        

    def inverse_mod(self, n):
        """
        Returns the inverse of self modulo $n$, if this inverse exists.
        Otherwise, raises a \exception{ZeroDivisionError} exception.
        
        INPUT:
           self -- Integer
           n -- Integer
        OUTPUT:
           x -- Integer such that x*self = 1 (mod m), or
                raises ZeroDivisionError.
        IMPLEMENTATION:
           Call the mpz_invert GMP library function.
        EXAMPLES:
            sage: a = Integer(189)
            sage: a.inverse_mod(10000)
            4709
            sage: a.inverse_mod(-10000)
            4709
            sage: a.inverse_mod(1890)
            Traceback (most recent call last):
            ...
            ZeroDivisionError: Inverse does not exist.
            sage: a = Integer(19)**100000
            sage: b = a*a
            sage: c=a.inverse_mod(b)
            Traceback (most recent call last):
            ...
            ZeroDivisionError: Inverse does not exist.
        """
        cdef mpz_t x
        cdef object ans
        cdef int r
        cdef Integer m
        m = Integer(n)

        if m == 1:
            return Integer(0)

        mpz_init(x)

        _sig_on
        r = mpz_invert(x, self.value, m.value)
        _sig_off
        
        if r == 0:
            raise ZeroDivisionError, "Inverse does not exist."
        ans = Integer()
        set_mpz(ans,x)
        mpz_clear(x)
        return ans
        
    def gcd(self, n):
        """
        Return the greatest common divisor of self and $n$.
        """
        cdef Integer _n
        cdef mpz_t g
        cdef object g0

        _n = Integer(n)
        mpz_init(g)
        

        _sig_on
        mpz_gcd(g, self.value, _n.value)
        _sig_off

        g0 = Integer()
        set_mpz(g0,g)
        mpz_clear(g)
        return g0

    def crt(self, y, m, n):
        """
        Return the unique integer between $0$ and $mn$ that is
        congruent to the integer modulo $m$ and to $y$ modulo $n$.  We
        assume that~$m$ and~$n$ are coprime.
        """
        cdef object g, s, t
        cdef Integer _y, _m, _n
        _y = Integer(y); _m = Integer(m); _n = Integer(n)
        g, s, t = _m.xgcd(_n)
        if not g.is_one():
            raise ArithmeticError, "CRT requires that gcd of moduli is 1."
        # Now s*m + t*n = 1, so the answer is x + (y-x)*s*m, where x=self.
        return (self + (_y-self)*s*_m) % (_m*_n)

    
        
ONE = Integer(1)

def integer(x):
    if isinstance(x, Integer):
        return x
    return Integer(x)

def factorial(unsigned long int n):
    """
    Return the factorial $n!=1 \\cdot 2 \\cdot 3 \\cdots n$. 
    The input integer $n$ must fit in an \\code{unsigned long int}. 
    """
    cdef mpz_t x
    cdef Integer z
    
    mpz_init(x)
    
    _sig_on
    mpz_fac_ui(x, n)
    _sig_off
    
    z = Integer()
    set_mpz(z, x)
    mpz_clear(x)
    return z

def _test():
    import doctest, sys
    doctest.testmod(sys.modules[__name__], verbose=True)

if __name__ ==  '__main__':
    _test()
