r"""
Interface to GAP

This interface is extremely flexible, since it's exactly like typing
into the GAP interpreter, and anything that works there should work
here.  Unfortunately, it's sometimes much slower than using C-library
calls.

\begin{notice}
The \code{pexpect} module must be installed, which is not standard
with Python (but is included with SAGE), and the \code{gap} command
must be in your path.
\end{notice}

AUTHORS: David Joyner and William Stein

EXAMPLES:
    sage: import sage.interfaces.gap as gap
    sage: Gap = gap.GAP()
    sage: i = "G:=Group((1,2),(1,2,3));"
    sage: Gap(i)
    'Group([ (1,2), (1,2,3) ])'
    sage: i = "Size(G);"
    sage: Gap(i)
    '6'

Next we give a Groebner basis example via GAP.  This will not work
unless the GAP singular package is installed and correctly configured.

    sage.: i = 'LoadPackage("singular");'
    sage.: o = Gap(i)
    sage.: i = 'R:= PolynomialRing( Rationals, ["x","y","z"] );;'
    sage.: o = Gap(i)
    sage.: i = 'i:= IndeterminatesOfPolynomialRing(R);;'
    sage.: o = Gap(i)
    sage.: i = "x:= i[1];; y:= i[2];; z:= i[3];;"
    sage.: o = Gap(i)
    sage.: i = "r:= [ x*y*z -x^2*z, x^2*y*z-x*y^2*z-x*y*z^2, x*y-x*z-y*z];;"
    sage.: o = Gap(i)
    sage.: i = "I:= Ideal( R, r );;"
    sage.: o = Gap(i)
    sage.: i = "GroebnerBasis( I );"
    sage.: o = Gap(i)
    sage.: print o
    
"""


#*****************************************************************************
#       Copyright (C) 2005 David Joyner and William Stein
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import pexpect
_GAP = "gap -n"
_PROMPT = "gap> "

_objsMath = {}
class _uniqMath(object):
    def __new__(cls):
        if _objsMath.has_key(0):
            return _objsMath[0]
        R = object.__new__(cls)
        R.restart()
        _objsMath[0] = R
        return R

class GAP(_uniqMath):
    def __init__(self):
        pass

    def __repr__(self):
        return "Interface to GAP: Groups, Algorithms, and Programming"
    
    def __call__(self, line, raw=True):
        return self.cmd(line, raw=raw) 

    def restart(self):
        if hasattr(self, "__expect"):
            self.__expect.close()
        self.__expect = pexpect.spawn(_GAP)
        self.__expect.timeout = 999999999
        self.__expect.expect(_PROMPT)
        
    def cmd(self, line, raw=True):
        line = line.replace("\n",";")
        self.__expect.sendline(line)
        self.__expect.expect(_PROMPT)
        out = self.__expect.before
        if out.find("  ***")!=-1 and out.find("^")!=-1:
            raise RuntimeError, "\n"+out[out.find("  ***"):]
        i = out.find("\n"); 
        out = out[i+1:].replace("\r\n","\n")
        if raw: return out[:-1]
        j = out.find("=")
        return out[j+2:-2].replace("\r\n","\n")
    

