"""
NTL mat_ZZ class
"""

#*****************************************************************************
#       Copyright (C) 2005 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

include 'interrupt.pxi'
include 'decl.pxi'
include 'misc.pxi'

cimport ntl_ZZ
import ntl_ZZ
cimport ntl_ZZX
import ntl_ZZX

cdef class ntl_mat_ZZ:
    # see mat_ZZ.pxd for data members
    r"""
    The \class{mat_ZZ} class implements arithmetic with matrices over $\Z$.
    """

    def __init__(self, unsigned long nrows=0, unsigned long ncols=0, v=None):
        cdef unsigned long i, j
        cdef ntl_ZZ.ntl_ZZ tmp
        if nrows is _INIT:
            return
        self.x = new_mat_ZZ(nrows, ncols)
        self.__nrows = nrows
        self.__ncols = ncols
        if v != None:
            _sig_on
            for i from 0 <= i < nrows:
                for j from 0 <= j < ncols:
                    tmp = ntl_ZZ.ntl_ZZ(v[i*ncols+j])                    
                    mat_ZZ_setitem(self.x, i+1, j+1, <ZZ*> tmp.x)
            _sig_off
            

    def __dealloc__(self):
        del_mat_ZZ(self.x)

    def __repr__(self):
        _sig_on
        return string(mat_ZZ_to_str(self.x))

    def __mul__(ntl_mat_ZZ self, other):
        cdef ntl_mat_ZZ y
        if not isinstance(other, ntl_mat_ZZ):
            other = ntl_mat_ZZ(other)
        y = other
        _sig_on
        return make_mat_ZZ(mat_ZZ_mul(self.x, y.x))
            
    def __sub__(ntl_mat_ZZ self, other):
        cdef ntl_mat_ZZ y
        if not isinstance(other, ntl_mat_ZZ):
            other = ntl_mat_ZZ(other)
        y = other
        _sig_on
        return make_mat_ZZ(mat_ZZ_sub(self.x, y.x))

    def __add__(ntl_mat_ZZ self, other):
        cdef ntl_mat_ZZ y
        if not isinstance(other, ntl_mat_ZZ):
            other = ntl_mat_ZZ(other)
        y = other
        _sig_on
        return make_mat_ZZ(mat_ZZ_add(self.x, y.x))
            
    def __pow__(ntl_mat_ZZ self, long e, ignored):
        _sig_on
        return make_mat_ZZ(mat_ZZ_pow(self.x, e))

    def nrows(self):
        return self.__nrows
    
    def ncols(self):
        return self.__ncols

    def __setitem__(self, ij, x):
        cdef ntl_ZZ.ntl_ZZ y
        cdef int i, j
        if not isinstance(x, ntl_ZZ.ntl_ZZ):
            x = ntl_ZZ.ntl_ZZ(x)
        y = x
        if not isinstance(ij, tuple) or len(ij) != 2:
            raise TypeError, 'ij must be a 2-tuple'
        i, j = ij
        if i < 0 or i >= self.__nrows or j < 0 or j >= self.__ncols:
            raise IndexError, "array index out of range"
        mat_ZZ_setitem(self.x, i+1, j+1, <ZZ*> y.x)

    def __getitem__(self, ij):
        cdef int i, j
        if not isinstance(ij, tuple) or len(ij) != 2:
            raise TypeError, 'ij must be a 2-tuple'
        i, j = ij
        if i < 0 or i >= self.__nrows or j < 0 or j >= self.__ncols:
            raise IndexError, "array index out of range"
        return make_ZZ(mat_ZZ_getitem(self.x, i+1, j+1))

    def determinant(self, deterministic=True):
        _sig_on
        return make_ZZ(mat_ZZ_determinant(self.x, deterministic))

    def HNF(self, D=None):
        r"""
        The input matrix A=self is an n x m matrix of rank m (so n >=
        m), and D is a multiple of the determinant of the lattice L
        spanned by the rows of A.  The output W is the Hermite Normal
        Form of A; that is, W is the unique m x m matrix whose rows
        span L, such that
    
        - W is lower triangular,
        - the diagonal entries are positive,
        - any entry below the diagonal is a non-negative number
          strictly less than the diagonal entry in its column.
    
        This is implemented using the algorithm of [P. Domich,
        R. Kannan and L. Trotter, Math. Oper. Research 12:50-59,
        1987].

        TIMINGS:
        NTL isn't very good compared to MAGMA, unfortunately:
        
            sage.: import ntl
            sage.: a=MatrixSpace(Q,200).random_element()    # -2 to 2
            sage.: A=ntl.mat_ZZ(200,200)
            sage.: for i in xrange(a.nrows()):
               ....:     for j in xrange(a.ncols()):
               ....:         A[i,j] = a[i,j]
               ....:
            sage.: time d=A.determinant()
            Time.: 3.89 seconds
            sage.: time B=A.HNF(d)
            Time.: 27.59 seconds

        In comparison, MAGMA does this much more quickly:
        \begin{verbatim}
            > A := MatrixAlgebra(Z,200)![Random(-2,2) : i in [1..200^2]];
            > time d := Determinant(A);
            Time: 0.710
            > time H := HermiteForm(A);
            Time: 3.080
        \end{verbatim}
        
        Nonetheless, NTL is superior to PARI, which can't even do a
        100x100 HNF at all, as far as I can tell.
        """
        cdef ntl_ZZ.ntl_ZZ _D
        if D == None:
            _D = self.determinant()
        else:
            _D = ntl_ZZ.ntl_ZZ(D)
        _sig_on
        return make_mat_ZZ(mat_ZZ_HNF(self.x, <ZZ*>_D.x))

    def charpoly(self):
        return make_ZZX(mat_ZZ_charpoly(self.x));
    
    
            
cdef make_mat_ZZ(mat_ZZ* x):
    cdef ntl_mat_ZZ y
    _sig_off
    y = ntl_mat_ZZ(_INIT)
    y.x = x
    return y

cdef make_ZZ(ZZ* x):
    cdef ntl_ZZ.ntl_ZZ y
    _sig_off
    y = ntl_ZZ.ntl_ZZ(ntl_ZZ._INIT)
    y.set(<void*> x)
    return y

cdef make_ZZX(ZZX* x):
    cdef ntl_ZZX.ntl_ZZX y
    _sig_off
    y = ntl_ZZX.ntl_ZZX(_INIT)
    y.set(<void*> x)
    return y
