"""
Matrices

Elements of matrix spaces are of class \\code{Matrix}.  They can be
either sparse or dense, and can be defined over any base ring.

EXAMPLES:

We create the $2\\times 3$ matrix
$$
   \\left(\\begin{matrix} 1&2&3\\\\4&5&6 \\end{matrix}\\right)
$$
as an element of a matrix space over $\\Q$:

    sage: M = MatrixSpace(RationalField(),2,3)
    sage: A = M([1,2,3, 4,5,6])
    sage: A
    [1 2 3]
    [4 5 6]
    sage: A.parent()
    Full MatrixSpace of 2 by 3 dense matrices over Rational Field

We next change the top-right entry of $A$.  Note that matrix indexing
is $0$-based in SAGE, so the top right entry is $(0,2)$, which should
be thought of as ``row number $0$, column number 2''.

    sage: A[0,2] = 389
    sage: A
    [  1   2 389]
    [  4   5   6]

Also notice how matrices print.  All columns have the same width and
entries in a given column are right justified.   Next we compute the
reduced row echelon form of $A$.

    sage: A.echelon_form()
    [      1       0 -1933/3]
    [      0       1  1550/3]


"""


#*****************************************************************************
#       Copyright (C) 2005 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import sage.rings.rings
import sage.rings.arith 
import sage.misc.misc as misc
import sage.linalg.dense_matrix_pyx as dense_matrix_pyx
import sage.linalg.sparse_matrix
import sage.linalg.vector 
import sage.linalg.vector_space
import sage.linalg.matrix_space
import matrix_space
import sage.libs.pari
import sage.rings.element
import berlekamp_massey

class Matrix:
    """
    The \\class{Matrix} class is the base class for all matrix
    classes.  To create a \\class{Matrix}, first create a
    \\class{MatrixSpace}, then coerce a list of elements into the
    \\class{MatrixSpace}.  See the documentation of
    \\class{MatrixSpace} for more details.
    """
    def __init__(self, parent):
        if not isinstance(parent, matrix_space.MatrixSpace):
            raise TypeError, "parent must be a matrix_space.MatrixSpace"
        self.__parent = parent
        self.__nrows = parent.nrows()
        self.__ncols = parent.ncols()
        
    def __repr__(self):
        nr = self.nrows()
        nc = self.ncols()
        if nr == 0 or nc == 0:
            return "[]"
        # compute column widths
        S = [str(x) for x in self.list()]
        width = max([len(x) for x in S])
        rows = []
        m = 0
        
        # compute rows
        for r in xrange(nr):
            s = ""
            for c in xrange(nc):
                if c == nc-1:
                    sep=""
                else:
                    sep=" "
                entry = S[r*nc+c]
                if c == 0:
                    m = max(m, len(entry))
                entry = " "*(width-len(entry)) + entry
                s += entry + sep
            rows.append(s)
            
        # Remove leading spaces
##        n = width-m
##        if n > 0:
##            for r in xrange(nr):
##                rows[r] = rows[r][n:]
        
        # Put brackets around in a single string
        s = "\n".join(["[%s]"%row for row in rows])
        return s
        
    def latex(self):
        nr = self.nrows()
        nc = self.ncols()
        if nr == 0 or nc == 0:
            return "()"
        # compute column widths
        S = [str(x) for x in self.list()]
        width = max([len(x) for x in S])
        rows = []
        m = 0
        
        # compute rows
        for r in xrange(nr):
            s = ""
            for c in xrange(nc):
                if c == nc-1:
                    sep=""
                else:
                    sep="&"
                entry = misc.latex(S[r*nc+c])
                if c == 0:
                    m = max(m, len(entry))
                entry = " "*(width-len(entry)) + entry
                s += entry + sep
            rows.append(s)
            
        # Put brackets around in a single string
        s = "\\\\\n".join(["%s"%row for row in rows])
        return "\\left(\\begin{matrix}\n" + s + "\n\\end{matrix}\\right)"

    def __str__(self):
        return self.__repr__()

    def __hash__(self):
        return hash(self.__str__())
    
    ###################################################
    ## PARI
    ###################################################
    def pari(self):
        return sage.libs.pari.pari.matrix(self.nrows(), self.ncols(), self.list())


    ###################################################
    ## Access:
    ##   These *must* be overridden by each subclass
    ###################################################
    def __getitem__(self, ij):
        """
        USAGE:
            A[i, j] -- the i,j of A, and
            A[i]    -- the i-th row of A.
        """
        raise NotImplementedError
    
    def get(self, ij):
        """
        Entry access, but possibly without bounds checking (for
        efficiency).
        USAGE:
            A[i, j] -- the i,j of A, and
            A[i]    -- the i-th row of A.
        """
        return self[ij]
    
    def __setitem__(self, ij, x):
        """
        USAGE:
            A[i, j] = value -- set the (i,j) entry of A
            A[i] = value    -- set the ith row of A
        """
        raise NotImplementedError
        
    def set(self, ij, x):
        self[ij] = x
        
    ###################################################
    ## Convenience
    ###################################################
    def matrix_space(self, nrows=None, ncols=None, sparse=None):
        if nrows == None: nrows = self.nrows()
        if ncols == None: ncols = self.ncols()
        if sparse is None:
            sparse = sparse=self.is_sparse()
        return self.parent().matrix_space(nrows, ncols, sparse=sparse)
        
    def new_matrix(self, nrows=None, ncols=None, entries=0, 
                coerce_entries=True, copy=True):
        """
        Create a matrix in the parent of this space with the given
        number of rows, columns, etc.

        WARNING: This function called with no arguments returns the 0
        matrix by default, not the matrix self.
        """
        return self.matrix_space(nrows, ncols).matrix(
                entries, coerce_entries, copy)
        
    ###################################################
    ## Properties
    ###################################################
    def base_ring(self):
        return self.__parent.base_ring()
        
    def is_matrix(self):
        return True
        
    def ncols(self):
        return self.__ncols
    
    def nrows(self):
        return self.__nrows
    
    def parent(self):
        return self.__parent

    ###################################################
    ## Functions
    ###################################################
    def add_multiple_of_row(self, i, j, s):
        """
        Replace row i by s times row j.
        """
        for c in xrange(self.ncols()):
            self[i,c] += s*self[j,c]
            
    def add_multiple_of_column(self, i, j, s):
        """
        Replace column i by s times column j.
        """
        for r in xrange(self.nrows()):
            self[r,i] += s*self[r,j]
    
    def augment(self, other):
        """
        Return the augmented matrix of the form [self | other].
        """
        if not isinstance(other, Matrix):
            raise TypeError, "other must be a Matrix"
        if self.base_ring() != other.base_ring():
            raise TypeError, "base rings must be the same"
        if self.nrows() != other.nrows():
            raise TypeError, "number of rows must be the same"
        Z = self.new_matrix(ncols = self.ncols() + other.ncols())
        for r in xrange(self.nrows()):
            for c in xrange(self.ncols()):
                Z[r,c] = self[r,c]
        nc = self.ncols()
        for r in xrange(other.nrows()):
            for c in xrange(other.ncols()):
                Z[r,c + nc] = other[r,c]
        return Z
        
    def block_sum(self, other):
        """
        Return the block matrix that has self and other on the diagonal:
        [self |    0  ]
        [  0  | other ]
        """
        if not isinstance(other, Matrix):
            raise TypeError, "other must be a Matrix"
        top = self.augment(self.new_matrix(ncols=other.ncols()))
        bottom = other.new_matrix(ncols=self.ncols()).augment(other)
        return top.stack(bottom)
        
    def change_ring(self, ring):
        if ring == self.base_ring():
            return self
        M = sage.linalg.matrix_space.MatrixSpace(ring, self.__nrows, self.__ncols)
        return M(self.list())
        
    def charpoly(self):
        """
        Return the characteristic polynomial of this matrix.
        """
        time = misc.verbose(t=0)
        H = self.hessenberg_form()
        n = self.nrows()
        c = [0 for i in range(n+1)]
        R = sage.rings.rings.PolynomialRing(self.base_ring())
        c[0] = R(1)
        X = R.gen()
        for m in range(1,n+1):
            c[m] = (X - R(H[m-1,m-1]))*c[m-1]
            t = 1
            for i in range(1,m):
                t = t*H[m-i, m-i-1]
                c[m] = c[m] - R(t*H[m-i-1,m-1])*c[m-i-1]
        misc.verbose('computed characteristic polynomial of %sx%s matrix'%
                     (self.nrows(), self.ncols()), time)
        return c[n]

    def columns(self):
        """
        Returns the list of columns of self, as vectors.
        EXAMPLES:
            sage: A = MatrixSpace(RationalField(), 2,3) (xrange(6))
            sage: A
            [0 1 2]
            [3 4 5]
            sage: A.columns()
            [(0 3), (1 4), (2 5)]
        """
        return self.transpose().rows()
        
    def column_space(self):
        raise NotImplementedError
        
    def commutator(self, other):
        return self*other - other*self
      
    def copy(self):
        return self.new_matrix(entries=self.entries(), coerce_entries=False)
        
    def decompose(self, is_diagonalizable=False):
        """
        Returns the decomposition of the vector space under the right
        action of this matrix. 
        
        Let A be the matrix acting from the on the vector space V of
        column vectors.  Assume that A is square.  This function
        computes maximal subspaces W_1, ..., W_n corresponding to
        Galois conjugacy classes of eigenvalues of A.  More precisely,
        let f(X) be the characteristic polynomial of A.  this function
        computes the subspace $W_i = ker(g_(A)^n)$, where g_i(X) is an
        irreducible factor of f(X) and g_i(X) exactly divides f(X).
        If the optional parameter is_diagonalizable is True, then we
        let W_i = ker(g(A)), since then we know that ker(g(A)) =
        $ker(g(A)^n)$.
        
        OUTPUT:
            list -- list of Vector Spaces
        """
        assert self.is_square(), "self must be square"
        f = self.charpoly()
        E = []
        
        # Idea: For optimization, could compute powers of self
        #       up to max degree of any factor.  Then get g(self)
        #       by taking a linear combination.
        
        for g, m in f.factor():
            if is_diagonalizable:
                B = g(self)
            else:
                B = g(self) ** m
            E.append((B.kernel(), m==1))
        return E    
        
    def decompose_subspace(self, V, is_diagonalizable=False):
        """
        Suppose A is a matrix acting on a vector space V, and W is a
        vector subspace of V. Return decomposition of subspace as a list
        of pairs (W, is_irred) where is_irred is True if the charpoly
        of A acting on W is irreducible.
        """
        assert self.is_square(), "matrix must be square"
        assert V.base_ring() == self.base_ring(), "base rings must be compatible"
        assert V.degree() == self.ncols(), \
               "V must be a subspace of an n-dimensional space, where matrix is nxn"
        time = misc.verbose(t=0)

        # 1. Restrict
        B = self.restrict(V)
        time0 = misc.verbose("restrict -- ", time)

        # 2. Decompose restriction
        D = B.decompose(is_diagonalizable=is_diagonalizable)

        # 3. Lift decomposition to subspaces of ambient vector space.
        # Each basis vector for an element of D defines a linear combination
        # of the basis of W, and these linear combinations define the
        # corresponding subspaces of the ambient space V.
        
        misc.verbose("decompose -- ", time0)        
        C = V.matrix()
        
        # [todo]: Here I'm assuming that get an echelon basis,
        # but I haven't written down a proof of this!!!
        D = [(sage.linalg.vector_space.VectorSpace(
                [x*C for x in W.basis()]), is_irred)
                for W, is_irred in D]
        misc.verbose(t=time)
        return D
        
    def denominator(self):
        if self.nrows() == 0 or self.ncols() == 0:
            return 1
        R = self.base_ring()
        x = self.list()
        d = x[0].denominator()
        for y in x:
            d = d.lcm(y.denominator())
        return d

    def dense_matrix(self):
        """
        If this matrix is sparse, return a dense matrix with
        the same entries.  If this matrix is dense, return this
        matrix (not a copy). 

        NOTE: The definition of "dense" and "sparse" in SAGE have
        nothing to do with the number of nonzero entries.  Sparse
        and dense are properties of the underlying representation
        of the matrix.

        EXAMPLES:
            sage: A = MatrixSpace(QQ,2, sparse=True)([1,2,0,1])
            sage: A.is_sparse()
            True
            sage: B = A.dense_matrix()
            sage: B.is_sparse()
            False
            sage: A*B
            [1 4]
            [0 1]
            sage: A.parent()
            Full MatrixSpace of 2 by 2 sparse matrices over Rational Field
            sage: B.parent()
            Full MatrixSpace of 2 by 2 dense matrices over Rational Field
            sage: (A*B).parent()
            Full MatrixSpace of 2 by 2 sparse matrices over Rational Field
            sage: (B*A).parent()
            Full MatrixSpace of 2 by 2 dense matrices over Rational Field
        """
        if self.is_dense():
            return self
        M = self.matrix_space(self.nrows(), self.ncols(), sparse=False)
        return M(self.list())
        
    def sparse_matrix(self):
        """
        If this matrix is dense, return a sparse matrix with
        the same entries.  If this matrix is sparse, return this
        matrix (not a copy). 

        NOTE: The definition of "dense" and "sparse" in SAGE have
        nothing to do with the number of nonzero entries.  Sparse
        and dense are properties of the underlying representation
        of the matrix.

        EXAMPLES:
            sage: A = MatrixSpace(QQ,2, sparse=False)([1,2,0,1])
            sage: A.is_sparse()
            False
            sage: B = A.sparse_matrix()
            sage: B.is_sparse()
            True
            sage: A
            [1 2]
            [0 1]
            sage: B
            [1 2]
            [0 1]
            sage: A*B
            [1 4]
            [0 1]
            sage: A.parent()
            Full MatrixSpace of 2 by 2 dense matrices over Rational Field
            sage: B.parent()
            Full MatrixSpace of 2 by 2 sparse matrices over Rational Field
            sage: (A*B).parent()
            Full MatrixSpace of 2 by 2 dense matrices over Rational Field
            sage: (B*A).parent()
            Full MatrixSpace of 2 by 2 sparse matrices over Rational Field
        """
        if self.is_sparse():
            return self
        M = self.matrix_space(self.nrows(), self.ncols(), sparse=True)
        return M(self.dict())

    def dense_row(self, n):
        if n < 0 or n >= self.nrows():
            raise IndexError, "n must be between 0 and the number of rows"
        V = sage.linalg.vector_space.VectorSpace(\
                 self.base_ring(), self.ncols(), sparse=False)
        return V([self[n, i] for i in xrange(self.ncols())])
        
    def determinant(self):
        if self.nrows() != self.ncols():
            raise ArithmeticError, "Matrix must be square, but is %sx%s"%(
                self.nrows(), self.ncols())
        # TODO: Stupid slow method!
        return (-1)**self.nrows() * self.charpoly()[0]

    def det(self):
        return self.determinant()

    def dict(self):
        try:
            return self.__dict
        except:
            v = {}
            for ij in self.nonzero_positions():
                v[ij] = self.get(ij)
            self.__dict = v
            return self.__dict        
            
    def entries(self):
        """
        This must be defined in the derived class.  It is the
        underlying representation of elements, which may be a list or
        dict or something else.
        """
        raise NotImplementedError
        
    def fcp(self):
        """
        Return the factorization of the characteristic polynomial of self.
        """
        return self.charpoly().factor()
        
    def height(self):
        raise NotImplementedError

    def hermite_form(self):
        # TODO -- this isn't very good -- pari has much better hfn options.
        if self.nrows() == 0 or self.ncols() == 0:
            return self
        v = self.pari().mathnf()
        # We can do a 'fast' change of the above into a list of ints,
        # since we know all entries are ints:
        v = eval(str(v).replace(';',','))
        return self.new_matrix(entries=v, coerce_entries=True)

    def hessenberg_form(self):
        """
        Compute and return the Hessenberg form of self.
        ALGORITHM: See Henri Cohen's book.
        """
        if not self.is_square():
            raise ArithmeticError, "self must be square"
        n = self.nrows()
        tm = misc.verbose("Computing Hessenberg Normal Form of %sx%s matrix"%(n,n))
        h = self.copy()
        for m in range(1,n-1):
            # Search for a non-zero entry in column m-1
            i = False
            for r in range(m+1,n):
                if h[r,m-1] != 0:
                    i = r
                    break
            if i: 
                # Found a nonzero entry in column m-1 that is strictly below row m
                # Now set i to be the first nonzero position >= m in column m-1 
                if h[m,m-1] != 0:
                    i = m
                t = h[i,m-1]
                if i>m:
                    h.swap_rows(i,m)
                    # We must do the corresponding column swap to 
                    # maintain the characteristic polynomial (which is
                    # an invariant of Hessenberg form)
                    h.swap_columns(i,m)  
                # Now the nonzero entry in position (m,m-1) is t.
                # Use t to clear the entries in column m-1 below m.
                for j in range(m+1,n):
                    if h[j,m-1] != 0:
                        u = h[j,m-1]/t
                        h.add_multiple_of_row(j, m, -u)
                        # To maintain charpoly, do the corresponding column operation,
                        # which doesn't mess up the matrix, since it only changes
                        # column m, and we're only worried about column m-1 right now.
                        # Add u*column_j to column_m.
                        h.add_multiple_of_column(m, j, u)
        misc.verbose("Finished Hessenberg Normal Form of %sx%s matrix"%(n,n),tm)
        return h
        
    def image(self):
        raise NotImplementedError
    
    def is_dense(self):
        return self.parent().is_dense()
        
    def is_invertible(self):
        return self.is_square() and self.rank() == self.nrows()
        
    def is_sparse(self):
        return self.parent().is_sparse()
        
    def is_square(self):
        return self.nrows() == self.ncols()
        
    def kernel(self):
        """
        ALGORITHM:
        Elementary row ops don't change kernel, since they are just
        left multiplication by an invertible matrix, so we instead
        compute kernel of echelon form, which is easy.  More
        precisely, there is a basis vector of the kernel that
        corresponds to each non-pivot column.  That vector has a 1
        at the non-pivot column, 0's at all other non-pivot columns,
        and for each pivot column, the negative of the entry at the
        non-pivot column in the row with that pivot element.
        
        Since we view matrices as acting on the right, we compute
        the echelon form of the transpose.
        """

        R = self.base_ring()
        if isinstance(R, sage.rings.rings.IntegerRing):
            A = self.pari()
            B = A.matkerint()
            if B.ncols() == 1:
                B = A.matker().mattranspose()
            if B.ncols() == 0:
                return []
            assert B.ncols()
            # Now B is a basis for the LLL-reduced integer kernel as a
            # PARI object.  The basis vectors or B[0], ..., B[n-1],
            # where n is the dimension of the kernel.
            n = self.nrows()
            Q = sage.rings.rings.RationalField()
            V = sage.linalg.vector_space.VectorSpace(Q, n)
            basis = [V([Q(x) for x in b]) for b in B]
            return basis

        elif isinstance(R, sage.rings.rings.NumberField):
            A = self.pari()
            B = A.matker()
            n = self.nrows()
            V = sage.linalg.vector_space.VectorSpace(R, n)
            basis = [V([R(x) for x in b]) for b in B]
            return V.subspace(basis)

            
        if not self.base_ring().is_field():
            raise ArithmeticError, "not able to compute kernel over %s"%self.base_ring()
        
        E = self.transpose().echelon_form()
        pivots = E.pivots()
        pivots_set = set(pivots)
        basis = []
        VS = sage.linalg.vector_space.VectorSpace
        V = VS(R, self.nrows())
        ONE = R(1)
        for i in xrange(self.nrows()):
            if not (i in pivots_set):
                v = V(0)
                v[i] = ONE
                for r in range(len(pivots)):
                    v[pivots[r]] = -E[r,i]
                basis.append(v)
        return V.subspace(basis)
        
    def linalg_matrix(self):
        z = self.base_ring()(0)
        M = sage.linalg.matrix_space.MatrixSpace(self.base_ring(), self.nrows(), 
                self.ncols())([z for _ in xrange(self.nrows()*self.ncols())])
        for i,j in self.nonzero_positions():
            M[i,j] = self[i,j]
        return M
        
    def list(self):
        return [self[i,j] for i in xrange(self.nrows()) for j in xrange(self.ncols())]
        
    def maxspin(self, v):
        """
        Computes the largest integer n such that the list of vectors
        $S=[v, A(v), ..., A^n(v)]$ are linearly independent, and returns
        that list.

        INPUT:
            self -- Matrix
            v -- Vector
        OUTPUT:
            list -- list of Vectors

        ALGORITHM:
            The current implementation just adds vectors to a vector
            space until the dimension doesn't grow.  This could be
            optimized by directly using matrices and doing an
            efficient Echelon form.  Also, when the base is Q, maybe
            we could simultaneously keep track of what is going on in
            the reduction modulo p, which might make things much
            faster.
        """
        if v == 0: return []
        VS = sage.linalg.vector_space.VectorSpace
        V = VS([v])
        w = v
        S = [v]
        while True:
            w = w*self
            W = V + VS([w])
            if W.dimension() == V.dimension():
                return S
            V = W
            S.append(w)
        
    def nonpivots(self):
        """
        Return the list of i such that the i-th column of self 
        is NOT a pivot column of the reduced row echelon form 
        of self.
        OUTPUT:
            list -- sorted list of integers
        """
        X = set(self.pivots())
        return [j for j in xrange(self.ncols()) if not (j in X)]
        
    def nonzero_positions(self):
        """
        Returns the set of pairs (i,j) such that self[i,j] != 0.
        """
        z = self.base_ring()(0)
        return set([(i,j) for i in xrange(self.nrows()) \
                          for j in xrange(self.ncols()) \
                          if self[i,j] != z])
        
    def nonzero_positions_in_column(self, i):
        """
        Return the integers j such that self[j,i] is nonzero, i.e.,
        such that the j-th position of the i-th column is nonzero.
        """
        z = self.base_ring()(0)
        return set([j for j in xrange(self.nrows()) if self[j,i] != z])
        
    def nonzero_positions_in_row(self, i):
        """
        Return the integers j such that self[i,j] is nonzero, i.e.,
        such that the j-th position of the i-th row is nonzero.
        """
        z = self.base_ring()(0)        
        return set([j for j in xrange(self.ncols()) if self[i,j] != z])
        
    def nullity(self):
        # Use that rank + nullity = number of columns
        return self.ncols() - self.rank()
        
    #def numeric_array(self, typecode=Numeric.Float64):
    #    A = Numeric.array(self.list(), typecode=typecode)
    #    return Numeric.resize(A,(self.nrows(), self.ncols()))
        
    def pivots(self):
        """
        Return the i such that the i-th column of self is a
        pivot column of the reduced row echelon form of self.
        OUTPUT:
            list -- sorted list of integers
        """
        try:
            return self.__pivots
        except AttributeError:
            self.echelon_form()
            return self.__pivots
        
    def rank(self):
        return len(self.pivots())
        
    def rational_decomposition(self):
        raise NotImplementedError
        
    def rational_reconstruction(self):
        raise NotImplementedError
        
    def echelon_form(self, *args):
        """
        Returns the reduced row echelon form

        INPUT:
            matrix -- an element A of a MatrixSpace

        OUTPUT:
            matrix -- The reduced row echelon form of A.
            Note that self is *not* changed by this command.

        EXAMPLES:
           sage: from sage.linalg.matrix_space import MatrixSpace
           sage: from sage.rings.rings import RationalField
           sage: MS = MatrixSpace(RationalField(),2,3)
           sage: C = MS.matrix([1,2,3,4,5,6])
           sage: C.rank()
           2
           sage: C.nullity()
           1
           sage: C.echelon_form()
           [ 1  0 -1]
           [ 0  1  2]
        """
        if not self.base_ring().is_field():
            raise ArithmeticError, "Base ring of self must be a field."
        t = misc.verbose("Generic echelon...")
        pivot_positions = []
        start_row = 0
        A = self.copy()
        nrows = A.nrows()
        ncols = A.ncols()
        cleared_a_column = False
        for c in range(ncols):
            misc.verbose("column %s of %s"%(c, ncols),t, level=2)
            for r in range(start_row, nrows):
                if A.get((r,c)) != 0:
                    pivot_positions.append(c)
                    # Divide row r through by 1/A[r,c], so leading coefficient
                    # is 1.
                    A.rescale_row(r, ~A.get((r,c)))
                    # Swap
                    A.swap_rows(r,start_row)
                    # Clear column
                    cleared_a_column = True
                    for i in range(nrows):
                        if i != start_row:
                            x = A.get((i,c))
                            if x != 0:
                                # Add -x times start row to i
                                A.add_multiple_of_row(i, start_row, -x)
                if cleared_a_column:
                    start_row += 1
                    cleared_a_column = False
                    break
        self.__pivots = pivot_positions
        A.__pivots = list(self.__pivots)
        misc.verbose("Finished generic echelon.",t)
        return A
        
    def rescale_row(self, i, s):
        """
        Replace i-th row of self by s times i-th row of self.
        """
        if s == 1: return
        for j in xrange(self.ncols()):
            self[i,j] *= s
        
    def restrict(self, V, check=False):
        """
        Returns the matrix that defines the action of self on the invariant 
        subspace V.
        
        This function returns an nxn matrix, where V has dimension n.
        It does *not* check that V is in fact invariant under self.
        """
        if not isinstance(V, sage.linalg.vector_space.VectorSpace):
            raise TypeError, "V must be a Vector Space"
        if V.base_ring() != self.base_ring():
            raise TypeError, "base rings must be the same"
        if V.degree() != self.nrows():
            raise IndexError, "degree of V must equal number of rows of self"
        if V.dimension() == 0:
            return self.new_matrix(nrows=0, ncols=0)
        if V.is_full():
            return self.copy()
            
        B = V.matrix()
        P = B.pivots()
        return B*self.submatrix_from_columns(P)
        
    def row(self, i):
        V = sage.linalg.vector_space.VectorSpace(self.base_ring(), self.ncols())
        return V([self[i,j] for j in xrange(self.ncols())])
        
    def rows(self):
        return [self.row(i) for i in xrange(self.nrows())]

    def linear_combination_of_rows(self, v):
        R = self.rows()
        return sum([v[i]*R[i] for i in xrange(len(v))], self.row_space()(0))
    
    def row_space(self):
        return sage.linalg.vector_space.VectorSpace(self.rows())
        
    def smith_normal_form(self):
        raise NotImplementedError
        
    def sparse_columns(self):
        try:
            return self.__sparse_columns
        except:
            C = [[] for _ in xrange(len(self.ncols()))]
            for i, j in self.nonzero_positions():
                x = self.get(i,j)
                C[j].append((i,x))
            self.__sparse_columns = C
            return self.__sparse_columns
            
    def sparse_rows(self):
        try:
            return self.__sparse_rows
        except:
            R = [[] for _ in xrange(len(self.ncols()))]
            for i, j in self.nonzero_positions():
                x = self.get(i,j)
                R[i].append((j,x))
            self.__sparse_rows = r
            return self.__sparse_rows
        
    def stack(self, other):
        """
        Return the augmented matrix self on top of other:
           [ self  ]
           [ other ]
        """
        if not isinstance(other, Matrix):
            raise TypeError, "other must be a matrix"
        if self.base_ring() != other.base_ring():
            raise TypeError, "base rings must be the same"
        if self.ncols() != other.ncols():
            raise TypeError, "number of columns must be the same"
        Z = self.new_matrix(nrows = self.nrows() + other.nrows())
        for r in xrange(self.nrows()):
            for c in xrange(self.ncols()):
                Z[r,c] = self[r,c]
        nr = self.nrows()
        for r in xrange(other.nrows()):
            for c in xrange(other.ncols()):
                Z[r+nr,c] = other[r,c]
        return Z

    def str_latex(self, hfill=False):
        """
        Print a latex matrix representation of self.
        For example, if self is [1,2;3,4], the following 
        latex is generated:
        \\begin{verbatim}
            \\left(\\begin{array}1&2\\\\3&4
            \\end{array}\\right)
        \\end{verbatim}
        """
        nr = self.nrows()
        nc = self.ncols()
        s = "\\left(\\begin{array}{" + "c"*nc + "}\n"
        for i in xrange(nr):
            for j in xrange(nc):
                if hfill:
                    s += "\\hfill{}"
                s += str(self[i,j])
                if j < nc - 1:
                    s += "&"
                elif i < nr - 1:
                    s += "\\\\\n"
        s += "\n\\end{array}\\right)"
        return s

    def str_sparse_latex(self, variable="x"):
        """
        Return a latex string that represents this matrix as a sparse
        matrix.  The rows are printed as sums sum a_i x_i, where
        x is the variable.
        """
        nr = self.nrows()
        nc = self.ncols()
        s = "\\left(\\begin{align*}\n"
        for i in xrange(nr):
            v = [(j, self[i,j]) for j in xrange(nc) if self[i,j] != 0]
            for j in xrange(len(v)):
                s += "%s*%s_{%s}"%(v[j][1], variable, v[j][0])
                if j == 0:
                    s += "& + "
                elif j < len(v) - 1:
                    s += " + "
                else:
                    s += "\\\\\n"
        s += "\n\\end{align*}"
        return s
        
    def submatrix_from_columns(self, columns):
        """
        Return the submatrix of self of columns col[i] for i in
        the list of columns.
        """
        if not isinstance(columns, list):
            raise TypeError, "columns must be a list of integers"
        A = self.new_matrix(ncols = len(columns))
        k = 0
        for i in columns:
            i = int(i)
            #if not isinstance(i, int):
            #    raise TypeError, "each element of columns must be an int"
            if i < 0 or i >= self.ncols():
                raise IndexError, "column %s out of range"%i
            for r in xrange(self.nrows()):
                A[r,k] = self[r,i]
            k += 1
        return A
        
    def submatrix_from_rows(self, rows):
        """
        Return the submatrix of self of rows row[i] for i in
        the list of rows.
        """
        if not isinstance(rows, list):
            raise TypeError, "rows must be a list of integers"
        A = self.new_matrix(nrows = len(rows))
        k = 0
        for i in rows:
            #if not isinstance(i, int):
            #    raise TypeError, "each element of rows must be an int"
            i = int(i)
            if i < 0 or i >= self.nrows():
                raise IndexError, "row %s out of range"%i
            for c in xrange(self.ncols()):
                A[k,c] = self[i,c]
            k += 1
        return A
        
    def swap_columns(self, c1, c2):
        """
        Swap columns c1 and c2 of self.
        """
        nc = self.ncols()
        if c1 < 0 or c1 >= nc:
            raise IndexError, "c1 must satisfy 0<=c1<%s"%nc
        if c2 < 0 or c2 >= nc:
            raise IndexError, "c2 must satisfy 0<=c2<%s"%nc
        if c1 == c2: return
        for r in xrange(self.nrows()):
            a, b = self.get((r,c2)), self.get((r,c1))
            self.set((r,c1), a)
            self.set((r,c2), b)
        
    def swap_rows(self, r1, r2):
        """
        Swap rows r1 and r2 of self.
        """
        nr = self.nrows()
        if r1 < 0 or r1 >= nr:
            raise IndexError, "r1 must satisfy 0<=r1<%s"%nr
        if r2 < 0 or r2 >= nr:
            raise IndexError, "r2 must satisfy 0<=r2<%s"%nr
        if r1 == r2: return
        for c in xrange(self.ncols()):
            a, b = self.get((r2,c)), self.get((r1,c))
            self.set((r1,c), a)
            self.set((r2,c), b)
        
    def trace(self):
        """
        Return the trace of self, which is the sum of the
        diagonal entries of self.
        INPUT:
            self -- a square matrix
        OUTPUT:
            element of the base ring of self
        """
        if self.nrows() != self.ncols():
            raise ArithmeticError, "matrix must be square"
        R = self.base_ring()
        return misc.add([self[i,i] for i in xrange(self.nrows())],R(0))
        
    def transpose(self):
        """
        Returns the transpose of self, without changing self.
        EXAMPLES:
        We create a matrix, compute its transpose, and note that the
        original matrix is not changed.
            sage: M = MatrixSpace(RationalField(), 2)
            sage: A = M([1,2,3,4])
            sage: B = A.transpose()
            sage: print B
            [1 3]
            [2 4]
            sage: print A
            [1 2]
            [3 4]
        """
        T = self.new_matrix(self.ncols(), self.nrows())
        for r in xrange(self.nrows()):
            for c in xrange(self.ncols()):
                T[c,r] = self[r,c]
        return T
        
    def vector_matrix_multiply(self, v):
        """
        Returns the vector times matrix product.

        INPUT:
             matrix -- an element A of a MatrixSpace
             vector -- an element v of a VectorSpace

        OUTPUT:
            The the vector times matrix product v*A.

        EXAMPLES:
            sage: MS = MatrixSpace(RationalField(), 2,2)
            sage: B = MS.matrix([1,2,1,2])
            sage: V = VectorSpace(RationalField(), 2)
            sage: v = V.vector([1,2])
            sage: B.vector_matrix_multiply(v)     # computes v*B
            (3 6)
            sage: Bt = B.transpose()
            sage: Bt.vector_matrix_multiply(v)    # computes B*v
            (5 5)
        """
        if not isinstance(v, sage.linalg.vector.Vector):
            raise TypeError, "v must be a Vector"
        if self.base_ring() != v.base_ring():
            raise ArithmeticError, "base rings must be equal"
        if self.nrows() != v.degree():
            raise ArithmeticError, "number of rows of matrix must equal degree of vector"
        s = sage.linalg.vector_space.VectorSpace( 
                    self.base_ring(), self.ncols())(0)
        for i in xrange(self.nrows()):
            if v[i] != 0:
                s += v[i]*self.row(i)
        return s

    def iterates(self, v, n):
        """
        Let A be this matrix.   Return a list consisting of
        $$
           v, Av, A^2v, ..., A^(n-1)v.
        $$
        """
        if self.nrows() != self.ncols():
            raise ArithmeticError, "matrix must be square."
        X = [v]
        for _ in range(n-1):
            X.append(X[len(X)-1]*self)
        M = self.matrix_space(n, self.ncols())
        return M(X)

    def weidemann(self, i, t=0):
        """
        Application of Weiedemann's algorithm to the i-th standard
        basis vector.

        If the optimal parameter t is nonzero, use only the first t
        linear recurrence relations.
        """
        i = int(i); t=int(t)
        if self.nrows() != self.ncols():
            raise ArithmeticError, "matrix must be square."
        n = self.nrows()
        v = sage.linalg.vector_space.VectorSpace(self.base_ring(), n).gen(i)
        tm = misc.verbose('computing iterates...')
        cols = self.iterates(v, 2*n).columns()
        tm = misc.verbose('computed iterates', tm)
        f = None
        # Compute the minimal polynomial of the linear recurrence
        # sequence corresponding to the 0-th entries of the iterates,
        # then the 1-th entries, etc.
        if t == 0:
            R = range(n)
        else:
            R = [t]
        for i in R:
            tm = misc.verbose('applying berlekamp-massey')
            g = berlekamp_massey.berlekamp_massey(cols[i].list())
            misc.verbose('berlekamp-massey done', tm)
            if f is None:
                f = g
            else:
                f = f.lcm(g)
            if f.degree() == n:
                break
        return f
        
        
        
    ###################################################
    ## Arithmetic
    ###################################################
    def __abs__(self):
        return self.determinant()
    
    def __add__(self, right):
        if not isinstance(right, Matrix):
            right = self.parent()(right)
        elif self.parent() != right.parent():
            raise TypeError, "self and right must have the same parent"
        A = self.new_matrix()
        for i in xrange(self.nrows()):
            for j in xrange(self.ncols()):
                A[i,j] = self[i,j] + right[i,j]
        return A
            
    def __radd__(self, right):
        if not isinstance(self, Matrix):
            self = right.parent()(self)
        return self + right

    def __cmp__(self, right):
        if not isinstance(right, Matrix):
            right = self.parent()(right)
        elif self.parent() != right.parent():
            return -1
        for i in xrange(self.nrows()):
            for j in xrange(self.ncols()):
                c = self[i,j].__cmp__(right[i,j])
                if c: return c
        return 0
        
    def __invert__(self):
        if not self.is_square():
            raise "self must be a square matrix"
        A = self.augment(self.parent().identity_matrix())
        B = A.echelon_form()
        if B[self.nrows()-1,self.ncols()-1] != 1:
            raise ZeroDivisionError, "self is not invertible"
        return B.submatrix_from_columns(range(self.ncols(), 2*self.ncols()))

    def __div__(self, right):
        if not isinstance(right, Matrix):
            right = self.parent()(~right)
        else:
            right = ~right
        return self*right
        
    def __mod__(self, p):
        raise NotImplementedError
    
    def _scalar_multiply(self, x):
        if not x in self.base_ring():
            x = self.base_ring()(x)
        M = self.new_matrix()
        nc = self.ncols()
        for i in xrange(self.nrows()):
            for j in xrange(self.ncols()):
                M[i,j] = x * self[i,j]
        return M
    
    def _right_scalar_multiply(self, x):
        if not x in self.base_ring():
            x = self.base_ring()(x)
        M = self.new_matrix()
        nc = self.ncols()
        for i in xrange(self.nrows()):
            for j in xrange(self.ncols()):
                M[i,j] = self[i,j] * x
        return M

    def __mul__(self, right):
        if not isinstance(right, Matrix):
            return self._right_scalar_multiply(right)
        if self.base_ring() != right.base_ring():
            raise ArithmeticError, "base rings must be equal"
        if self.ncols() != right.nrows():
            raise ArithmeticError, "number of columns of self must equal number of rows of right."
        nr = self.nrows()
        snc = self.ncols()
        nc = right.ncols()
        M = self.new_matrix(nr, nc)
        if snc == 0: return M
        zero = self.base_ring()(0)
        for i in xrange(nr):
            for j in xrange(nc):
                M[i,j] = misc.add([self[i,k]*right[k,j] for k in range(snc)],zero)
        return M
        
    def __neg__(self):
        return self.__rmul__(self.base_ring()(-1))
        
    def __pos__(self):
        return self
        
    def __pow__(self, n):
        """
        EXAMPLES:
            sage: MS = MatrixSpace(RationalField(), 3, 3)
            sage: A = MS([0, 0, 1, 1, 0, '-2/11', 0, 1, '-3/11'])
            sage: A * A**(-1) == 1
            True
            sage: A**4
            [      -3/11     -13/121   1436/1331]
            [    127/121   -337/1331 -4445/14641]
            [    -13/121   1436/1331 -8015/14641]
        """
        if self.nrows() != self.ncols():
            raise ArithmeticError, "self must be square"
        #if not isinstance(n, int):
        #    raise TypeError, "n must be an int"
        n = int(n)
        if n == 0:
            return self.parent()(1)
        if n < 0:
            return (~self)**(-n)
        ans = self.parent()(1)
        apow = self
        while n != 0:
            if n%2 != 0:
                ans = ans * apow
            n /= 2
            if n == 0: break   # to save extra multiplication
            apow = apow * apow
        return ans
        
    def __rmul__(self, left):
        A = self.new_matrix()
        for i, j in self.nonzero_positions():
            A[i,j] = left*self[i,j]
        return A
        
    def __sub__(self, right):
        if not isinstance(right, Matrix):
            right = self.parent()(right)
        elif self.parent() != right.parent():
            raise TypeError, "self and other must have the same parent"
        A = self.new_matrix()
        for i in xrange(self.nrows()):
            for j in xrange(self.ncols()):
                A[i,j] = self[i,j] - right[i,j]
        return A
 
        raise NotImplementedError
    

#############################################
## Generic dense matrices over any field
############################################# 
def _convert_dense_entries_to_list(entries):
    # Create matrix from a list of vectors
    e = []
    for v in entries:
        e += v.list()
    copy = False
    return e

class Matrix_generic_dense(Matrix):
    """
    The \\class{Matrix_generic_dense} class derives from
    \\class{Matrix}, and defines functionality for dense matrices over
    any base ring.  Matrices are represented by a list of elements in
    the base ring, and element access operations are implemented in
    this class.
    """
    def __init__(self, parent, entries=0, 
                       coerce_entries=True, 
                       copy=True):
        Matrix.__init__(self, parent)
        
        R = self.base_ring()
        if not isinstance(entries, list):
            x = R(entries)
            zero = R(0)
            entries = [zero for _ in xrange(self.nrows()*self.ncols())]
            if x != zero:
                if self.nrows() != self.ncols():
                    raise TypeError, "scalar matrix must be square"
                
                nc = self.ncols()
                for i in xrange(self.nrows()):
                    entries[i*nc + i] = x
            
        if len(entries) > 0 and hasattr(entries[0], "is_vector"):
            entries = _convert_dense_entries_to_list(entries)
            
        if len(entries) != self.nrows() * self.ncols():
            if len(entries) != self.nrows() * self.ncols():
                raise ArithmeticError, "entries must be a list of length %s"%\
                       (self.nrows()*self.ncols())
        if coerce_entries:
            try:
                entries = [R(x) for x in entries]
            except TypeError:
                raise TypeError, "Unable to coerce entries to %s"%R
        elif copy:
            # Make a copy
            entries = list(entries) 
            
        self.__entries = entries
        
    def __getitem__(self, ij):
        """
        USAGE:
            A[i, j] -- the i,j of A, and
            A[i]    -- the i-th row of A.
        """
        if isinstance(ij, int):
            return self.row(ij)
        if not isinstance(ij, tuple) or len(ij) != 2:
            raise TypeError, "index must be a 2-tuple (i,j)"
        i,j=ij
        if i < 0 or i >= self.nrows():
            raise IndexError, "row index (i=%s) is out of range"%i
        if j < 0 or j >= self.ncols():
            raise IndexError, "colums index (j=%s) is out of range"%j
        return self.__entries[i*self.ncols() + j]
    
    def __setitem__(self, ij, value):
        """
        USAGE:
            A[i, j] = value -- set the (i,j) entry of A
            A[i] = value    -- set the ith row of A
        """
        if isinstance(ij, int):
            # Set the whole row.
            if ij < 0 or ij >= self.nrows():
                raise IndexError, "row index (=%s) is out of range"%ij                
            for j in self.ncols():
                self[ij,j] = value[j]
            return
        if not isinstance(ij, tuple) or len(ij) != 2:
            raise TypeError, "index must be a 2-tuple (i,j)"
        i,j=ij
        if i < 0 or i >= self.nrows():
            raise IndexError, "row index (i=%s) is out of range"%i
        if j < 0 or j >= self.ncols():
            raise IndexError, "colums index (j=%s) is out of range"
        self.__entries[i*self.ncols() + j] = self.base_ring()(value)

    def entries(self):
        return self.__entries
        
    def list(self):
        return self.__entries
        
#############################################
## Generic sparse matrices over any field
#############################################  
def _sparse_dot_product0(v, w):
    """
    v and w are sorted lists of pairs (i,x), where i is the position
    and x is a ring element.
    """
    s = 0
    i = 0
    j = 0
    while i < len(v) and j < len(w):
        a = v[i]; b = w[j]
        if a[0] < b[0]:
            i += 1
            continue
        elif a[0] > b[0]:
            j += 1
            continue
        s += a[1] * b[1]
        i += 1
        j += 1
    return s
    
def _sparse_dot_product(v, w):
    """
    v and w are dictionaries with integer keys.
    """
    x = set(v.keys()).intersection(set(w.keys()))
    return sum([v[k]*w[k] for k in x])
    
def _convert_sparse_entries_to_dict(entries):
    e = {}
    for i in xrange(len(entries)):
        for j, x in (entries[i].dict()).iteritems():
            e[(i,j)] = x
    return e
    
class Matrix_generic_sparse(Matrix):
    """
    The \\class{Matrix_generic_sparse} class derives from
    \\class{Matrix}, and defines functionality for dense matrices over
    any base ring.  A generic sparse matrix is represented using a
    dictionary with keys pairs $(i,j)$ and values in the base ring.
    """
    ##WARNING: We do not check that the i,j pairs satisfy
    ##    0 <= i < nrows,  0 <= j < ncols.
    def __init__(self, parent, 
                 entries=0, 
                 coerce_entries=True, 
                 copy=True):
        Matrix.__init__(self, parent)
        R = self.base_ring()
        
        if not isinstance(entries, (list, dict)):
            x = R(entries)
            zero = R(0)
            entries = {}
            if x != zero:
                if self.nrows() != self.ncols():
                    raise TypeError, "scalar matrix must be square"
                for i in xrange(self.nrows()):
                    entries[(i,i)] = x
            
        if isinstance(entries, list) and len(entries) > 0 and \
                hasattr(entries[0], "is_vector"):
                #isinstance(entries[0], sage.linalg.vector.Vector):
            entries = _convert_sparse_entries_to_dict(entries)

        if isinstance(entries, list):
            if len(entries) != self.nrows() * self.ncols():
                raise TypeError, "entries has the wrong length"
            x = entries
            entries = {}
            k = 0
            for i in xrange(self.nrows()):
                for j in xrange(self.ncols()):
                    if x[k] != 0:
                        entries[(i,j)] = x[k]
                    k += 1
            copy = False
            
        if not isinstance(entries, dict):
            raise TypeError, "entries must be a dict"
        if coerce_entries:
            try:
                for k, x in entries.iteritems():
                    entries[k] = R(x)
            except TypeError:
                raise TypeError, "Unable to coerce entries to %s"%R
        elif copy:
            # Make a copy
            entries = dict(entries)
        self.__entries = entries
        self.__zero = R(0)
        
    def __getitem__(self, ij):
        """
        USAGE:
            A[i, j] -- the i,j of A, and
            A[i]    -- the i-th row of A.
        """
        if isinstance(ij, int):
            return self.row(ij)
        if not isinstance(ij, tuple) or len(ij) != 2:
            raise TypeError, "index must be a 2-tuple (i,j)"
        # the ij might not be int's, which would cause a problem:
        ij = (int(ij[0]), int(ij[1]))   
        (i,j) = ij[0], ij[1]
        if i < 0 or i >= self.nrows():
            raise IndexError, "row index (i=%s) is out of range"%i
        if j < 0 or j >= self.ncols():
            raise IndexError, "colums index (j=%s) is out of range"%j
        if self.__entries.has_key(ij):
            return self.__entries[ij]
        return self.__zero
        
    def dict(self):
        return self.entries()
                
    def get(self, ij):
        """
        Like __getitem__ but with no type or bounds checking.
        For (i,j) access, returns 0 if access is out of bounds.
        """
        if isinstance(ij, int):
            return self.row(ij)
        if self.__entries.has_key(ij):
            return self.__entries[ij]
        return self.__zero
        
    def set(self, ij, x):
        """
        Like __setitem__ but with no type or bounds checking.  Only
        works for single entries, not whole rows.
        """
        if x == 0:
            if self.__entries.has_key(ij):
                del self.__entries[ij]
            return
        self.__entries[ij] = x
    
    def __setitem__(self, ij, value):
        """
        USAGE:
            A[i, j] = value -- set the (i,j) entry of A
            A[i] = value    -- set the ith row of A
        """
        if isinstance(ij, int):
            # Set the whole row.
            if ij < 0 or ij >= self.nrows():
                raise IndexError, "row index (=%s) is out of range"%ij
            try:
                for j in value.nonzero_positions():
                    self[ij,j] = value[j]
            except AttributeError:
                for j in self.ncols():
                    self[ij,j] = value[j]
            return
        
        if not isinstance(ij, tuple) or len(ij) != 2:
            raise TypeError, "index must be a 2-tuple (i,j)"
        # the ij might not be int's, which would cause a problem, so we coerce
        ij = (int(ij[0]), int(ij[1]))   
        i,j=ij
        if i < 0 or i >= self.nrows():
            raise IndexError, "row index (i=%s) is out of range"%i
        if j < 0 or j >= self.ncols():
            raise IndexError, "colums index (j=%s) is out of range"
        x = self.base_ring()(value)
        if x == 0:
            if self.__entries.has_key(ij):
                del self.__entries[ij]
            return
        self.__entries[ij] = x
        
    def __mul__(self, right):
        if not isinstance(right, Matrix):
            return self._right_scalar_multiply(right)        
        if not isinstance(right, Matrix_generic_sparse):
            right = self.matrix_space(right.nrows(), right.ncols())(right)
        if self.base_ring() != right.base_ring():
            raise ArithmeticError, "base rings must be equal"
        if self.ncols() != right.nrows():
            raise ArithmeticError, "number of columns of self must equal number of rows of right."
        nr = self.nrows()
        snc = self.ncols()
        nc = right.ncols()
        zero = self.base_ring()(0)
        E = {}
        rows = self.sparse_rows()
        cols = right.sparse_columns()
        for i in range(nr):
            row = rows[i]
            if len(row) == 0: continue
            for j in range(nc):
                x = _sparse_dot_product(row, cols[j])
                if x != 0: E[(i,j)] = x
        return self.new_matrix(nr, nc, entries = E, coerce_entries=False, copy=False)

    def hessenberg_form(self):
        """
        Return the Hessenberg form of this matrix.
        """
        # Current implementation is way too slow on sparse matrix.
        # On the dense it is reasonable.
        return self.dense_matrix().hessenberg_form()
        
    def sparse_columns(self):
        try:
            return self.__sparse_columns
        except:
            C = [{} for _ in xrange(self.ncols())]
            for ij, x in self.__entries.iteritems():
                C[ij[1]][ij[0]] = x
            self.__sparse_columns = C
            return self.__sparse_columns
            
    def sparse_rows(self):
        R = [{} for _ in xrange(self.nrows())]
        for ij, x in self.__entries.iteritems():
            R[ij[0]][ij[1]] = x
        return R
    
    def __rmul__(self, left):
        X = {}
        left = self.base_ring()(left)
        for ij, x in self.entries().iteritems():
            X[ij] = left*x
        return self.new_matrix(entries=X, copy=False, coerce_entries=False)
        
    def denominator(self):
        R = self.base_ring()
        x = self.entries()
        if len(x) == 0:
            return 1
        Z = x.iteritems()
        d = Z.next()[1].denominator()
        for _, y in Z:
            d = d.lcm(y.denominator())
        return d
        
    def entries(self):
        return self.__entries
        
    def nonzero_positions(self):
        """
        Returns the set of pairs (i,j) such that self[i,j] != 0.
        """
        return set(self.__entries.keys())
    
    def submatrix_from_columns(self, columns):
        """
        Return the submatrix of self of columns col[i] for i in
        the list of columns.
        """
        if not isinstance(columns, list):
            raise TypeError, "columns must be a list of integers"
            
        # Let A be this matrix and let B denote this matrix with
        # the columns deleted.
        # ALGORITHM:
        # 1. Make a table that encodes the function
        #          f : Z --> Z,
        #    f(j) = column of B where the j-th column of A goes
        # 2. Build new list of entries and return resulting matrix.
        C = set(columns)
        X = []
        j = 0
        for i in xrange(self.ncols()):
            if i in C:
                X.append(j)
                j += 1
            else:
                X.append(-1)    # column to be deleted.
        entries = {}
        E = self.entries()
        for ij in E.iterkeys():
            if ij[1] in C:
                i,j=ij
                entries[(i,X[j])] = E[ij]
                
        return self.new_matrix(ncols = len(columns), entries = entries, 
                    copy=False, coerce_entries=False)

    def submatrix_from_rows(self, rows):
        """
        Return the submatrix of self of rows row[i] for i in
        the list of rows.
        """
        if not isinstance(rows, list):
            raise TypeError, "rows must be a list of integers"
        R = set(rows)
        if not R.issubset(set(xrange(self.nrows()))):
            raise ArithmeticError, "Invalid rows."
        X = []
        i = 0
        for j in xrange(self.nrows()):
            if j in R:
                X.append(i)
                i += 1
            else:
                X.append(-1)    # row to be deleted.
        entries2 = []
        E = self.entries()
        for ij in E.iterkeys():
            if ij[0] in R:
                i,j=ij
                entries[(X[i],j)] = E[ij]
                
        return self.new_matrix(
                    nrows = len(rows),
                    entries = entries, 
                    copy=False, coerce_entries=False)
        
    def swap_rows(self, r1, r2):
        """
        Swap rows r1 and r2 of self.
        """
        nr = self.nrows()
        if r1 < 0 or r1 >= nr:
            raise IndexError, "r1 must satisfy 0<=r1<%s"%nr
        if r2 < 0 or r2 >= nr:
            raise IndexError, "r2 must satisfy 0<=r2<%s"%nr
        if r1 == r2: return
        for c in xrange(self.ncols()):
            self[r1,c], self[r2,c]  =  self[r2,c], self[r1,c]
        
    def transpose(self):
        X = {}
        for ij, x in self.__entries.iteritems():
            X[(ij[1],ij[0])] = x
        return self.new_matrix(nrows = self.ncols(), ncols = self.nrows(),
                           entries = X, copy=False, coerce_entries=False)
        

#############################################
## Dense matrices over the rational numbers
#############################################        
class Matrix_rational_dense(Matrix):
    """
    The \\class{Matrix_rational_dense} class derives from
    \\class{Matrix}, and defines functionality for dense matrices over
    the field $\\Q$ of rational numbers.
    """
    def __init__(self, 
                    parent, 
                    entries=0, 
                    coerce_entries=True, 
                    copy=True):
        Matrix.__init__(self, parent)
                
        if isinstance(entries, dense_matrix_pyx.Matrix_rational):
            if copy:
                entries = entries.copy()
            self.__matrix = entries
            return
            
        if not isinstance(entries, list):
            entries = self.base_ring()(entries)
            
        if isinstance(entries, list) and len(entries) > 0 and hasattr(entries[0],"is_vector"):
            #isinstance(entries[0], sage.linalg.vector.Vector):
            entries = _convert_dense_entries_to_list(entries)
        
        if isinstance(entries, list):
            if len(entries) != parent.nrows() * parent.ncols():
                raise TypeError, "entries has wrong length"

        self.__matrix = dense_matrix_pyx.Matrix_rational(
                    parent.nrows(), parent.ncols(), entries)
            
    def __getitem__(self, ij):
        """
        Use \\code{A[i,j]} to obtain the the $(i,j)$th entry of $A$, and
        \\code{A[i]} to obtain the $i$-th row of $A$.
        """
        if isinstance(ij, int):
            return self.row(ij)
        return self.__matrix[ij]
        
    def __setitem__(self, ij, x):
        """
        Use \\code{A[i,j]=x} to set the $(i,j)$th entry of $A$ to $x$,
        and \\code{A[i]=v} to set the $i$th row of $A$ to the entries
        of $v$.
        """
        if isinstance(ij, int):
            # Set the whole row.
            if ij < 0 or ij >= self.nrows():
                raise IndexError, "row index (=%s) is out of range"%ij                
            for j in self.ncols():
                self.__matrix[ij,j] = value[j]
            return
        self.__matrix[ij] = x
               
    def __mul__(self, right):
        if not isinstance(right, Matrix_rational_dense):
            if isinstance(right, Matrix):
                right = self.matrix_space(right.nrows(), right.ncols())(right)
            else:
                right = self.parent()(right)
        M = self.matrix_space(self.nrows(), right.ncols())
        A = self.__matrix * right.__matrix
        return Matrix_rational_dense(M, A, copy=False)

    def iterates(self, v, n):
        """
        Let A be this matrix.   Return a matrix with rows
        $$
           v, Av, A^2v, ..., A^(n-1)v.
        $$
        """
        if self.nrows() != self.ncols():
            raise ArithmeticError, "matrix must be square."
        I =  self.__matrix.iterates(list(v), n)
        M = self.matrix_space(n, self.ncols())
        return Matrix_rational_dense(M, I, copy=False)
        
        
    def hessenberg_form(self):
        """
        Return the Hessenberg form of this matrix.
        """
        time = misc.verbose(t=0)
        A = self.copy()
        A.__matrix.hessenberg_form()
        misc.verbose("finished computing Hessenberg form",time)
        return A

    def charpoly(self, bound=None):
        """
        Return the characteristic polynomial of this matrix, computed
        using the standard multimodular Hessenberg algorithm.

        The multimodular algorithm works by first computing a bound B,
        then computing the characteristic polynomial (using Hessenberg
        form) modulo enough primes so that their product is bigger
        than B.  One then uses the Chinese Remainder Theorem to
        recover the characteristic polynomial.  If the optional bound
        is specified, that bound is used for B instead of a potentially
        much worse general bound.
        """
        if self.nrows() != self.ncols():
            raise ArithmeticError, "the matrix must be square."
        d = self.denominator()
        if d != 1:
            A = self*d
        else:
            A = self
        if bound is None:
            # The following code (inspired by NTL's mat_poly_ZZ.c) computes
            # a bound that is vastly better than the Hadamard bound.  Victor
            # Shoup's comment, 'This bound is computed via interpolation
            # through complex roots of unity' means.  Reference: Mathieu and
            # Ford (1990, Section 6), 'On p-adic computation of the rational
            # form of a matrix.'.
            time = misc.verbose("computing bound")
            bound = 1
            zero = sage.rings.rings.IntegerRing()(0)
            for i in range(A.nrows()):
                t1 = zero
                for j in range(A.ncols()):
                    t1 += (self[i,j]*self[i,j]).numerator()
                t1 += abs(self[i,i].numerator()) + 1
                if t1 > 1:
                    t1 = t1.isqrt() + 1
                bound *= t1
            bound += 1
            misc.verbose("bound = %s"%bound, time)
        else:
            bound = int(bound)

        p = 46337   # todo -- don't hardcode, so on 64-bit machine will be twice as fast!
        
        prod = 1
        X = []
        primes = []
        while prod < bound:
            time = misc.verbose("using p = %s"%p)
            B = A.__matrix.matrix_modint_nodenom(p)
            primes.append(p)
            fmodp = B.charpoly()   # vector of entries of the characteristic polynomial
            misc.verbose("got charpoly mod %s"%p, time)
            X.append(fmodp)
            p = sage.rings.arith.previous_prime(p)
            prod *= p
            
        time = misc.verbose("final CRT")
        w = sage.rings.arith.CRT_vectors(X, primes)
        misc.verbose("done with CRT",time )
        modulus = misc.prod(primes)
        m = modulus // 2
        for i in range(len(w)):
            if w[i] >= m:
                w[i] = w[i] - modulus
        g = sage.rings.rings.PolynomialRing(sage.rings.rings.RationalField())(w)
        f = ((1/d)**self.nrows()) * g.rescale(d)
        return f
        
    def list(self):
        return self.__matrix.list()
        
    def entries(self):
        return self.__matrix
        
    def pivots(self):
        try:
            return self.__pivots
        except AttributeError:
            self.echelon_form()
            return self.__pivots
        
    def echelon_form(self):
        A = self.__matrix.echelon_modular()
        self.__pivots = A.pivots()
        B = Matrix_rational_dense(self.parent(), A)
        B.__pivots = list(self.__pivots)
        return B
        
        
#############################################
## Sparse matrices over the rational numbers
#############################################
class Matrix_rational_sparse(Matrix_generic_sparse):
    """
    The \\class{Matrix_rational_sparse} class derives from
    \\class{Matrix}, and defines functionality for sparse matrices
    over the field $\\Q$ of rational numbers.
    """
    def __init__(self, 
                    parent, 
                    entries=0, 
                    coerce_entries=True, 
                    copy = True):
        Matrix_generic_sparse.__init__(self, parent, entries, 
                        coerce_entries, copy)
            
    def _sparse_list(self):
        try:
            return self.__sparse_list
        except AttributeError:
            self.__sparse_list = [(ij[0], ij[1], x) for 
                ij, x in self.entries().iteritems()]
            self.__sparse_list.sort()
        return self.__sparse_list
            
    def _sparse_matrix_rational(self):
        try:
            return self.__sparse_matrix_rational
        except AttributeError:
            self.__sparse_matrix_rational = sage.linalg.sparse_matrix.SparseMatrix(
                self.base_ring(), self.nrows(), self.ncols(), self._sparse_list(), 
                coerce=False, sort=False, copy=False)
            return self.__sparse_matrix_rational

    def pivots(self):
        try:
            return self.__pivots
        except AttributeError:
            self.echelon_form()
            return self.__pivots
        
    def echelon_form(self, height_guess=None):
        t0 = misc.verbose()
        A = self._sparse_matrix_rational()
        t = misc.verbose("start")
        B = A.echelon_form(height_guess=height_guess)
        t1=misc.verbose("done", t)
        self.__pivots = B.pivots()
        X = B.entries()
        Y = {}
        for i,j,x in X:
            Y[(i,j)] = x
        C = self.new_matrix(coerce_entries=False, entries = Y, copy=False)
        C.__pivots = list(self.__pivots)
        misc.verbose("overhead", t0+(t1-t))
        return C

    def hessenberg_form(self):
        """
        Return the Hessenberg form of this sparse matrix.
        """
        return self.dense_matrix().hessenberg_form()

    def charpoly(self):
        """
        Return the characteristic polynomial of this matrix. 
        """
        return self.dense_matrix().charpoly()
    
        
#############################################
import sage.rings.polynomial_ring
def berlekamp(v):
    """
    Find minimal polynomial of a linear recurrence sequence.
    """
    if len(v) % 2 != 0 or len(v) == 0:
        raise TypeError, "len of v must be even and positive"
    n = len(v)/2
    S = sage.rings.polynomial_ring.PolynomialRing(v[0].parent())
    x = S._0
    R0 = x**(2*n)
    R1 = S(v)
    V0 = S(0)
    V1 = S(1)
    while R1.degree() >= n:
        Q, R = R0/R1   
        print "\nQ = ", Q
        print "R = ", R
        assert R0 == Q*R1 + R
        V0, V1, R0, R1 = V1, V0-Q*V1, R1, R
        print "V0 = ", V0
        print "V1 = ", V1
        print "R0 = ", R0
        print "R1 = ", R1
    d = max(V1.degree(), 1+R1.degree())
    print "d = ", d
    P = V1.reverse()
    print "p rev = ", P
    P = x**(d-P.degree())*P
    c = P.leading()
    print "c = ", c
    return (1/c)*P
    
        
#############################################
## Unit Tests 
#############################################
## def test_rank_rational_dense(n=10):
##     M = sage.linalg.matrix_space.MatrixSpace(rings.RationalField(),n)
##     A = M(range(n*n))
##     assert A.rank() == 2
##     A = M(range(n*n)).transpose()
##     assert A.rank() == 2
    
## def test_rank_rational_sparse(n=10):
##     M = sage.linalg.matrix_space.MatrixSpace(rings.RationalField(),n, sparse=True)
##     A = M(0)
##     for i in xrange(n-2):
##         A[i,i] = i
##         A[i,i+2] = i+2
##     print A.rank()
##     assert A.rank() == n-2
##     B = A.transpose()
##     print B.rank()
##     assert A.rank() == B.rank()
##     assert B.rank() == n-2
    
## def test_dense_echelon_consistent(n=10):
##     M = sage.linalg.matrix_space.MatrixSpace(rings.RationalField(), n)
##     A = M.random_element(prob=0.1)
##     print "A = ", A
##     print "A.rank() =", A.rank()
##     assert A.rank() == A.transpose().rank()
    
if __name__ ==  '__main__':
    import doctest, sys
    from sage.linalg.matrix_space import MatrixSpace
    from sage.linalg.vector_space import VectorSpace
    from sage.rings.rings import RationalField
    QQ = RationalField()
    doctest.testmod(sys.modules[__name__])
 
