"""
Space of matrices over a ring.
"""

# System imports
import random
import weakref

# SAGE imports
import sage.rings.rings 
import sage.structure.gens as gens
import matrix

_objs = {}
class _uniq(object):
    def __new__(cls, base_ring, nrows, ncols=None, sparse=False):
        if ncols == None: ncols = nrows
        key = (base_ring, nrows, ncols, sparse)
        if _objs.has_key(key):
            m = _objs[key]()
            if m != None: return m
        O = object.__new__(cls)
        _objs[key] = weakref.ref(O)
        return O
    
class MatrixSpace(_uniq, gens.Gens):
    """
    The space of all nrows x ncols matrices over base_ring.
    """
    def __init__(self,  base_ring,             
                        nrows, 
                        ncols=None, 
                        sparse=False):
        """
        Create with the command

              MatrixSpace(base_ring , nrows [, ncols] [, sparse])

        The default value of the optional argument sparse is False.
        The default value of the optional argument ncols is nrows.

        INPUT:
             base_ring -- a ring
             nrows -- int, the number of rows
             ncols -- (default nrows) int, the number of columns 
             sparse -- (default false) whether or not matrices are given
                       a sparse representation
        OUTPUT:
            The space of all nrows x ncols matrices over base_ring.

        EXAMPLES:
            sage: from sage.rings.rings import RationalField
            sage: from sage.linalg.matrix_space import *
            sage: QQ = RationalField()
            sage: MS = MatrixSpace(QQ,2,2)
            sage: MS.base_ring()
            Rational Field
            sage: MS.dimension()
            4
            sage: MS.dims()
            (2, 2)
            sage: B = MS.basis()
            sage: B
            [[1 0]
            [0 0], [0 1]
            [0 0], [0 0]
            [1 0], [0 0]
            [0 1]]
            sage: B[0]
            [1 0]
            [0 0]
            sage: B[1]
            [0 1]
            [0 0]
            sage: B[2]
            [0 0]
            [1 0]
            sage: B[3]
            [0 0]
            [0 1]
            sage: A = MS.matrix([1,2,3,4])
            sage: A
            [1 2]
            [3 4]
            sage: MS2 = MatrixSpace(QQ,2,3)
            sage: B = MS2.matrix([1,2,3,4,5,6])
            sage: A*B
            [ 9 12 15]
            [19 26 33]
        """
        if not isinstance(base_ring, sage.rings.rings.Ring):
            raise TypeError, "base_ring must be a sage.rings.rings.Ring"
        if ncols == None: ncols = nrows
        nrows = int(nrows)
        #if not isinstance(nrows, int):
        #    raise TypeError, "nrows must be an int"
        ncols = int(ncols)
        #if not isinstance(ncols, int):
        #    raise TypeError, "ncols must be an int"    
        if nrows < 0:
            raise ArithmeticError, "nrows must be nonnegative"
        if ncols < 0:
            raise ArithmeticError, "ncols must be nonnegative"
        self.__base_ring = base_ring
        self.__nrows = nrows
        self.__is_sparse = sparse
        if ncols == None:
            self.__ncols = nrows
        else:
            self.__ncols = ncols
        self.__set_matrix_class()
                
    def __call__(self, entries=0, coerce_entries=True, copy=True):
        return self.matrix(entries, coerce_entries, copy)

    def __cmp__(self, other):
        if isinstance(other, MatrixSpace) and \
           self.__base_ring == other.__base_ring and \
           self.__nrows == other.__nrows and \
           self.__ncols == other.__ncols and \
           self.__is_sparse == other.__is_sparse:
            return 0
        return -1
            
        
    def __contains__(self, x):
        return isinstance(x, matrix.Matrix) and x.parent() == self
 
    def __repr__(self):
        if self.is_sparse():
            s = "sparse"
        else:
            s = "dense"
        return "Full MatrixSpace of %s by %s %s matrices over %s"%(
                    self.__nrows, self.__ncols, s, self.__base_ring)
          
    def __set_matrix_class(self):
        if self.is_dense() and isinstance(self.base_ring(), sage.rings.rings.RationalField):
            self.__matrix_class = matrix.Matrix_rational_dense
            return
        if self.is_sparse():
            if isinstance(self.base_ring(), sage.rings.rings.RationalField):
                self.__matrix_class = matrix.Matrix_rational_sparse
                return
            self.__matrix_class = matrix.Matrix_generic_sparse
            return
        self.__matrix_class = matrix.Matrix_generic_dense
            
    def base_ring(self):
        return self.__base_ring
        
    def basis(self):
        try:
            return self.__basis
        except AttributeError:
            self.__basis = [self() for _ in range(self.dimension())]
            one = self.base_ring()(1)
            i = 0
            for r in range(self.__nrows):
                for c in range(self.__ncols):
                    self.__basis[i][r,c] = one
                    i += 1
            return self.__basis
    
    def dimension(self):
        return self.__nrows * self.__ncols
    
    def dims(self):
        return (self.__nrows, self.__ncols)
        
    def identity_matrix(self):
        if self.__nrows != self.__ncols:
            raise TypeError, "self must be a space of square matrices"
        A = self(0)
        for i in xrange(self.__nrows):
            A[i,i] = 1
        return A
        
    def is_dense(self):
        return not self.__is_sparse
        
    def is_sparse(self):
        return self.__is_sparse

    def gen(self, n):
        return self.basis()[n]

    def ngens(self):
        return self.dimension()
 
    def matrix(self, x=0, coerce_entries=True, copy=True):
        if isinstance(x, xrange):
            x = list(x)
        elif isinstance(x, (int, sage.rings.rings.Integer)) and x==1:
            return self.identity_matrix()
        if isinstance(x, matrix.Matrix):
            if x.parent() == self:
                return x.copy()
            x = x.list()
        return self.__matrix_class(self, x, coerce_entries, copy)
     
    def matrix_space(self, nrows, ncols, sparse=False):
        return MatrixSpace(self.__base_ring, nrows, ncols, 
                        sparse=sparse)
 
    def ncols(self):
        return self.__ncols
        
    def nrows(self):
        return self.__nrows
    
    def random_element(self, X=[-2,-1,1,2], prob=1.0, coerce=True):
        """
        Returns a random element of self.
        """
        prob=float(prob)
        if not isinstance(X, list):
            raise TypeError, "X must be a list"
        if not isinstance(coerce, bool):
            raise TypeError, "coerce must be a bool"
        R = self.base_ring()
        if coerce:
            X = [R(a) for a in X]
        A = self(0)
        for i in range(self.nrows()):
            for j in range(self.ncols()):
                if random.random() <= prob:
                    A[i,j] = R(random.choice(X))
        return A

    random = random_element
        
if __name__ ==  '__main__':
    import doctest, sys
    doctest.testmod(sys.modules[__name__])
