"""
Functional notation

These are function so that you can write foo(x) instead of x.foo() in
certain common cases.
"""

#*****************************************************************************
#       Copyright (C) 2004 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************
__doc_exclude = ["RealField", "ComplexField", "R", "C"]

from sage.rings.rings import RealField, ComplexField, PolynomialRing, RationalField
import sage.rings.integer_ring
R = RealField()
C = ComplexField()

import math

##############################################################################
# There are many functions on elements of a ring, which mathematicians
# usually write f(x), e.g., it is weird to write x.log() and natural
# to write log(x).  The functions below allow for the more familiar syntax.
##############################################################################
def arg(x):
    """
    Return the argument of a complex number x.
    """
    try: return x.arg()
    except AttributeError: return C(x).arg()

def base_ring(x):
    """
    Return the base ring over which x is defined.
    """
    return x.base_ring()

def base_field(x):
    """
    Return the base field over which x is defined.
    """
    return x.base_field()

def basis(x):
    """
    Return the fixed basis of x.
    """
    return x.basis()

def charpoly(x):
    """
    Return the characteristic polynomia of x.
    """
    return x.charpoly()

def conductor(x):
    """
    Return the conductor of x.
    """
    return x.conductor()

def cos(x):
    """
    Return the cosine of x.
    """
    try: return x.cos()
    except AttributeError: return math.cos(float(x))

def cyclotomic_polynomial(n):
    return PolynomialRing(RationalField()).cyclotomic_polynomial(n)

def decomposition(x):
    """
    Return the decomposition of x.
    """
    return x.decomposition()

def det(x):
    """
    Return the determinant of x.
    """
    return x.det()

def discriminant(x):
    return x.discriminant()

disc = discriminant

def exp(x):
    """
    Return the value of the exponentation function at x.
    """
    try: return x.exp()
    except AttributeError: return math.exp(x)

def factor(x):
    """
    Return the prime factorization of x.
    """
    try: return x.factor()
    except AttributeError: return sage.rings.integer_ring.factor(x)

def gens(x):
    """
    Return the generators of x.
    """
    return x.gens()

def hecke_operator(x,n):
    """
    Return the n-th Hecke operator T_n acting on x.
    """
    return x.hecke_operator(n)

def image(x):
    """
    Return the image of x.
    """
    return x.image()

def is_even(x):
    """
    Return whether or not x is even, e.g., divisible by 2.
    """
    try: return x.is_even()
    except AttributeError: return x%2==0

def is_odd(x):
    """
    Return whether or not x is odd.  This is by definition the
    complement of is_even.
    """
    return not is_even(x)

def j_invariant(x):
    """
    Return the j_invariant of x.
    """
    return x.j_invariant()

def kernel(x):
    """
    Return the kernel of x.
    """
    return x.kernel()

def log(x,b=None):
    """
    Return the log of x to the base b.  The default base is e.

    INPUT:
        x -- number
        b -- base (default: None, which means natural log)
    OUTPUT:
        number

    WARNING: In MAGMA, the order of arguments is reversed from in
    SAGE, i.e., the base is given first.  We use the opposite
    ordering, so the base can be viewed as an optional second
    argument.
    """
    if b is None:
        try: return x.log()
        except AttributeError:
            return math.log(x)
    else:
        try: return x.log(b)
        except AttributeError:
            return math.log(x)/math.log(b)
    

def ngens(x):
    """
    Return the number of generators of x.
    """
    return x.ngens()

def norm(x):
    """
    Return the norm of x.
    """
    return x.norm()

def one(R):
    """
    Return the one element of the ring R.
    """
    return R(1)

def order(x):
    """
    Return the order of x.
    """
    return x.order()

def rank(x):
    """
    Return the rank of x.
    """
    return x.rank()

def real(x):
    """
    Return the real part of x.
    """
    try: return x.real()
    except AttributeError: return C(x).real()

def regulator(x):
    """
    Return the regulator of x.
    """
    return x.regulator()

def imag(x):
    """
    Return the imaginary part of x.
    """
    try: return x.imag()
    except AttributeError: return C(x).imag()

def sqrt(x):
    """
    Return a square root of x.
    """
    try: return x.sqrt()
    except AttributeError: return math.sqrt(x)

def isqrt(x):
    """
    Return an integer square root, i.e., the floor of a
    square root.
    """
    try: return x.isqrt()
    except AttributeError:
        raise NotImplementedError

def sin(x):
    """
    Return the sin of x.
    """
    try: return x.sin()
    except AttributeError: return math.sin(float(x))
    
def tan(x):
    """
    Return the tangent of x.
    """
    try: return x.tan()
    except AttributeError: return math.tan(float(x))

def zero(R):
    """
    Return the zero element of the ring R.
    """
    return R(0)



#################################################################
# Generic parent
#################################################################
def parent(x):
    """
    Return x.parent() if defined, or type(x) if not.

    EXAMPLE:
        sage: Z = parent(int(5))
        sage: Z(17)
        17
        sage: Z
        <type 'int'>
    """
    try:
        return x.parent()
    except AttributeError:
        return type(x)

#################################################################
# Latex printing of SAGE objects
#################################################################
def latex(x):
    """
    Output x formated for inclusion in a LaTeX document.
    """
    try:
        return x.latex()
    except AttributeError:

        if isinstance(x, list):
            return "[" + ", \n".join([latex(v) for v in x]) + "]"
        
        return str(x)

