"""
Cusps: the set P^1(Q)
"""

#*****************************************************************************
#       Copyright (C) 2005 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.rings.rings import Infinity, infinity
from sage.rings.integer_ring import IntegerRing
from sage.rings.rational_field import RationalField
import sage.rings.rational as rational
import sage.rings.integer as integer

ZZ = IntegerRing(); QQ = RationalField()

class Cusp:
    """
    A cusp.

    A cusp is either a rational number or infinity, i.e., an element
    of the projective line over Q.  A Cusp is stored as a pair (a,b),
    where gcd(a,b)=1 and a,b are of type Integer.
    """

    def __init__(self, a, b=ZZ(1)):
        """
        Create the cusp a/b in P^1(Q), where if b=0 it is the point at
        infinity.

        EXAMPLES:
            sage: Cusp(2,3)
            2/3
            sage: Cusp(3,6)
            1/2
            sage: Cusp(1,0)
            Infinity
            sage: Cusp(infinity)
            Infinity
            sage: Cusp(5)
            5
            sage: Cusp(QQ("1/2"))             # rational number
            1/2
            sage: Cusp(1.5)
            Traceback (most recent call last):
            ...
            TypeError: Unable to coerce 1.5,1 to a Cusp
        """
        if isinstance(a, Infinity):
            self.__a = ZZ(1)
            self.__b = ZZ(0)
            return

        if isinstance(a, rational.Rational):
            self.__a = a.numer()
            self.__b = a.denom()
            return
        
        if isinstance(a, (int,long, integer.Integer)) and isinstance(b, (int,long, integer.Integer)):
            a = ZZ(a)
            b = ZZ(b)

        if isinstance(a, integer.Integer) and isinstance(b, integer.Integer):
            if b < 0:
                b *= -1
                a *= -1
            g = a.gcd(b)
            self.__a = a//g
            self.__b = b//g
            return
        
        else:
            
            raise TypeError, "Unable to coerce %s,%s to a Cusp"%(a,b)

    def numerator(self):
        """
        Return the numerator of the cusp a/b.

        sage: x=Cusp(6,9); x
        2/3
        sage: x.numerator()
        2
        """
        return self.__a

    def denominator(self):
        """
        Return the denominator of the cusp a/b.

        sage: x=Cusp(6,9); x
        2/3
        sage: x.denominator()
        3
        """        
        return self.__b

    def __repr__(self):
        if self.__b.is_zero():
            return "Infinity"
        if self.__b != 1:
            return "%s/%s"%(self.__a,self.__b)
        else:
            return str(self.__a)

    def __neg__(self):
        return Cusp(-self.__a, self.__b)

    def is_gamma0_equiv(self, other, N, transformation = False):
        """
        Return whether self and other are equivalent modulo the
        action of Gamma_0(N) via linear fractional transformations.
        
        INPUT:
            other -- Cusp
            N -- an integer (specifies the group Gamma_0(N))
            
            transformation -- bool (default: False), if True, also
                              return upper left entry of a matrix in
                              Gamma_0(N) that sends self to other.

        OUTPUT:
            bool -- True if self and other are equivalent
            Integer -- returned only if transformation is True

        EXAMPLES:
            sage: x = Cusp(2,3)
            sage: y = Cusp(4,5)
            sage: x.is_gamma0_equiv(y, 2)
            True
            sage: x.is_gamma0_equiv(y, 2, True)
            (True, 1)
            sage: x.is_gamma0_equiv(y, 3)
            False
            sage: x.is_gamma0_equiv(y, 3, True)
            (False, None)
            sage: Cusp(1,0)
            Infinity
            sage: z = Cusp(1,0)
            sage: x.is_gamma0_equiv(z, 3, True)
            (True, 2)

        ALGORITHM:
            See Proposition 2.2.3 of Cremona's book "Algorithms for Modular
            Elliptic Curves".
        """
        N = ZZ(N)
        u1 = self.__a
        v1 = self.__b
        u2 = other.__a
        v2 = other.__b
        if (u1,v1) != (ZZ(0),ZZ(1)):
            if v1 in [ZZ(0),ZZ(1)]:
                s1 = ZZ(1)
            else:
                s1 = u1.inverse_mod(v1)
        else:
            s1 = 0
        if (u2,v2) != (ZZ(0),ZZ(1)):
            if v2 in [ZZ(0),ZZ(1)]:
                s2 = 1
            else:
                s2 = u2.inverse_mod(v2)
        else:
            s2 = 0
        g = (v1*v2).gcd(N)
        a = s1*v2 - s2*v1
        if a%g != ZZ(0):
            if transformation:
                return False, None
            else:
                return False

        if not transformation:
            return True

        # Now we know the cusps are equivalent.  Use the proof of Prop 2.2.3
        # of Cremona to find a matrix in Gamma_0(N) relating them.
        dum, s2, r2 = u2.xgcd(-v2)
        assert dum.is_one()
        dum, s1, r1 = u1.xgcd(-v1)
        assert dum.is_one()
        a = s1*v2 - s2*v1
        assert (a%g).is_zero()
        # solve x*v1*v2 + a = 0 (mod N).
        d,x0,y0 = (v1*v2).xgcd(N)          # x0*v1*v2 + y0*N = d = g.
        # so x0*v1*v2 - g = 0 (mod N)
        x = -x0 * (a/g)
        # now  x*v1*v2 + a = 0 (mod N)
        s1p = s1+x*v1
        return (True, (u2*s1p-r2*v1)%N)

    def is_gamma1_equiv(self, other, N):
        """
        """
        raise NotImplementedError

    def is_chi_equiv(self, other, chi):
        """
        """
        raise NotImplementedError        

    
if __name__ ==  '__main__':
    import doctest, sys
    from sage.all import QQ
    doctest.testmod(sys.modules[__name__])
