"""
Hecke algebras and modules 
"""

#*****************************************************************************
#       Copyright (C) 2004 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************


import math
import sage.rings.arith as arith
import sage.modular.dims as dims
import sage.misc.misc as misc
import sage.structure.function as function

class UniversalHeckeAlgebra:

    def __init__(self, weight):
        self.__weight = weight

    def __repr__(self):
        return "Universal Hecke algebra of weight %s"%self.__weight

class UniversalHeckeOperator:

    def __init__(self, parent, x):
        self.__parent = parent
        self.__x = x

    def __repr__(self):
        return "Universal Hecke operator %s in %s"%\
               (self.__parent, self.__x)


class HeckeAlgebra:

    def __init__(self, hecke_module, anemic=False):
        if not isinstance(hecke_module, HeckeModule):
            raise TypeError, "hecke_module (=%s) must be a HeckeModule"%hecke_module
        self.__hecke_module = hecke_module
        self.__bound = None

    def __repr__(self):
        return "Hecke algebra acting on %s"%self.__hecke_module

    def hecke_module(self):
        return self.__hecke_module

    def __call__(self, n):
        n = int(n)
        return self.hecke_operator(n)

    def hecke_operator(self, n):
        n = int(n)
        return HeckeOperator(self, self.__hecke_module.hecke_matrix_function(n))

    def decomposition(self):
        return self.__hecke_module.decomposition()

    def generators(self):
        raise NotImplementedError

    def discriminant(self):
        raise NotImplementedError

    def bound(self):
        """
        An integer n so that the Hecke operators T_1,...,T_n generate
        the Hecke algebra as a Z-module.
        """
        if self.__bound != None:
            return self.__bound
        M = self.hecke_module()
        k = M.weight()
        N = M.level()
        self.__bound = int(math.ceil(k*dims.idxG0(N)/12.0))
        return self.__bound

    
class HeckeOperator:

    def __init__(self, parent, x, index=None):
        if not isinstance(parent, HeckeAlgebra):
            raise TypeError, "parent (=%s) must be a HeckeAlgebra"%parent
        if not isinstance(x, function.MatrixFunction):
            raise TypeError, "x (=%s) must be a MatrixFunction"%x
        if index != None and not isinstance(index, (int,long)):
            raise TypeError, "index (=%s) must be an int or long"%index
        self.__parent = parent
        self.__x = x
        self.__index = index
        self.is_matrix = True
        
    def __repr__(self):
        if self.__index != None:
            return "Hecke operator T_%s on %s"%(self.__index, self.domain())
        return "Hecke operator on %s"%self.domain()

    def __add__(self, other):
        if not (isinstance(other, HeckeOperator) and other.__parent == self.__parent):
            raise ArithmeticError, "Hecke operators must be on the same space."
        return HeckeOperator(self.__parent, self.matrix_function() + other.matrix_function())

    def __sub__(self, other):
        if not (isinstance(other, HeckeOperator) and other.__parent == self.__parent):
            raise ArithmeticError, "Hecke operators must be on the same space."
        return HeckeOperator(self.__parent, self.matrix_function() - other.matrix_function())

    def __mul__(self, other):
        if not (isinstance(other, HeckeOperator) and other.__parent == self.__parent):
            try:
                other = self.domain().base_ring()(other)
                return HeckeOperator(self.__parent, self.matrix_function() * other)
            except:
                pass
            raise ArithmeticError, "Hecke operators must be on the same space."
        ind = None
        if arith.gcd(self.__index, other.__index) == 1:
            ind = self.__index * other.__index
        else:
            P = set(arith.prime_divisors(self.domain().level()))
            if P.issubset(set(arith.prime_divisors(self.__index))) and \
               P.issubset(set(arith.prime_divisors(other.__index))):
                ind = self.__index * other.__index
        return HeckeOperator(self.__parent, self.matrix_function() * other.matrix_function(), index=ind)

    def codomain(self):
        """
        Return the codomain of this Hecke operator, e.g., a space of
        modular symbols or modular forms.
        """
        return self.matrix_function().domain()

    def charpoly(self):
        """
        Return the characteristic polynomial of this Hecke operator.
        """
        return self.matrix().charpoly()

    def decompose(self):
        """
        Decompose the Hecke module under the action of this Hecke
        operator.
        """
        if arith.GCD(self.__index, self.domain().level()) == 1:
            return self.matrix_function().decompose(is_diagonalizable=True)
        # TODO: There are other weaker hypotheses that imply diagonalizability.
        return self.matrix_function().decompose()
    
    def det(self):
        """
        Return the determinant of this Hecke operator.
        """
        return self.matrix().det()

    def domain(self):
        """
        Return the domain of this Hecke operator, e.g., a space of
        modular symbols or modular forms.
        """
        return self.matrix_function().domain()

    def fcp(self):
        """
        Return the factorization of the characteristic polynomial of
        this Hecke operator.
        """
        return self.matrix().fcp()
    
    def index(self):
        """
        Return the index of this Hecke operator or None if this Hecke
        operator does not have an index.

        For each integer n>0, the Hecke operator T_n has index n.
        However, T_n + T_m is also a Hecke operator, but has no
        well-defined index.
        """
        return self.__index
    
    def matrix(self):
        """
        Return the matrix underlying this Hecke operator.
        """
        return self.__x.matrix()

    def matrix_function(self):
        return self.__x

    def parent(self):
        """
        Return the parent of this Hecke operator.  The parent is a
        Hecke algebra.
        """
        return self.parent()

    def trace(self):
        """
        Return the trace of this Hecke operator.
        """
        return self.matrix().trace()



        
class HeckeModule:
    
    def __init__(self):
        # true is anemic decomposition, and false is non-anemic one
        self.__decomposition = {True:None, False:None}  
        
    def dimension(self):
        raise NotImplementedError

    def set_factor_number(self, i):
        self.__factor_number = i

    def factor_number(self):
        try:
            return self.__factor_number
        except:
            return -1
    
    def level(self):
        raise NotImplementedError
    
    def weight(self):
        raise NotImplementedError
    
    def base_ring(self):
        raise NotImplementedError
    
    def hecke_matrix(self, n):
        raise NotImplementedError

    def hecke_matrix_function(self, n):
        n = int(n)
        return function.MatrixFunction(self.hecke_matrix(n), self, self, "Hecke operator T_%s"%n)
    
    def vector_space(self):
        raise NotImplementedError

    def decomposition(self, anemic=True, bound=None):
        """
        Returns the complete direct-sum decomposition of the Hecke
        module self under the action of Hecke operators.  This is the
        finest direct-sum decomposition of self that we can obtain
        using splittings defined by Hecke operators.

        Each factor in the decomposition is a Hecke submodule obtained
        as the kernel of f(T_p)^r acting on self, where p is coprime
        to the level and r=1, or f(X)^r exactly divides the
        characteristic polynomial, so ker(f(T_p)^r) is a generalized
        rational space (direct sum of conjugate generalized
        eigenspaces).

        INPUT:
            anemic -- bool (default: True),  if True, use only Hecke operators
                      of index coprime to the level.  This is desirable, since
                      
            bound -- int or None, (default: None).  If None, use all Hecke operators
                     up to the Sturm bound, and hence obtain the same result as
                     one would obtain by using every element of the Hecke ring.
                     If a fixed integer, decompose using only Hecke operators
                     T_p, with p prime, up to bound.

        OUTPUT:
            list -- a list of subspaces of self.
                      
        """        
        if not isinstance(anemic, bool):
            raise TypeError, "anemic must be of type bool."
        if self.__decomposition[anemic] != None:
            return self.__decomposition[anemic]
        time = misc.verbose("Decomposing %s"%self)
        T = self.hecke_algebra()
        if bound is None:
            bound = T.bound()
        else:
            bound = int(bound)
        D = []
        U = [self.vector_space()]
        p = 2
        while len(U) > 0 and p <= bound:
            misc.verbose(mesg="p=%s"%p,t=time)
            if anemic:
                while arith.GCD(p, self.level()) != 1:
                    p = arith.next_prime(p)
            misc.verbose("Decomposition using p=%s"%p)
            t = T(p).matrix()
            Uprime = []
            for V in U:
                for W, is_irred in t.decompose_subspace(V, is_diagonalizable=(self.level()%p != 0)):
                    if is_irred:
                        D.append(self(W))
                    else:
                        Uprime.append(W)
                    #end
                #end
            #end
            p = arith.next_prime(p)
            U = Uprime
        #end
        for W in U:
            D.append(self(W))
        for A in D:
            if anemic:
                A.__is_splittable_anemic = False
                A.__is_splittable = False
            else:
                A.__is_splittable = False
        self.__is_splittable = len(D) > 1
        if anemic:
            self.__is_splittable_anemic = len(D) > 1
            
        self.__decomposition[anemic] = D
        self.__decomposition[anemic].sort()
        for i in range(len(D)):
            self.__decomposition[anemic][i].set_factor_number(i)
        return self.__decomposition[anemic]
        
    def is_splittable(self):
        """
        Returns true if and only if only it is possible to split
        off a nontrivial generalized eigenspace of self as the
        kernel of some Hecke operator.
        """
        if not hasattr(self, "__is_splittable"):
            self.decomposition(anemic=False)
        return self.__is_splittable
        
    def is_splittable_anemic(self):
        """
        Returns true if and only if only it is possible to split
        off a nontrivial generalized eigenspace of self as the
        kernel of some Hecke operator of index coprime to the level.
        """
        if not hasattr(self,"__is_splittable_anemic"):
            self.decomposition(anemic=True)
        return self.__is_splittable_anemic

