"""
Space of boundary modular symbols.

Used mainly for computing the cuspidal subspace of modular symbols. 
"""

#*****************************************************************************
#       SAGE: System for Algebra and Geometry Experimentation
#
#       Copyright (C) 2005 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

# Python imports
import operator
import weakref

# SAGE imports
import sage.linalg.vector_space as vector_space
from   sage.misc.misc import repr_lincomb

import sage.modular.congroup as congroup
import sage.modular.dirichlet as dirichlet
import sage.rings.rings as rings
import sage.rings.coerce as coerce
import sage.structure.gens as gens

import sage.modular.cusps as cusps

import sage.modular.modsym.ambient 
import sage.modular.modsym.element as element
import sage.modular.modsym.manin_symbols as manin_symbols

QQ = rings.RationalField()


class BoundarySymbol:
    """
    A boundary symbol. 
    """
    def __init__(self, parent, x):
        """
        Create a boundary symbol.

        INPUT:
            parent -- BoundarySymbols; a space of boundary modular symbols
            x -- a dict with integer keys and values in the base
                 field of parent.
        """
        self.__parent = parent
        self.__x = x
        self.is_vector = True  # so arithmetic works right. 

    def __repr__(self):
        """
        Returns a string representation for self for printing purposes.
        """
        g = self.__parent._known_gens_repr()
        z = [0 for _ in xrange(len(g))]
        for i,c in self.__x.items():
            z[i] = c
        return repr_lincomb(g, z)

    def __add__(self, other):
        if not isinstance(other, BoundarySymbol):
            return coerce.bin_op(self, other, operator.add)
        z = dict(other.__x)
        for i, c in self.__x.items():
            if z.has_key(i):
                z[i] += c
            else:
                z[i] = c
        return BoundarySymbol(self.__parent, z)

    def __radd__(self, left):
        if not isinstance(left, BoundarySymbol):
            return coerce.bin_op(left, self, operator.add)
        assert False

    def __sub__(self, other):
        if not isinstance(other, BoundarySymbol):
            return coerce.bin_op(self, other, "-")
        z = dict(self.__x)
        for i, c in other.__x.items():
            if z.has_key(i):
                z[i] -= c
            else:
                z[i] = -c
        return BoundarySymbol(self.__parent, z)

    def __rsub__(self, left):
        if not isinstance(left, BoundarySymbol):
            return coerce.bin_op(left, self, "-")
        assert False
        
    def __mul__(self, other):
        x = {}
        for i, c in self.__x.items():
            x[i] = c*other
        return BoundarySymbol(self.__parent, x)

    def __rmul__(self, left):
        return self*left

    def __neg__(self):
        return self*(-1)

    def parent(self):
        return self.__parent

    def vector(self):
        return self.__parent.vector_space()(self.__x)


_objsBoundarySymbols = {}
class _uniqBoundarySymbols(object):
    def __new__(cls, group=1, weight=2, sign=0,
                base_field = QQ,
                character = None):
        key = (group, weight, sign, base_field, character)
        if _objsBoundarySymbols.has_key(key):
            x = _objsBoundarySymbols[key]()
            if x != None:
                return x
        R = object.__new__(cls)
        _objsBoundarySymbols[key] = weakref.ref(R)
        return R

class BoundarySymbols(_uniqBoundarySymbols, gens.Gens):
    """
    Space of boundary symbols for a congruence subgroup of SL_2(Z).
    
    This class is an abstract base class, so only derived classes should 
    be instantiated. 

    INPUT:
        weight -- int, the weight
        group -- congroup.CongruenceGroup, a congruence subgroup. 
        sign -- int, either -1, 0, or 1
        base_field -- rings.Ring (defaults to the rational numbers)
    """
    def __init__(self,
                 group = congroup.Gamma0(1),
                 weight = 2,
                 sign = 0,
                 base_field = QQ,
                 character = None):
        """
        Initialize a space of boundary symbols.
        """
        weight = int(weight)
        if not isinstance(group, congroup.CongruenceSubgroup):
            raise TypeError, "group must be a congruence subgroup"
        sign = int(sign)
        if not isinstance(base_field, rings.Ring) and base_field.is_field():
            raise TypeError, "base_field must be of type Ring and be a field"
        if character == None and isinstance(group, congroup.Gamma0):
            character = dirichlet.TrivialCharacter(group.level(), base_field)
        (self.__group, self.__weight, self.__character,
          self.__sign, self.__base_field) = (group, weight,
                                             character, sign, base_field)

    def group(self):
        """
        Return the congruence subgroup associated to this space of
        boundary modular symbols.
        """
        return self.__group

    def level(self):
        """
        Return the level of the congruence subgroup of this space.
        """
        return self.__group.level()
    
    def weight(self):
        """
        Return the weight of this space of boundary modular symbols.
        """
        return self.__weight

    def character(self):
        """
        Return the Dirichlet character assocaited to this space of
        boundary modular symbols.
        """
        return self.__character
    
    def sign(self):
        """
        Return the sign of the complex conjugation involution on this
        space of boundary modular symbols.
        """
        return self.__sign

    def base_field(self):
        """
        Return the base field of this space of boundary modular
        symbols.
        """
        return self.__base_field

    def gens(self, i):
        """
        Return the i-th generator of this space.
        """
        return BoundarySymbol(self, {i:1})

    def _known_gens(self):
        """
        Return the known generators of this space.
        """
        raise NotImplementedError

    def __len__(self):
        return len(self._known_gens())

    def vector_space(self):
        return vector_space.VectorSpace(self.__base_field, len(self._known_gens()), sparse=True)

    def dimension(self):
        return len(self._known_gens())

    #####################################################################
    # Coercion
    #####################################################################
    def __call__(self, x):
        """
        Coerce x into a boundary symbol space.

        If x is a modular symbol (with the same group, weight,
        character, sign, and base field), this returns the image of
        that modular symbol under the boundary map.
        """
        raise NotImplementedError

    def __repr__(self):
        return ("Space of Boundary Modular Symbols of weight %s for" + \
                " %s with sign %s and character %s over %s")%(
                 self.weight(), self.group(), self.sign(),
                 self.character(), self.base_field())

    def element(self, x):
        """
        Creates and returns an element of self from a modular or
        sage symbol, if possible.
        INPUT:
           x -- an object of one of the following types:
                ModularSymbol, ManinSymbol.  We ignore the group,
                weight, and character of the input modular or sage symbol.
        OUTPUT:
           ModularSymbol -- a modular symbol with parent self.
        """
        raise NotImplementedError

    def filename(self):
        """
        Returns the filename of self that should be used to store self
        in the database.
        INPUT:
           self -- space of modular symbols
        OUTPUT:
           str -- a string
        """
        return "boundsym-%s_%s_%s_%s_%s"%\
               (self.__group.name(),self.__weight,self.__character,\
                self.__sign,self.__base_field.name())

class BoundarySymbols_wtk_g0_Q(BoundarySymbols):
    """
    Boundary symbols for Gamma_0(N) of integer weight k > 2 over the field Q.
    """
    def __init__(self, N, k, sign=0):
        """
        Initialize a space of boundary symbols of weight k for
        Gamma_0(N), over Q.
        
        For weight 2, it is faster to use BoundarySymbols_wt2_g0_Q.
        
        INPUT:
            N -- int, the level
            k -- integer weight >= 2.
            sign -- int, either -1, 0, or 1
        """
        N = int(N)
        k = int(k)
        sign = int(sign)
        if not sign in [-1,0,1]:
            raise TypeError, "sign must be an int in [-1,0,1]"
            
        BoundarySymbols.__init__(self,
                                 weight = k,
                                 group  = congroup.Gamma0(N),
                                 sign   = sign,
                                 base_field = QQ)
        
    def __repr__(self):
        return ("Space of Boundary Modular Symbols for %s of weight %s " + \
                "and over %s")%(self.group(),
                                             self.weight(),
                                             self.base_field())
                    

    
class BoundarySymbols_wt2_g0_Q(BoundarySymbols_wtk_g0_Q):
    """
    Modular symbols for Gamma_0(N) of integer weight 2 over the field Q.
    """
    def __init__(self, N, sign=0):
        """
        Initialize a space of boundary modular symbols.
        INPUT:
            N -- int, the level
            sign -- int, either -1, 0, or 1
        OUTPUT:
            The space of boundary modular symbols of weight 2, trivial
            character, level N and given sign.
        """            
        BoundarySymbols_wtk_g0_Q.__init__(self, 
                N=N, k=2, sign=sign)
        self.__known_gens = []
        self.__known_gens_repr = []
        self.__N = N
        self.__sign = sign

    def __call__(self, x):
        if isinstance(x, int) and x == 0:
            return BoundarySymbol(self, {})

        elif isinstance(x, cusps.Cusp):
            N = self.__N
            g = self.__known_gens
            for i in xrange(len(g)):
                if x.is_gamma0_equiv(g[i], N):
                    return BoundarySymbol(self, {i:1})
            g.append(x)
            self.__known_gens_repr.append("[%s]"%x)
            return BoundarySymbol(self, {(len(g)-1):1})

        elif isinstance(x, manin_symbols.ManinSymbol):
            alpha, beta = x.endpoints(self.level())
            return self(beta) - self(alpha)
        
        elif isinstance(x, element.ModularSymbolElement):
            M = x.parent()
            if not isinstance(M,
                              sage.modular.modsym.ambient.ModularSymbolsAmbient_wt2_g0_Q):
                raise TypeError, "x (=%s) must be an element of a space of modular symbols of type ModularSymbolsAmbient_wt2_g0_Q"%x
            if M.level() != self.level():
                raise TypeError, "x (=%s) must have level %s but has level %s"%(
                    x, self.level(), M.level())
            S = x.manin_symbol()
            if len(S) == 0:
                return self(0)
            return sum([c*self(v) for c, v in S])

        raise TypeError, "Coercion of %s into %s not (yet) defined."%(x,self)

    def _known_gens(self):
        return self.__known_gens
    
    def _known_gens_repr(self):
        return self.__known_gens_repr
            
class BoundarySymbols_wtk_g1_Q(BoundarySymbols):
    def __init__(self, level, weight, sign=0):
        """
        Initialize a space of boundayr modular symbols for Gamma1(N).
        INPUT:
            level -- int, the level
            weight -- int, the weight >= 2
            sign -- int, either -1, 0, or 1
        EXAMPLES:
            sage: BoundarySymbols_wtk_g1_Q(17,2)
            Boundary Modular Symbols space for Gamma_1(17) of weight 2 and dimension 25 over Rational Field
            sage: [BoundarySymbols_wtk_g1_Q(7,k).dimension() for k in [2,3,4,5]]
            [5, 8, 12, 16]
        """            
        BoundarySymbols.__init__(self, 
                weight = weight, 
                group  = congroup.Gamma1(level),
                sign   = sign, 
                base_field = QQ)
                
    def __repr__(self):
        return ("Boundary Modular Symbols space for Gamma_1(%s) of weight %s " + \
                "and dimension %s over %s")%(self.level(),self.weight(),
                    self.dimension(), self.base_field())
 
        
class BoundarySymbols_wtk_eps(BoundarySymbols):
    def __init__(self, eps, weight, sign=0):
        """
        Space of boundary modular symbols with given weight, character, and sign.

        INPUT:
            eps -- dirichlet.DirichletCharacter, the "Nebentypus" character.
            weight -- int, the weight >= 2
            sign -- int, either -1, 0, or 1
        EXAMPLES:
        
        """            
        level = eps.modulus()
        BoundarySymbols.__init__(self, 
                weight = weight, 
                group = congroup.Gamma1(level),
                sign = sign, 
                base_field = eps.base_ring(),
                character = eps)
                
    def __repr__(self):
        return ("Boundary Modular Symbols space of level %s, weight %s, character %s " + \
                "and dimension %s over %s")%(self.level(), self.weight(), 
                    self.character(), self.dimension(), self.base_field())
 



