"""
Subspace of ambient spaces of modular symbols.
"""

#*****************************************************************************
#       SAGE: System for Algebra and Geometry Experimentation
#
#       Copyright (C) 2005 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

# Design thought: Make an attribute is_hecke_equivariant.  If True,
# then compute Hecke operators on space and don't have to check that
# they are well defined and space is preserved.  If false, must check
# this, or return operator from space to ambient space (or image under
# Hecke operator?)    

import weakref

import ambient
import space
import sage.modular.congroup as congroup

class ModularSymbolsSubspace(space.ModularSymbolsSpace):
    """
    Subspace of ambient space of modular symbols.
    """
    
    def __init__(self, ambient_space, vector_space):
        self.__ambient_space = ambient_space
        self.__vector_space = vector_space
        A = ambient_space
        space.ModularSymbolsSpace.__init__(self, A.group(), A.weight(), \
                                     A.character(), A.sign(), A.base_field())

    def __repr__(self):
        return "Subspace of Modular Symbols of dimension %s and level %s"%(
                    self.dimension(), self.level())

    def __add__(self, other):
        if not isinstance(other, space.ModularSymbolsSpace):
            raise TypeError, "other (=%s) must be a ModularSymbolsSpace."%other
        if self.ambient_space() != other.ambient_space():
            raise ArithmeticError, "Sum only defined for subspaces of a common ambient modular symbols space."
        if other.is_ambient():
            return other
        # Neither is ambient        
        V = self.embedded_vector_space() + other.embedded_vector_space()
        M = self.ambient_space()
        return M(V)

    def intersect(self, other):
        """
        Returns the intersection of self and other, which must both
        lie in a common ambient space of modular symbols.
        EXAMPLES:
            sage: M = ModularSymbols(43, sign=1)
            sage: A = M[0] + M[1]
            sage: B = M[1] + M[2]
            sage: A.dimension(), B.dimension()
            (2, 3)
            
            SAGE: C = A.intersect(B); C.dimension()  # TODO
            1
        """
        if self.ambient_space() != other.ambient_space():
            raise ArithmeticError, "Intersection only defined for subspaces of"\
                  + " a common ambient modular symbols space."
        if other.is_ambient():
            return self
        # Neither is ambient
        V = self.embedded_vector_space().intersect(other.embedded_vector_space())
        M = self.ambient_space()
        return M(V)


    def _compute_hecke_matrix(self, n):
        A = self.ambient_space().hecke_matrix(n)
        T = A.restrict(self.embedded_vector_space())
        self._hecke_matrices[n] = T
        return T

    def ambient_space(self):
        return self.__ambient_space

    def ambient_vector(self, v):
        """
        Given a vector v in self.vector_space(), return the
        corresponding vector in the ambient modular symbols space.
        """
        B = self.__vector_space.basis()
        zero = self.__vector_space.ambient_space()(0)
        return sum([v[i]*B[i] for i in range(len(B))], zero)

    def dimension(self):
        return self.__vector_space.dimension()

    def embedded_vector_space(self):
        return self.__vector_space
        
    def is_ambient(self):
        return False



if __name__ ==  '__main__':
    import doctest, sys
    from modsym import ModularSymbols
    from sage.modular.dirichlet import DirichletGroup
    doctest.testmod(sys.modules[__name__])
