"""
A finite field
"""

#*****************************************************************************
#
#   SAGE: System for Algebra and Geometry Experimentation    
#
#       Copyright (C) 2005 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************


import random, weakref

import arith
import ring
import sage.rings.rings
import sage.structure.gens as gens
import sage.tables.conway

import sage.libs.pari

import finite_field_element
import integer_mod_ring



_objsFiniteField = {}
class FiniteField(object, ring.Ring, gens.Gens):
    def __new__(cls, modulus, var="a"):
        key = (modulus,var)
        if _objsFiniteField.has_key(key):
            x = _objsFiniteField[key]()
            if x != None: return x
        if arith.is_prime(modulus):
            R = object.__new__(FiniteField_prime_modn)
        else:
            R = object.__new__(FiniteField_ext_pari)
        _objsFiniteField[key] = weakref.ref(R)
        return R

    def __init__(self):
        raise NotImplementedError

    def __contains__(self, x):
        if not isinstance(x, finite_field_element.FiniteFieldElement):
            return False
        return x.parent() == self

    def __cmp__(self, other):
        """
        Compares this finite field with other.  Two finite fields are
        equal if and only if they have the same cardinality *and* the
        defining polynomials are the same.

        EXAMPLES:
            sage: FiniteField(3**2) == FiniteField(3**3)
            False
            sage: FiniteField(3**2) == FiniteField(3**2)
            True
            sage: FiniteField(3**2,'beta') == FiniteField(3**2,'alpha')
            False
            sage: FiniteField(3**2,'beta') == FiniteField(3**2,'beta')
            True
        """
        if self is other: return 0
        if not isinstance(other, FiniteField):
            return -1
        if self.characteristic() < other.characteristic():
            return -1
        elif self.characteristic() > other.characteristic():
            return 1
        if self.variable() != other.variable():
            return -1
        if self.order() < other.order():
            return -1
        elif self.order()== other.order() and \
                 (self.degree() == 1 or self.polynomial() == other.polynomial()):
            return 0
        return 1

    def __getitem__(self, n):
        """

        Returns the n-th element of the field.  The elements of the
        field are

             [a, a^2, a^3, ... , 0]

        where a is the generator for the multiplicative group returned
        by unit_gen().

        EXAMPLES:
            sage: k = GF(8)
            sage: k[0]
            1
            sage: k[1]
            a
            sage: k[7]
            0
        """
        if n < 0 or n >= self.order():
            raise IndexError, "n (=%s) must be between 0 and the order %s of the field."%(\
                n, self.order())
        if n == self.order() - 1:
            return self(0)
        return self.gen()**n

    def __iter__(self):
        """

        Return iterator over elements of the finite field, ending with 0.

        EXAMPLES:
            sage: k = GF(9)
            sage: i = 0
            sage: for x in k: print x, k[i]; i+= 1
            1 1
            a a
            a + 1 a + 1
            2*a + 1 2*a + 1
            2 2
            2*a 2*a
            2*a + 2 2*a + 2
            a + 2 a + 2
            0 0
        """
        a = self.gen()
        b = self(1)
        yield b
        for i in xrange(self.order() - 2):
            b = b*a
            yield b
        yield self(0)

    def gen(self):
        raise NotImplementedError

    def ngens(self):
        """
        The number of generators of the finite field.  Always 1.

        EXAMPLES:
            sage: k = FiniteField(3**4)
            sage: k.ngens()
            1
        """
        return 1
        
    def is_field(self):
        """
        Returns whether or not the finite field is a field, i.e.,
        always returns True.

        EXAMPLES:
            sage: k = FiniteField(3**4)
            sage: k.is_field()
            True
        """
        return True

    def is_finite(self):
        return True

    def order(self):
        raise NotImplementedError
        
    def unit_group_exponent(self):
        """
        The exponent of the unit group of the finite field.  For a
        finite field, this is always the order minus 1.

        EXAMPLES:
            sage: k = GF(2**10)
            sage: k.order()
            1024
            sage: k.unit_group_exponent()
            1023
        """
        return self.order() - 1
    

    def random(self):
        """
        A random element of the finite field.

        EXAMPLES:
            >> k = GF(2**10)
            >> k.random()
            a^9 + a
        """
        return self[random.randrange(0,self.order())]

    def polynomial(self):
        raise NotImplementedError

    def prime_field(self):
        """
        Returns the prime subfield of this field.
        
        EXAMPLES:
            sage: k = GF(9)
            sage: k.prime_field()
            Finite field of size 3
        """
        return FiniteField(self.characteristic())

    prime_subfield = prime_field

    def variable(self):
        return 'x'

    def polynomial_ring(self):
        """
        Returns the polynomial ring over the prime subfield in the
        same variable as this finite field.
        
        EXAMPLES:
            sage: k = FiniteField(3**4, "alpha")
            sage: k.polynomial_ring()
            Univariate Polynomial Ring in alpha over Finite field of size 3
        """
        try:
            return self.__polynomial_ring
        except AttributeError:
            self.__polynomial_ring = sage.rings.rings.PolynomialRing(
                FiniteField(self.characteristic()), self.variable())
            return self.__polynomial_ring
        

class FiniteField_ext_pari(FiniteField):
    """
    Finite field of order q, where q is a nontrivial prime power.
    (Implemented using PARI mod's.)
    
    Create with the command
          FiniteField(modulus)

    INPUT:
        q -- int, prime power, order of the finite field
    OUTPUT:
        FiniteField -- finite field of order q.

    EXAMPLES:
        sage: k = FiniteField(9)
        sage: k
        Finite field of size 3^2
        sage: k.is_field()
        True
        sage: k.characteristic()
        3
        sage: a = k.gen()
        sage: a 
        a
        sage: a.parent()
        Finite field of size 3^2
        sage: a.charpoly()
        x^2 + 2*x + 2
        sage: [a**i for i in range(8)]
        [1, a, a + 1, 2*a + 1, 2, 2*a, 2*a + 2, a + 2]

    Fields can be coerced into sets or list and iterated over:
        sage: list(k)
        [1, a, a + 1, 2*a + 1, 2, 2*a, 2*a + 2, a + 2, 0]
        sage: set(k)
        set([a, 2*a + 1, a + 2, a + 1, 0, 1, 2*a + 2, 2, 2*a])
        sage: [x for x in k]
        [1, a, a + 1, 2*a + 1, 2, 2*a, 2*a + 2, a + 2, 0]
        
    Next we compute with the finite field of order 16, where
    the variable is named b.
        sage: k16 = FiniteField(16, "b")
        sage: z = k16.gen()
        sage: z
        b
        sage: z.charpoly()
        x^4 + x + 1
        sage: k16.is_field()
        True
        sage: k16.characteristic() 
        2
        sage: z.order()
        15

    Of course one can also make prime finite fields.
        sage: k = FiniteField(7)

    Note that the generator is a primitive root of unity.
        sage: k.gen()
        3
        sage: k.gen().order()
        6
    """

    def __init__(self, q, var="a"):
        """
        Create finite field of order q with variable var.

        INPUT:
            q -- integer, size of the finite field, not prime
            var -- (optional, defaults to 'a') variable used for
                   printing element of the finite field.  Also,
                   two finite fields are considered equivalent
                   if they have the same variable.
        OUTPUT:
            FiniteField -- a finite field of order q with variable var.
        EXAMPLES:
            sage: FiniteField(17)
            Finite field of size 17
            sage: FiniteField(2**10)
            Finite field of size 2^10
            sage: FiniteField(3**5, "b")
            Finite field of size 3^5
            sage: FiniteField(3**5, "b").gen()
            b

        You can also create a finite field using GF, which is a synonym 
        for FiniteField.
            sage: GF(19**2)
            Finite field of size 19^2
        """
        q = int(q)
        var = str(var)
        if q < 2:
            raise ArithmeticError, "q (=%s) must be a prime power"%q
        F = arith.factor(q, int_=True)
        if len(F) > 1:
            raise ArithmeticError, "q (=%s) must be a prime power"%q
        ring.Ring.__init__(self)
        self.__var = var
        self.__char = F[0][0]
        self.__pari_one = sage.libs.pari.pari(1).Mod(self.__char)
        self.__degree = F[0][1]
        if F[0][1] <= 1:
            raise ValueError, "The size of the finite field must not be prime."
        self.__order = q
        self.__is_field = True
        if exists_conway_polynomial(self.__char, self.__degree):
            f = conway_polynomial(self.__char, self.__degree)
        else:
            # The following is fast/deterministic, but has serious problems since
            # it crashes on 64-bit machines, and I can't figure out why.
            # So instead we iterate through random polys until we find an irreducible one.
            # self.__pari_modulus = sage.libs.pari.finitefield_init(self.__char, self.__degree, var)
            R = sage.rings.rings.PolynomialRing(GF(self.__char))
            while True:
                f = R.random(self.__degree)
                if f.is_irreducible():
                    break
        f = sage.libs.pari.pari(str(f))
        self.__pari_modulus = f.subst('x',var) * self.__pari_one
        self.__gen = finite_field_element.FiniteFieldElement(self, sage.libs.pari.pari(var))

    def _pari_one(self):
        """
        The PARI object Mod(1,p).  This is implementation specific
        and should be ignored by users.
        """
        return self.__pari_one

    def _pari_modulus(self):
        """
        The polynomial mod p that defines the finite field, as a PARI
        object.  This is implementation specific, and some finite fields
        might not be implemented using PARI, so you should avoid using
        this function.

        INPUT:  nothing
        OUTPUT:
            gen -- a pari polynomial gen
        EXAMPLES:
            sage: GF(19**2)._pari_modulus()
            Mod(1, 19)*a^2 + Mod(18, 19)*a + Mod(2, 19)

            sage: GF(13**3)._pari_modulus()
            Mod(1, 13)*a^3 + Mod(2, 13)*a + Mod(11, 13)
    
            sage: FiniteField(2**4, "b")._pari_modulus()
            Mod(1, 2)*b^4 + Mod(1, 2)*b + Mod(1, 2)
        """
        return self.__pari_modulus

    def gen(self, n=0):
        """
        Return chosen generator of the finite field.  This generator
        is a root of the defining polynomial of the finite field, and
        is guaranteed to be a generator for the multiplicative group.  

        INPUT:
            nothing
        
        OUTPUT:
            FiniteFieldElement -- field generator of finite field
            
        EXAMPLES:
            sage: FiniteField(2**4, "b").gen()
            b
            sage: k = FiniteField(3**4, "alpha")
            sage: a = k.gen()
            sage: a
            alpha
            sage: a**4
            alpha^3 + 1
        """
        return self.__gen

    def variable(self):
        return self.__var

    def characteristic(self):
        """
        Returns the characteristic of the finite field, which is a
        prime int.

        EXAMPLES:
            sage: k = FiniteField(3**4)
            sage: k.characteristic()
            3
        """
        return self.__char
    
    def degree(self):
        """
        Returns the degree of the finite field, which is a positive
        integer.

        EXAMPLES:
            sage: FiniteField(3).degree()
            1
            sage: FiniteField(3**20).degree()
            20
        """
        return self.__degree
    
    def __repr__(self):
        return "Finite field of size %s^%s"%(self.__char, self.__degree)
    
    def name(self):
        """
        Short description string that determines the finite field.
        EXAMPLES:
            sage: FiniteField(3**4).name()
            'GF3_4'
        """
        return "GF%s_%s"%(self.__char,self.__degree)
    

    def __call__(self, x):
        """
        Coerce x into the finite field.
        
        INPUT:
            x -- object

        OUTPUT:
            FiniteFieldElement -- if possible, makes a finite field element from x.
            
        EXAMPLES:
            sage: k = GF(3**4,'a')
            sage: b = k(5)
            sage: b.parent()
            Finite field of size 3^4
            sage: a = k.gen()
            sage: k(a + 2)
            a + 2
        
        WARNING: Finite fields are currently implemented somewhat
        naively using polynomials modulo p and the PARI ffinit
        function.  In particular, we do not use Conway polynomials and
        define natural consistent inclusion maps between different
        finite fields.
        """
        if isinstance(x, finite_field_element.FiniteFieldElement):
            if x.parent() is self:
                return x
            else:
                # This is where we *would* do coercion from one finite field to another...
                raise TypeError, "No coercion of %s into %s defined."%(x, self)
            
        if isinstance(x, (int, long, sage.rings.rings.Integer, sage.rings.rings.Rational, sage.libs.pari.gen)):
            return finite_field_element.FiniteFieldElement(self, x)
        raise TypeError, "No coercion of %s into %s defined."%(x, self)

    def __len__(self):
        """
        The number of elements of the finite field.

        EXAMPLES:
            sage: k = GF(2**10)
            sage: k
            Finite field of size 2^10
            sage: len(k)
            1024
        """
        return self.__order

    def order(self):
        """
        The number of elements of the finite field.

        EXAMPLES:
            sage: k = GF(2**10)
            sage: k
            Finite field of size 2^10
            sage: k.order()
            1024
        """
        return self.__order

    def polynomial(self):
        """
        Return the irreducible characteristic polynomial of the
        generator of this finite field, i.e., the polynomial f(x) so
        elements of the finite field as elements modulo f.

        EXAMPLES:
            sage: k = FiniteField(17)
            sage: k.polynomial()
            x
            sage: k = FiniteField(9)
            sage: k.polynomial()
            x^2 + 2*x + 2
        """
        try:
            return self.__polynomial
        except:
            R = sage.rings.rings.PolynomialRing(FiniteField(self.characteristic()))
            self.__polynomial = R(self._pari_modulus())
        return self.__polynomial


class FiniteField_prime_modn(FiniteField, integer_mod_ring.IntegerModRing):
    def __init__(self, p):
        p = int(p)
        if not arith.is_prime(p):
            raise ArithmeticError, "p (=%s) must be prime."%p
        integer_mod_ring.IntegerModRing.__init__(self, p)
        self.__char = p
        self.__gen = self(int(sage.libs.pari.pari(p).znprimroot().lift()))
        
    def characteristic(self):
        return self.__char

    def modulus(self):
        return self.__char

    def polynomial(self):
        try:
            return self.__polynomial
        except:
            self.__polynomial = sage.rings.rings.PolynomialRing(self)([0,1])
            return self.__polynomial

    def order(self):
        return self.__char

    def gen(self, n=0):
        return self.__gen

    def __repr__(self):
        return "Finite field of size %s"%self.order()

    def degree(self):
        """
        Returns the degree of the finite field, which is a positive
        integer.

        EXAMPLES:
            sage: FiniteField(3).degree()
            1
            sage: FiniteField(3**20).degree()
            20
        """
        return 1



##################################################################
    
GF = FiniteField

def conway_polynomial(p, n):
    """
    Return the Conway polynomial of degree n over GF(p), which is
    loaded from a table.

    If the requested polynomial is not known, this function raises a
    RuntimeError exception.

    INPUT:
        p -- int
        n -- int
        
    OUTPUT:
        Polynomial -- a polynomial over the prime finite field GF(p).

    NOTE: The first time this function is called a table is read from
    disk, which takes a fraction of a second.  Subsequent calls do not
    require reloading the table.

    See also the ConwayPolynomials() object, which is a table of
    Conway polynomials.   For example, if c=ConwayPolynomials, then
    c.primes() is a list of all primes for which the polynomials are
    known, and for a given prime p,  c.degree(p) is a list of all
    degrees for which the Conway polynomials are known.

    EXAMPLES:
        sage: conway_polynomial(2,5)
        x^5 + x^2 + 1
        sage: conway_polynomial(101,5)
        x^5 + 2*x + 99
        sage: conway_polynomial(97,101)
        Traceback (most recent call last):
        ...
        RuntimeError: Conway polynomial over F_97 of degree 101 not in database.
    """
    (p,n)=(int(p),int(n))
    R = sage.rings.rings.PolynomialRing(GF(p), 'x')
    try:
        return R(sage.tables.conway.ConwayPolynomials()[p][n])
    except KeyError:
        raise RuntimeError, \
              "Conway polynomial over F_%s of degree %s not in database."%(p,n)

def exists_conway_polynomial(p, n):
    """
    Return True if the Conway polynomial over F_p of degree n is in the
    database and False otherwise.

    If the Conway polynomial is in the database, to obtain it use the
    command conway_polynomial(p,n).

    EXAMPLES:
        sage: exists_conway_polynomial(2,3)
        True
        sage: exists_conway_polynomial(2,-1)
        False
        sage: exists_conway_polynomial(97,200)
        False
        sage: exists_conway_polynomial(6,6)
        False
    """
    return sage.tables.conway.ConwayPolynomials().has_polynomial(p,n)

    
if __name__ ==  '__main__':
    import doctest, sys
    doctest.testmod(sys.modules[__name__])
