"""
The Integer Ring

The class \\class{IntegerRing} represents
the ring $\\mathbf{Z}$ of (arbitrary precision) integers.  Each integer
is an instance of the class \\class{Integer}, which is defined
in a Pyrex extension module that wraps GMP integers 
(the \\class{mpz_t} type in GMP). 

    sage: Z = IntegerRing(); Z
    Integer Ring
    sage: Z.characteristic()
    0
    sage: Z.is_field()
    False

There is a unique instances of class \\class{IntegerRing}.  To create
an \\class{Integer}, coerce either a Python int, long, or a string.
Various other types will also coerce to the integers, when it makes
sense.

    sage: a = Z(1234); b = Z(5678); print a, b
    1234 5678
    sage: type(a)
    <type '_integer.Integer'>
    sage: a + b
    6912
    sage: Z('94803849083985934859834583945394')
    94803849083985934859834583945394
"""

#*****************************************************************************
#
#   SAGE: System for Algebra and Geometry Experimentation    
#
#       Copyright (C) 2005 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import random

import sage.rings.arith 
import sage.misc.misc as misc
import sage.rings.ring as ring
import sage.rings.integer
import sage.rings.integer_mod

_obj_int = {}
class _uniq_int(object):
    def __new__(cls):
        if _obj_int.has_key(0):
            return _obj_int[0]
        O = object.__new__(cls)
        _obj_int[0] = O
        return O

class IntegerRing(_uniq_int, ring.Ring):
    """
    The ring of integers.
    """    
    
    def __init__(self):
        """

        In order to introduce the ring $\\Z$ of integers, we illustrate
        creation, calling a few functions, and working with its
        elements.

            sage: Z = IntegerRing(); Z
            Integer Ring
            sage: Z.characteristic()
            0
            sage: Z.is_field()
            False

        We next illustrate basic arithmetic in $\\Z$:
        
            sage: a = Z(1234); b = Z(5678); print a, b
            1234 5678
            sage: type(a)
            <type '_integer.Integer'>
            sage: a + b
            6912
            sage: b + a
            6912
            sage: a * b
            7006652
            sage: b * a
            7006652
            sage: a - b
            -4444
            sage: b - a
            4444

        When we divide to integers using /, the result is
        automatically coerced to the field of rational numbers, even
        if the result is an integer.
        
            sage: a / b
            617/2839
            sage: type(a/b)
            <type '_rational.Rational'>
            sage: a/a
            1
            sage: type(a/a)
            <type '_rational.Rational'>

        For floor division, instead using the // operator:
            sage: a // b
            0
            sage: type(a//b)
            <type '_integer.Integer'>

        Next we illustrate arithmetic with automatic coercion.
        The types that coerce are: str, int, long, Integer.
            sage: a + 17
            1251
            sage: a * 374
            461516
            sage: 374 * a
            461516
            sage: a/19
            1234/19
            sage: 0 + Z(-64)
            -64

        Integers can be coerced:
            sage: a = Z(-64)
            sage: int(a)
            -64
            sage: long(a)
            -64L
        """
        pass

    def __contains__(self, x):
        return isinstance(x, sage.rings.integer.Integer)
    
    def __repr__(self):
        return "Integer Ring"
    
    def __call__(self, x):
        if isinstance(x, sage.rings.integer_mod.IntegerMod):
            x = x.lift()
        return sage.rings.integer.Integer(x)
        #z.set_str("%s"%hex(x).replace("0x","").replace("L",""),16)
        #return z

    def is_atomic_repr(self):
        """
        Return True, since elements of the integers do not have
        to be printed with paranethesis around them, when they
        are coefficients, e.g., in a polynomial.
        """
        return True
    
    def is_field(self):
        """
        Return False.
        """
        return False
    
    def name(self):
        """
        Return the string "Z".
        """
        return "Z"
    
    def characteristic(self):
        """
        Return 0 as a Python int.
        """
        return 0
    
    def random(self, bound=5):
        """
        Return a random integer between -bound and bound,
        including both endpoints.
        """
        return self(random.randint(-bound,bound))
    
    def __cmp__(self, other):
        if isinstance(other, IntegerRing):
            return 0
        return -1
    
    def zeta(self):
        """
        Return the integer -1, which is the root of unity of largest
        order in Z.
        """
        return sage.rings.integer.Integer(-1)


Z = IntegerRing()

def factor(n):
    """
    Return the factorization of the positive integer $n$ as a list of
    tuples $(p_i,e_i)$ such that $n=\prod p_i^{e_i}$.
    """
    return [(Z(p),e) for p, e in sage.rings.arith.factor(long(n))]
    
def crt_basis(X, xgcd=None):
    """
    Compute and return a Chinese Remainder Theorem basis for the list
    X of coprime integers.
    
    INPUT:
        X -- a list of Integers that are coprime in pairs
    OUTPUT:
        E -- a list of Integers such that E[i] = 1 (mod X[i])
             and E[i] = 0 (mod X[j]) for all j!=i.

    The E[i] have the property that if A is a list of objects, e.g.,
    integers, vectors, matrices, etc., where A[i] is moduli X[i], then
    a CRT lift of A is simply
                       sum E[i] * A[i].

    ALGORITHM:
    To compute E[i], compute integers s and t such that
    
                s * X[i] + t * (prod over i!=j of X[j]) = 1.   (*)

    Then E[i] = t * (prod over i!=j of X[j]).  Notice that equation
    (*) implies that E[i] is congruent to 1 modulo X[i] and to 0
    modulo the other X[j] for j!=i.

    COMPLEXITY: We compute len(X) extended GCD's.
    """
    if not isinstance(X, list):
        raise TypeError, "X must be a list"
    if len(X) == 0:
        return []

    P = misc.mul(X)

    Y = []
    # 2. Compute extended GCD's
    ONE=X[0].parent()(1)
    for i in range(len(X)):
        p = X[i]
        prod = P/p
        g,s,t = p.xgcd(prod)
        if g != ONE:
            raise ArithmeticError, "The elements of the list X must be coprime in pairs."
        Y.append(t*prod)
    return Y
             
    
if __name__ ==  '__main__':
    import doctest, sys
    doctest.testmod(sys.modules[__name__])
