"""
Multivariate Polynomials
"""

import weakref

from ring import Ring

import multi_polynomial
import polylist, polydict
import sage.structure.gens as gens


def MPolynomialRing(base_ring, n=1, var="x", repr='dict'):
    """
    Create a Multivariate polynomial ring.

    INPUT:
        base_ring -- Ring
        n -- int, number of variables  (default: 1)
        var -- list or str; list of n variable names or a string;
               if a string, names the variables var_0, var_1, etc.
               (default: 'x')
        repr -- str; choice of underlying representation of polynomials.
               (default: 'dict')
               The choices are
                   'dict' -- Use a Python dictionary-based representation  (faster)
                   'list' -- Use a Python list-based representation
                   
    EXAMPLES:
        sage: R = MPolynomialRing(RationalField(), 3)
        sage: R
        Polynomial ring in x0, x1, x2 over Rational Field
        sage: x0,x1,x2 = R.gens()
        sage: x0.element()
        PolyDict with representation {(1, 0, 0): 1}
        sage: x0 + x1 + x2
        x2 + x1 + x0
        sage: (x0 + x1 + x2)**2
        x2^2 + 2*x1*x2 + x1^2 + 2*x0*x2 + 2*x0*x1 + x0^2

    Next we specify all the variables and do some additional examples of arithmetic.
    
        sage: R = MPolynomialRing(RationalField(), 3, repr='list', var=['a','b','c'])
        sage: R
        Polynomial ring in a, b, c over Rational Field
        sage: a,b,c = R.gens()
        sage: a.element()
        PolyList with distributive representation [[1, [1, 0, 0]]]
        sage: (a+b)**2
        2*a*b + a^2 + b^2
        sage: (a + 2*b + 3*c**2)**3
        27*a*c^4 + 8*b^3 + 27*c^6 + 9*a^2*c^2 + 36*a*b*c^2 + a^3 + 54*b*c^4 + 6*a^2*b + 12*a*b^2 + 36*b^2*c^2


    We can construct multi-variate polynomials rings over completely
    arbitrary SAGE rings.  In this example, we construct a polynomial
    ring S in 3 variables over a polynomial ring in 2 variables over
    GF(9).  Then we construct a polynomial ring in 20 variables over S!
    
        sage: R = MPolynomialRing(GF(9),2, var=['n1','n2']); n1,n2=R.gens()
        sage: n1**2 + 2*n2
        2*n2 + n1^2
        sage: S = MPolynomialRing(R,3, var='a'); a0,a1,a2=S.gens()
        sage: S
        Polynomial ring in a0, a1, a2 over Polynomial ring in n1, n2 over Finite field of size 3^2
        sage: x = (n1+n2)*a0 + 2*a1**2
        sage: x
        2*a1^2 + (n2 + n1)*a0
        sage: x**3
        2*a1^6 + (n2^3 + n1^3)*a0^3
        sage: T = MPolynomialRing(S, 20)
        sage: T
        Polynomial ring in x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12, x13, x14, x15, x16, x17, x18, x19 over Polynomial ring in a0, a1, a2 over Polynomial ring in n1, n2 over Finite field of size 3^2
    """
    try:
        if not base_ring.is_ring():
            raise TypeError, "Base ring (=%s) must be a Ring."%base_ring
    except AttributeError:
        raise TypeError, "Base ring (=%s) must be a Ring."%base_ring
    if repr == 'list':
        return MPolynomialRing_polylist(base_ring, n, var)
    elif repr == 'dict':
        return MPolynomialRing_polydict(base_ring, n, var)


class MPolynomialRing_base(object, Ring, gens.Gens):
    def __init__(self, base_ring, n=1, var="x"):
        Ring.__init__(self)
        self.__base_ring = base_ring
        n = int(n)
        if n <= 0:
            raise ValueError, "Multivariate Polynomial rings must " + \
                  "have more than 0 variables."
        self.__ngens = n
        if not isinstance(var, list):
            self.__var_names = ["%s%s"%(var,i) for i in range(n)]
        else:
            self.__var_names = list(var)
        if len(self.__var_names) != n:
            raise RuntimeError, "Number of variables (=%s) must equal length of variable name list (=%s)."%(n, self.__var_names)
        self.__repr = "Polynomial ring in %s over %s"%(
            ", ".join(self.__var_names), base_ring)

    def __repr__(self):
        return self.__repr

    def base_ring(self):
        return self.__base_ring

    def characteristic(self):
        """
        Return the characteristic of this polynomial ring. 
        EXAMPLES:
            sage: R = PolynomialRing(RationalField(),3)
            sage: R.characteristic()
            0
            sage: R = PolynomialRing(GF(7),20)
            sage: R.characteristic()
            7
        """
        return self.__base_ring.characteristic()

    def gen(self, n=0):
        if n < 0 or n >= self.ngens():
            raise ValueError, "Generator %s not defined."%n
        return self._gens[n]

    def ngens(self):
        return self.__ngens

    def variable(self):
        return self.__variable

    def variable_names(self):
        return self.__var_names

    def _monomial_order_function(self):
        raise NotImplementedError


class MPolynomialRing_polylist(MPolynomialRing_base):

    def __init__(self, base_ring, n=1, var="x"):
        MPolynomialRing_base.__init__(self, base_ring, n, var)
        self.__monomial_order_function = polylist.lex_monord
        # Construct the generators
        v = [0 for _ in xrange(n)]
        one = base_ring(1)
        self._gens = []
        for i in xrange(n):
            v[i] = 1
            self._gens.append(multi_polynomial.MPolynomial_polylist(self, [[one, list(v)]]))
            v[i] = 0
        self.__zero_list = v
            
    def _monomial_order_function(self):
        return self.__monomial_order_function

    def __call__(self, x):
        """
        Coerce x into this multivariate polynomial ring.
        """
        if isinstance(x, multi_polynomial.MPolynomial) and x.parent() == self:
            return x
        if isinstance(x, polylist.PolyList):
            return multi_polynomial.MPolynomial_polylist(self, x)
        c = self.base_ring()(x)
        return multi_polynomial.MPolynomial_polylist(self, [[c, list(self.__zero_list)]])

                
class MPolynomialRing_polydict(MPolynomialRing_base):

    def __init__(self, base_ring, n=1, var="x"):
        MPolynomialRing_base.__init__(self, base_ring, n, var)
        # Construct the generators
        v = [0 for _ in xrange(n)]
        one = base_ring(1);
        self._gens = []
        for i in xrange(n):
            v[i] = 1  # int's!
            self._gens.append(multi_polynomial.MPolynomial_polydict(self, {tuple(v):one}))
            v[i] = 0
        self.__zero_tuple = tuple(v)
            
    def _monomial_order_function(self):
        return self.__monomial_order_function

    def __call__(self, x):
        """
        Coerce x into this multivariate polynomial ring.
        """
        if isinstance(x, multi_polynomial.MPolynomial) and x.parent() == self:
            return x
        if isinstance(x, polydict.PolyDict):
            return multi_polynomial.MPolynomial_polydict(self, x)
        c = self.base_ring()(x)
        return multi_polynomial.MPolynomial_polydict(self, {self.__zero_tuple:c})

if __name__ ==  '__main__':
    import doctest, sys
    from sage.rings.rational_field import RationalField
    from sage.rings.rings import GF
    PolynomialRing = MPolynomialRing
    doctest.testmod(sys.modules[__name__])
