"""
Polylist functionality -- an implementation of the underlying
 arithmetic for multi-variate polynomial rings.
"""

#*****************************************************************************
#       Copyright (C) 2005 David Joyner (wdj@usna.edu)
#                          William Stein (was@math.harvard.edu)
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************


import sage.rings.arith as arith
from sage.misc.misc import cputime

"""
SUMMARY:
    The functions in this file use the "distributive representation"
    of multivariate polynomials

        [[c1,[e1,...,er]],...] <-> c1*x1^e1*...*xr^er+...,

    which we call a polylist.   This file implements arithmetic
    functionality for polylists.

EXAMPLES:
    sage: polylist1 = [[2,[2,3]],[3,[1,2]],[4,[2,1]]]
    sage: polylist2 = [[-1,[1,3]],[-3,[1,2]],[2,[1,1]]]
    sage: P1 = PolyList(polylist1)
    sage: P2 = PolyList(polylist2)
    sage: P1+P2
    PolyList:[[2, [2, 3]], [4, [2, 1]], [-1, [1, 3]], [0, [1, 2]], [2, [1, 1]]]
    sage: P3 = P1+P2
    sage: print P3
    2x_0^2x_1^3 + 4x_0^2x_1 + -1x_0x_1^3x_0x_1^2 + 2x_0x_1
    sage: P4 = P1*P2
    sage: P4
    PolyList with distributtive representation:[[-2, [3, 6]], [-6, [3, 5]], [0, [3, 4]], [-3, [2, 5]], [-9, [2, 4]], [6, [2, 3]], [-12, [3, 3]], [8, [3, 2]]]
    sage: print P4
    -2x_0^3x_1^6 + -6x_0^3x_1^5x_0^3x_1^4 + -3x_0^2x_1^5 + -9x_0^2x_1^4 + 6x_0^2x_1^3 + -12x_0^3x_1^3 + 8x_0^3x_1^2

    sage: polylist5 = [[2,[1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20]],[3,[1,0,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20]],[4,[2,1,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20]]]
    sage: P5 = PolyList(polylist5)
    sage: P5
    PolyList with distributive representation:[[2, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20]], [3, [1, 0, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20]], [4, [2, 1, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20]]]
    sage: print P5
    2x_0x_1^2x_2^3x_3^4x_4^5x_5^6x_6^7x_7^8x_8^9x_9^10x_10^11x_11^12x_12^13x_13^14x_14^15x_15^16x_16^17x_17^18x_18^19x_19^20 + 3x_0x_1^0x_2^3x_3^4x_4^5x_5^6x_6^7x_7^8x_8^9x_9^10x_10^11x_11^12x_12^13x_13^14x_14^15x_15^16x_16^17x_17^18x_18^19x_19^20 + 4x_0^2x_1x_2^3x_3^4x_4^5x_5^6x_6^7x_7^8x_8^9x_9^10x_10^11x_11^12x_12^13x_13^14x_14^15x_15^16x_16^17x_17^18x_18^19x_19^20


    sage: from sage.rings.finite_feild import GF
    sage: F = GF(5) a = F.gen()
    sage: polylist1 = [[2,[2,3]],[3,[1,2]],[4,[2,1]]]
    sage: polylist2  =[[-1,[1,3]],[-3,[1,2]],[2,[1,1]]]
    sage: P1 = PolyList(polylist1, F, "YY")
    sage: P2 = PolyList(polylist2, F, "YY")
    sage: P1 * P2
    ???
    sage: P1 + P2
    ???
    sage: R = MPolynomialRing(F,5,"t")
    sage: R
    ???

AUTHORS:  David Joyner and William Stein 
"""

def dirac_delta(i,j):
    """
    INPUT:
        i,j, objects
    OUTPUT:
        int -- 1 if i==j, 0 otherwise
    """
    return int(i==j)


def flatten(lists):
    """
    used to convert [[row1],[row2],...,[rowk]] matrix notation
    to [row1,...,rowk] (in fact, row sizes need not be the same)
 
    EXAMPLES:
        sage: flatten([[1,2],[3,4]])
        [1, 2, 3, 4]
    """
    return sum(lists,[])

def list_union(self, other):
    """
    INPUT:
        self -- list
        other -- list
    OUTPUT:
        list -- their union, with order preserved
    """
    res = list(self)  
    for x in other:
       if not(x in res):
           res.append(x)
    return res

def list_intersection(self, other):
    """
    INPUT:
        self -- list
        other -- list
    OUTPUT:
        list --  their intersection (again, order preserved)
    """ 
    res = list(self)
    cap = []
    for x in other:
        if x in res:
            cap.append(x)
    return cap


def list_filtered(L,boolprop):
    """
    INPUT:
        L -- list
        boolprop -- a Boolean-valued function on the elements of list
    OUTPUT:
        list
    """
    return [x for x in L if boolprop(x)]

def list_positions(L,x):
    """
    INPUT:
        L -- list
        x -- anything
    OUTPUT:
        list -- the list of indices i where x = L[i]
    """
    ans = []
    n = len(L)
    for i in range(n):
        if x==L[i]:
            ans.append(i)
    return ans
    # WS: If you do this more than a few times on a single list it
    # WS: would be better that this list is sorted, and then one
    # WS: would use sage.misc.search

def list_total(L):
    """
    INPUT:
        L -- list
    OUTPUT:
        the sum of the entries of L
    """
    return sum(L,0)   # WS: The extra 0 is so the empty list returns a sum of 0.

def vector_add(list1,list2):
    """
    INPUT: two lists of numbers of the same length
        list1 -- list
        list2 -- list
        
    OUTPUT:
        number -- the component-wise sum, as though list1 and list2 were vectors
    """
    n = len(list1)           # does not check if lengths agree
    return [list1[i]+list2[i]  for i in range(n)]


def lex_monord(left,right):
    """
    INPUT:  left and right are monomial polylists
        left -- list
        right -- list
    OUTPUT:
        function -- comp fucntion (i.e., takes values 1, 0, -1) on polylists

    EXAMPLES:
        sage: polylist1 = [[2,[2,3]],[3,[1,2]],[4,[2,1]]]
        sage: polylist1.sort(lex_monord)
        sage: polylist1
        [[2, [2, 3]], [4, [2, 1]], [3, [1, 2]]]
    """
    p = left[1]
    q = right[1]
    if p == q:
        return 0
    ans = p < q
    return int(2*int(ans)-1)

def standard_basis_vector(a,i,n):
    """
    Returns list of length n with the integer a in (i-1)th position and zeros elsewhere.

    INPUT:
        a -- int
        i -- int, >= 1
        n -- int
    OUTPUT:
        list
    """
    v = [0]*n
    v[i-1] = a
    return v
    # WAS: The above three lines are about 4 times faster (depending on exact parameters)
    # than the code below:
    # return [a*int(j==i-1) for j in range(n)]

def degrevlex_monord(left,right):
    # Do the linear algebra import during function to avoid circular import problems
    import sage.linalg.vector_space, sage.linalg.matrix_space, sage.rings.rings
    p = left[1]
    q = right[1]
    r = len(p)  # assume this  =  len(q)
    V = sage.linalg.vector_space.VectorSpace(sage.rings.rings.RationalField(),r)
    p = V.vector(p)
    q = V.vector(q)
    M = sage.linalg.matrix_space.MatrixSpace(sage.rings.rings.RationalField(),r,r)
    B = [[1 for i in range(r)]]
    for i in range(r-1):
        B.append(standard_basis_vector(-1,r-i,r))

    A = M.matrix(flatten(B))  # This is the matrix repn of the degrevlex
    At = A.transpose()
    p = p*At                  # monomial ordering - see p52 of 
    q = q*At                  # Computational Commutative Algebra 1,
                              # M. Kreuzer, L. Robbiano.
    if p == q:
        return 0
    ans = (p < q)
    return int(2*int(ans)-1)

def coeffs_polylist(polylist):
    n = len(polylist)
    coeffs = [polylist[i][0] for i in range(n)]
    return coeffs

def expons_polylist(polylist):
    n = len(polylist)
    expons = [polylist[i][1] for i in range(n)]
    return expons


class PolyList(object):
    def __init__(self, plist, vars="x", morder=lex_monord, zero=0, remove_zero=False):
        """
        INPUT:
            plist -- list, which represents a multi-variable polynomial
                     with the distribute representation
            base_ring -- the ring in which the coefficients lie
            vars -- list; list of variable names, each a string
            morder -- function: a monomial order
        """
        if not isinstance(plist, list):
            if isinstance(plist, dict):
                plist = [[c,list(e)] for e,c in plist.iteritems() if c != zero]
            else:
                raise TypeError, "plist (=%s) must be a list."%plist
        if remove_zero:
            plist = [x for x in plist if x[0] != zero]
        n = len(plist)
        self.__dist_repn  = list(plist)
        self.__zero = zero
        if isinstance(vars, str):
            if len(plist) == 0:
                vars = [vars]
            else:
                m = len(plist[0][1])
                vars = ["%s_%s"%(vars, i) for i in range(m)]


        # this is the the distributive representation, with entries "coerced" into the base ring
        self.__variables = vars
        self.__monomial_ordering = morder

    def dict(self):
        """
        Return a dict that defines self.  
        """
        x = {}
        for v in self.__dist_repn:
            c, e = v[0], tuple(v[1])
            if x.has_key(e):
                x[e] += c
            else:
                x[e] = c
        return x

    def coefficients(self):
        return [t[0] for t in self.__dist_repn]

    def exponents(self):
        return [t[1] for t in self.__dist_repn]

    def _set_repn(self):
        try:
            return self.__set_repn
        except AttributeError:
            self.__set_repn = set(self.__dist_repn)
        
    def concat(self, value):           # value: list, Set...
        S = self._set_repn()
        for x in value:                # removes duplicates
            if not x in S:
                self.__dist_repn.append(x)

    def __len__(self):
        return len(self.__dist_repn)        # len(self)

    def __getitem__(self, i):
        return self.__dist_repn[int(i)]        # self[i]

    def __repr__(self):
        return 'PolyList with distributive representation %s'%self.__dist_repn

    def poly_repr(self):
        vars = self.__variables
        n = len(vars)
        poly = ""
        for c, e in self.__dist_repn:
            is_const_term = set(e) == set([0])
            if c != 0:
                sign_switch = False
                # First determine the multinomial:
                multi = ""
                for j in range(n):
                    if e[j] != 0:
                        if len(multi) > 0:
                            multi += "*"
                        multi += vars[j]
                        if e[j] != 1:
                            multi += "^%s"%e[j]
                # Next determine coefficient of multinomial
                if len(multi) == 0:
                    multi = str(c)
                elif c != 1:
                    if len(poly) > 0 and c == -1:
                        sign_switch = True
                    else:
                        multi = "%s*%s"%(c,multi)

                # Now add on coefficiented multinomials
                if len(poly) > 0:
                    if sign_switch:
                        poly += " - "
                    else:
                        poly += " + "
                poly += multi
        poly = poly.lstrip().rstrip()
        if len(poly) == 0:
            return "0"
        return poly

    def _add_polylists(left, right):
        """
        Uses the "distributive representation" of polynomials
        left = [[c1,[e1,...,er]],...] <-> c1*x1^e1*...*xr^er+...

        """
        zero = left.__zero
        D = left.dict()
        R = right.dict()
        for e,c in R.iteritems():
            if D.has_key(e):
                D[e] += c
            else:
                D[e] = c
        F = PolyList(D, left.__variables, morder=left.__monomial_ordering, zero=zero)
        return F

    def __add__(self, other):
        if not isinstance(other, PolyList):
            raise TypeError, "other (=%s) must be a PolyList."%other
        return self._add_polylists(other) # self + other

    def _mult_polylists(left, right):
        """
        Uses the "distributive representation" of polynomials
        left = [[c1,[e1,...,er]],...] <-> c1*x1^e1*...*xr^er+...
        """
        if not isinstance(right, PolyList):
            raise TypeError, "other (=%s) must be a PolyList."%other
        zero = left.__zero
        vars = left.__variables
        mord = left.__monomial_ordering
        newpoly = {}
        for lpoly in left.__dist_repn:
            for rpoly in right.__dist_repn:
                e = tuple(vector_add(lpoly[1],rpoly[1]))
                c = lpoly[0]*rpoly[0]
                if not newpoly.has_key(e):
                    newpoly[e] = c
                else:
                    newpoly[e] += c
        return PolyList(newpoly, vars, mord, zero)

    def __mul__(self, other):
        if not isinstance(other, PolyList):
            raise TypeError, "other (=%s) must be a PolyList."%other
        return self._mult_polylists(other)

    def scalar_mult(self, s):
        v = [[x[0]*s ,x[1]] for x in self.__dist_repn]
        return PolyList(v, self.__variables, self.__monomial_ordering, self.__zero)

    def __sub__(self, other):
        if not isinstance(other, PolyList):
            raise TypeError, "other (=%s) must be a PolyList."%other
        return self + other.scalar_mult(-1)

    def _nvars(self):
        return len(self.__variables)

    def _one(self):
        v = [[self.__zero+1, [0]*self._nvars()]]
        return PolyList(v, self.__variables,
                        self.__monomial_ordering, self.__zero)

    def __pow__(self, n):
        return arith.generic_power(self, int(n), self._one())


# doctests
if __name__ ==  '__main__':
    import doctest, sys
    # Run the doctests.
    doctest.testmod(sys.modules[__name__])
