"""
Univariate Polynomial Rings
"""


#*****************************************************************************
#       Copyright (C) 2005 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************


import random
import weakref

from ring import Ring
import sage.structure.gens as gens
import polynomial 
import rational_field
import integer_ring
from sage.libs.pari import pari

_objsPolynomialRing = {}
class _uniqPolynomialRing(object):
    def __new__(cls, base_ring, variable="x", sparse=False):
        key = (base_ring, variable, sparse)
        if _objsPolynomialRing.has_key(key):
            x = _objsPolynomialRing[key]()
            if x != None: return x
        try:
            if not base_ring.is_ring():
                raise TypeError, "Base ring (=%s) must be a Ring."%base_ring
        except AttributeError:
            raise TypeError, "Base ring (=%s) must be a Ring."%base_ring            
        try:
            if base_ring.is_finite():
                R = object.__new__(PolynomialRing_finite_base)
            else:
                R = object.__new__(cls)
        except NotImplementedError:
            R = object.__new__(cls)
            
        _objsPolynomialRing[key] = weakref.ref(R)
        return R

class PolynomialRing(_uniqPolynomialRing, Ring, gens.Gens):
    """
    Univariate polynomial ring.
    
    EXAMPLES:
        sage: import rings
        sage: Q = rings.RationalField()
        sage: R = PolynomialRing(Q)
        sage: x = R.gen()
        sage: R(-1) + R(1)
        0
        sage: (x - '2/3')*(x**2 - 8*x + 16)
        x^3 - 26/3*x^2 + 64/3*x - 32/3
    """
    def __init__(self, base_ring, variable="x", sparse=False):
        try:
            if not base_ring.is_ring():
                raise TypeError, "Base ring (=%s) must be a Ring."%base_ring
        except AttributeError:
            raise TypeError, "Base ring (=%s) must be a Ring."%base_ring            
        self.__base_ring = base_ring
        self.__variable = variable
        self.__is_sparse = sparse
        Ring.__init__(self)
        self.__set_polynomial_class()
        self.__generator = self.polynomial([0,1], is_gen=True)
        self.__cyclopoly_cache = {}
        
    def __call__(self, *args, **kwds):
        return self.__polynomial_class(self, *args, **kwds)

    def __cmp__(self, other):
        if not isinstance(other, PolynomialRing):
            return -1
        if self.__variable == other.__variable and \
               self.__base_ring == other.__base_ring:
            return 0
        elif self.__variable < other.__variable:
            return -1
        return +1
        
    def __contains__(self, x):
        return isinstance(x, polynomial.Polynomial) and x.parent() == self
        
    def __repr__(self):
        s = "Univariate Polynomial Ring in %s over %s"%(
                self.__variable, self.__base_ring)
        if self.is_sparse():
            s = "Sparse " + s
        return s
        
    def __set_polynomial_class(self):
        if self.__is_sparse:
            self.__polynomial_class = polynomial.Polynomial_generic_sparse
        else:
            if isinstance(self.base_ring(), rational_field.RationalField):
                self.__polynomial_class = polynomial.Polynomial_rational_dense
            elif isinstance(self.base_ring(), integer_ring.IntegerRing):
                self.__polynomial_class = polynomial.Polynomial_integer_dense
            else:
                self.__polynomial_class = polynomial.Polynomial_generic_dense
            
    def base_ring(self):
        return self.__base_ring
            
    def characteristic(self):
        return self.__base_ring.characteristic()
        
    def cyclotomic_polynomial(self, n):
        """
        The nth cyclotomic polynomial.

        EXAMPLES:
            sage: R = PolynomialRing(QQ)
            sage: R.cyclotomic_polynomial(8)
            x^4 + 1
            sage: R.cyclotomic_polynomial(12)
            x^4 - x^2 + 1
            sage: S = PolynomialRing(GF(7))
            sage: S.cyclotomic_polynomial(12)
            x^4 + 6*x^2 + 1
        """
        if n <= 0:
            raise ArithmeticError, "n=%s must be positive"%n
        f = pari.polcyclo(n)
        C = self.__polynomial_class 
        if C == polynomial.Polynomial_rational_dense:
            return self(f, construct=True)
        coeffs = str(f.Vec())
        if C == polynomial.Polynomial_integer_dense:
            return self(coeffs, construct=True)

        coeffs = eval(coeffs)
        return self(coeffs, check=True)

    def gen(self, n=0):
        """
        If this is R[x], this intrinsic returns x.
        """
        if n != 0:
            raise IndexError, "Generator %s not defined."%n
        return self.__generator
            
    def is_field(self):
        return False
        
    def is_sparse(self):
        return self.__is_sparse
        
    def name(self):
        return "poly-%s-%s"%(self.__variable,self.__base_ring)
    
    def ngens(self):
        return 1
        
    def polynomial(self, *args, **kwds):
        return self.__polynomial_class(self, *args, **kwds)
    
    def random(self, degree, bound=10):
        """
        Return a random polynomial.
        
        INPUT:
            degree -- an int
            bound -- an int

        OUTPUT:
            Polynomial -- A polynomial whose coefficients of x^i, for
            i up to degree, are coercisions to the base ring of random
            integers between -bound and bound.
           
        """
        return self([random.randrange(-bound,bound) for _ in xrange(degree+1)])

    def variable(self):
        """
        Returns the string which is used to print the generator of the polynomial ring.
        OUTPUT:
            str -- generator name
        """
        return self.__variable

    def set_variable(self, variable):
        self.__variable = variable

class PolynomialRing_finite_base(PolynomialRing):
    def __init__(self, base_ring, variable="x", sparse=False):
        PolynomialRing.__init__(self, base_ring, variable, sparse)

    def random(self, degree, bound=None):
        """
        Return a random polynomial of the given degree.
        
        INPUT:
            degree -- an int
            bound -- default = None; if specified, entries are integers between -bound and bound

        OUTPUT:
            Polynomial
        """
        if bound != None:
            return PolynomialRing.random(self, degree, bound)
        K = self.base_ring()
        return self([K.random() for _ in xrange(degree+1)])


def polygen(base_ring, var="x"):
    if isinstance(var, list):
        return polygens(base_ring, var)
    return PolynomialRing(base_ring, var).gen()

def polygens(base_ring, vars, n=1):
    if not isinstance(vars, list):
        if not isinstance(vars, str):
            n = int(vars)
            v = "x"
        else:
            v = str(vars)
        vars = ["%s%s"%(v,i) for i in xrange(n)]
    if len(vars) == 0:
        raise IndexError, "must be at least one var"
    x = polygen(base_ring, vars[0])
    GENS = [x]
    for v in vars[1:]:
        x = polygen(x.parent(), v)
        GENS.append(x)
    return tuple(GENS)
        
    
# doctests
if __name__ ==  '__main__':
    import doctest, sys
    from sage.rings.rings import GF, Q as QQ
    doctest.testmod(sys.modules[__name__])
    
