"""
The field of real numbers.
"""

##################################################################
# (Implementation not done.  Read about Decimal type
# here: http://www.python.org/dev/doc/devel/whatsnew/node8.html)
#  -- or -- use GMP?  MPFR ??
##################################################################

import gmpy

import ring
import sage.rings.real_number
from sage.libs.pari import pari

class RealField(ring.Ring, object):
    """
    The field of real numbers.
    """

    def __init__(self):
        """
        EXAMPLES:
            sage: R = RealField_mpf(); R
            Multi-precision Real Field
            sage: R('1/3')
            0.333333333333333333333

        Note that the second argument is the number of *bits* of precision,
        not the number of digits of precision:
            sage.: R('1/3',100)
            0.3333333333333333333333333333333333333333
            sage.: R('1/3',200)
            0.333333333333333333333333333333333333333333333333333333333333333333333

        If we create a real without quotes, we loose precision, because the real is
        turned into a Python float:
            sage: R(0.333333333333333333333)
            0.33333333333333331483

        We can also coerce rational numbers and integers into R, but
        coercing a polynomial in raising an exception.
            sage: Q = RationalField()
            sage: R(Q('1/3'))
            0.333333333333333333333
            sage: S = PolynomialRing(Q)
            sage: R(S.gen())
            Traceback (most recent call last):
            ...
            TypeError: unable to coerce x to a RealNumber_mpf
            sage: R.is_field()
            True
            sage: R.characteristic()
            0
            sage: R.name()
            'R'
            sage: R == R
            True
            sage: R == 5
            False
        """
        pass
    
    def __new__(cls, *args, **kwds):
        if cls != RealField:
            return object.__new__(cls, *args ,**kwds)
        return RealField_mpf(*args, **kwds)

    def is_atomic_repr(self):
        return True

    def __call__(self, x, prec=0):
        raise NotImplementedError

    def __cmp__(self, other):
        raise NotImplementedError

    def __repr__(self):
        return "Real Field"

    def characteristic(self):
        """
        Returns the characteristic of the real field, which is 0.
        """
        return 0
        
    def is_field(self):
        """
        Returns True, since the real field is a field.
        """
        return True

    def name(self):
        """
        Returns a short string 'R' that describes the real field.
        """
        return "R"
        
    def pi(self, n=0):
        """
        Returns pi to (at least) the given precision.

        EXAMPLES:
            sage: RealField().pi(3)
            3.14
        """
        import sage.functions.transcendental
        n = int(n)
        return self(sage.functions.transcendental.pi(n), n)

    def euler_constant(self, n=0):
        """
        Returns Euler's constant
            gamma = 0.57721566... = lim 1 + 1/2 + 1/3 + 1/4 + 1/5 + ... + 1/m - log m.
        to at most 219 decimal digits of precision.

        IMPLEMENTATION:
            PARI C library.
            
        EXAMPLES:
            sage: RealField_mpf().euler_constant()
            0.577215664901532865549
        """
        n = int(n)
        p = n/4 + 1
        return self(pari.euler(p).python(p), n)



_obj_RR_decimal = {}
class _uniq_RR_decimal(object):
    def __new__(cls):
        if _obj_RR_decimal.has_key(0):
            return _obj_RR_decimal[0]
        O = object.__new__(cls)
        _obj_RR_decimal[0] = O
        return O
        
class RealField_decimal(RealField, _uniq_RR_decimal):
    """
    Representation of the real field using the Python decimal type.
    """
    def __init__(self):
        """
        EXAMPLES:
            sage: R = RealField_decimal(); S = RealField_mpf()
            sage: R(1) + S(1)
            2.0
            sage: S(1.1) - R(1.2)
            -0.0999999999999999111822
            sage: R.pi() * S.pi()
            9.86965056
            sage: R(2)*S(2)
            4.0
            sage: S(2) / R(3)
            0.666666666666666666667
            sage: R(2) < S(3)
            True
            sage: S(2.99) > R(3)
            False
        """
    def __call__(self, x, prec=0):
        return sage.rings.real_number.RealNumber_decimal(x)

    def __cmp__(self, other):
        if isinstance(other, RealField_decimal):
            return 0
        return -1
    
    def __repr__(self):
        return "Infinite Precision Real Field"


_obj_RR_mpf = {}
class _uniq_RR_mpf(object):
    def __new__(cls):
        if _obj_RR_mpf.has_key(0):
            return _obj_RR_mpf[0]
        O = object.__new__(cls)
        _obj_RR_mpf[0] = O
        return O
        
class RealField_mpf(RealField, _uniq_RR_mpf):
    """
    Representation of the real field using the GMP MPFR type.
    """

    def __init__(self):
        """
        EXAMPLES:
            sage: R = RealField_decimal(); S = RealField_mpf()
            sage: R(1) + S(1)
            2.0
            sage: R(2)*S(2)
            4.0
        """
    def __call__(self, x, prec=0):
        prec=int(prec)
        return sage.rings.real_number.RealNumber_mpf(x,prec)

    def __cmp__(self, other):
        if isinstance(other, RealField_mpf):
            return 0
        return -1
        
    def __repr__(self):
        return "Multi-precision Real Field"

RR_mpf = RealField_mpf()
RR_decimal = RealField_decimal()

# doctests
if __name__ ==  '__main__':
    import doctest, sys
    from rational_field import RationalField
    from polynomial_ring import PolynomialRing
    doctest.testmod(sys.modules[__name__])
    
