"""
The Stein-Watkins table of elliptic curves.

SAGE includes a subset of the Stein-Watkins table of elliptic curves.
This is a large table of curves of The unique instance of the class
SteinWatkinsDatabase() gives access to the database.

This database covers a wide range of conductors, but unlike
CremonaDatabase(), this database need not list all curves of a given
conductor.  It lists the curves whose coefficients aren't ``too
large'' (see [Stein-Watkins, Ants 5]).

As of July 2005, the database exported by SAGE goes up to conductor
200000.
"""

#*****************************************************************************
#       
#       SAGE: Copyright (C) 2005 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import bz2, os, random

import sage.ellcurve.constructor
import sage.tables.db   # very important that this be fully qualified
import sage.misc.misc

class SteinWatkinsIsogenyClass:
    def __init__(self, conductor):
        self.conductor = conductor
        
    def __repr__(self):
        return "Stein-Watkins isogeny class of conductor %s"%self.conductor

    def __len__(self):
        try:
            return len(self.curves)
        except AttributeError:
            return 0

    def __iter__(self):
        try:
            for E in self.curves:
                yield E
        except AttributeError:
            return


def _lines(s):
    while True:
        i = s.find("\n")
        if i == -1:
            yield ""
            return 
        line = s[:i]
        s = s[i+1:]
        yield line

        
    
    

class SteinWatkinsDatabase(sage.tables.db.Database):
    """
    The Stein-Watkins database of elliptic curves.

    EXAMPLES:
        sage: c = SteinWatkinsDatabase()
        sage: c
        The Stein-Watkins database of elliptic curves
        
        sage.: c[large number]
        ???
    """
    
    def __init__(self, read_only=True):
        """
        Initialize the database.

        INPUT:
            read_only -- bool (default: True), if True, then the
                         database is read_only and changes cannot be
                         commited to disk.
        """
        sage.tables.db.Database.__init__(self,
             name="stein-watkins-ecdb", read_only=read_only, thresh=1024)

    def __iter__(self):
        """
        Returns an the database.
        """
        return self.iter(11, self.largest_conductor()+1)
    
    def __getitem__(self, conductor):
        """
        Return all data about curves of given conductor in the database.
        """
        try:
            N = int(conductor)
        except TypeError:
            raise TypeError, "the conductor (=%s) must be an integer."%conductor
        if self.root.has_key(N):
            return sage.tables.db.Database.__getitem__(self, N)
        else:
            return []

    def __repr__(self):
        """
        String representation of this database.
        OUTPUT:
           str
        """
        return "The Stein-Watkins database of elliptic curves"

    def largest_conductor(self):
        """
        The largest conductor of a curve appearing in the database.
        OUTPUT:
            int -- largest conductor
        """
        if not self.has_key('largest_conductor'):
            return 0
        return sage.tables.db.Database.__getitem__(self, 'largest_conductor')

    def smallest_conductor(self):
        """
        The smallest conductor of a curve in the database (always 11).
        
        OUTPUT:
            int -- smallest conductor
        """
        if not self.has_key('largest_conductor'):
            return 0
        return 11

    def conductor_range(self):
        """
        Return the range of conductors that are covered by the database.
        
        OUTPUT:
            int -- smallest cond
            int -- largest
        """
        return self.smallest_conductor(), self.largest_conductor()

    def number_of_curves(self,  N=0):
        """
        Returns the number of curves stored in the database with conductor N.
        If N = 0, returns the total number of curves in the database.
        
        INPUT:
            N -- int
        OUTPUT:
            int

        EXAMPLES:
            sage: c = SteinWatkinsDatabase()
            sage: c.number_of_curves(11)
            3
            sage: c.number_of_curves(37)
            4
            sage: c.number_of_curves(990)
            38
        """
        if N == 0:
            if not self.has_key('number_of_curves'):
                return 0
            return sage.tables.db.Database.__getitem__(self,'number_of_curves')
        return sum(len(X.curves) for X in self[N])


    def number_of_isogeny_classes(self, N=0):
        """
        Returns the number of isogeny classes of curves in the database
        of conductor N.  If N is 0, return the total number of curves
        in the database. 

        INPUT:
            N -- int
        OUTPUT:
            int
        EXAMPLES:
            sage: c = SteinWatkinsDatabase()
            sage: c.number_of_isogeny_classes(11)
            1
            sage: c.number_of_isogeny_classes(37)
            2
        """
        if N == 0:
            if not self.has_key('number_of_isogeny_classes'):
                return 0
            return sage.tables.db.Database.__getitem__(self,'number_of_isogeny_classes')
        return len(self[N])


    def random(self):
        """
        Returns a random curve from the database.
        """
        N = random.randint(11, self.largest_conductor())
        A = self.allcurves(N)
        while len(A) == 0:
            N += 1
            if N > self.largest_conductor():
                N = 11
            A = self.allcurves(N)
        id = A.keys()[random.randrange(len(A))]
        return self.elliptic_curve(str(N) + id)


    ###############################################################################
    # Functions for loading data from the a.xxx src files,
    # which should be in $SAGE_HOME/data/src/stein-watkins-ecdb
    ###############################################################################
    def _read(self, file=None):
        if file == None:
            for f in ["a.000.bz2", "a.001.bz2"]:
                self._read(f)
            return
        name = "%s/data/src/stein-watkins-ecdb/%s"%(os.environ["SAGE_HOME"],file)
        print "Stein-Watkins database: reading %s"%name
        numiso = self.number_of_isogeny_classes()
        numcurves = self.number_of_curves()
        raw = open(name).read()
        data = bz2.decompress(raw)
        for L in data.split('\n'):
            if len(L) == 0: continue
            if L[0] != '[': # new curve
                numiso += 1
                x = L.split()
                N = int(x[0])
                #if N > 5000: break
                C = SteinWatkinsIsogenyClass(N)
                if self.root.has_key(N):
                    self[N] = self[N]  # so database knows it changed
                    self[N].append(C)
                else:
                    self[N] = [C]
                if N%100==0:
                    print "conductor = %s, numcurves = %s"%(N, numcurves)
                C.rank = int(x[2])
                C.leading_coefficient = x[3]
                C.isogeny_number = x[4]
                C.modular_degree = x[5]
                C.data = x[4:]
                C.curves = []
            else:
                w = L.split()
                C.curves.append([eval(w[0]), w[1], w[2], w[3]])
                numcurves += 1
                self[N] = self[N]

        print "Stein-Watkins database: committing %s"%name
        print "N=",N
        if N > self.largest_conductor():
            self['largest_conductor'] = N
        self['number_of_curves'] = numcurves
        self['number_of_isogeny_classes'] = numiso
        self.commit()
        


class SteinWatkinsData:
    """
    Class for iterating through one of the Stein-Watkins database files.
    """
    def __init__(self, num):
        num = int(num)
        if num < 0:
            raise RuntimeError, "num (=%s) must be a nonnegative integer"%num
        name = str(num)
        name = '0'*(3-len(name)) + name
        file = "%s/data/src/stein-watkins-ecdb/a.%s.bz2"%(os.environ["SAGE_HOME"],name)
        try:
            raw = open(file).read()
        except IOError:
            raise RuntimeError, "The Stein-Watkins data file %s is not installed."%file
        self.__data = bz2.decompress(raw)
        self.__iter = self.__iter__()

    def __repr__(self):
        return "Stein-Watkins Database Iterator"

    def __iter__(self, start=None, position=None):
        C = None
        if position != None:
            data = self.__data[position:]            
        elif start != None:
            i = self.__data.find("%s ["%start)
            if i == -1:
                raise RuntimeError, "the start level (%s) must be in the data"%start
            data = self.__data[i:]
        else:
            data = self.__data
        for L in data.split('\n'):
            if len(L) == 0:
                continue
            if L[0] != '[': # new curve
                if C != None:
                    yield C
                x = L.split()
                N = int(x[0])
                C = SteinWatkinsIsogenyClass(N)
                C.rank = int(x[2])
                C.leading_coefficient = x[3]
                C.isogeny_number = x[4]
                C.modular_degree = x[5]
                C.data = x[4:]
                C.curves = []
            else:
                w = L.split()
                C.curves.append([eval(w[0]), w[1], w[2], w[3]])
        yield C

    def iter_classes(self, start=None):
        return self.__iter__(start)

    def next(self):
        return self.__iter.next()

    def __contain__(self, conductor):
        """
        Return True if there is a curve of the given conductor
        in this table.
        """
        return self.__data.find("%s ["%conductor) != -1

    def __getitem__(self, N):
        """ 
        Return the curves of conductor N in this table.  (Very slow)
        """
        try:
            I = self.iter(N)
        except RuntimeError:
            return []
        X = []
        while True:
            C = I.next()
            if C.conductor > N:
                return X
            else:
                X.append(C)

    def iter_levels(self):
        """
        Iterate through the data in lists of all curves of each
        conductor.
        """
        iter = self.__iter__()
        C = []
        N = 0
        while True:
            E = iter.next()
            if E.conductor != N:
                if C != []:
                    yield C
                C = [E]
                N = E.conductor
            else:
                C.append(E)
        yield C

    def list(self, min_level=None, max_level=None):
        if min_level != None:
            i = self.__data.find("%s ["%int(min_level))
            if i != -1:
                iter = self.__iter__(position=i)
            else:
                iter = self.__iter__()
        v = []
        for C in iter:
            if not (max_level is None) and C.conductor > max_level:
                return v
            if not (min_level is None) and C.conductor < min_level:
                continue
            v.append(C)
        return v
           
        
class SteinWatkinsPrimeData(SteinWatkinsData):
    def __init__(self, num):
        num = int(num)
        if num < 0:
            raise RuntimeError, "num (=%s) must be a nonnegative integer"%num
        name = str(num)
        name = '0'*(2-len(name)) + name
        file = "%s/data/src/stein-watkins-ecdb/p.%s.bz2"%(os.environ["SAGE_HOME"],name)
        try:
            raw = open(file).read()
        except IOError:
            raise RuntimeError, "The Stein-Watkins data file %s is not installed."%file
        self._SteinWatkinsData__data = bz2.decompress(raw)
        self._SteinWatkinsData__iter = self.__iter__()

    def __repr__(self):
        return "Stein-Watkins Prime Conductor Database Iterator"




##############################################################
# For checking that documentation strings are correct.
##############################################################
if __name__ ==  '__main__':
    import doctest, sys
    doctest.testmod(sys.modules[__name__])
    
